#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"

int flt,fld,rpsf=6,sub=5,rphot,*circ[2],s_EXT=-1,ssky=0;
float rap=3,**cwt,ssh,ct,bg,sig=0;

float **opsf,**npsf=NULL,**isig2=NULL,**data=NULL;

//taken from numerical recipes;
#define ITMAX 1000
#define EPS 3.0e-8
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
float zbrent(float (*func)(float), float x1, float x2, float tol)
{
	int iter;
	float a,b,c,d=0,e=0,min1,min2;
	float fa,fb,fc,p,q,r,s,tol1,xm;

	a=x1;
	b=c=x2;
	fa=(*func)(a);
	fb=(*func)(b);
	if (fa==0) return a;
	if (fb==0) return b;
	while (fa>0) {
	   a+=(a-b);
	   fa=(*func)(a);
	}
	while (fb<0) {
	   b+=(b-a);
	   fb=(*func)(b);
	}
	fc=fb;
	for (iter=1;iter<=ITMAX;iter++) {
		if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
			c=a;
			fc=fa;
			e=d=b-a;
		}
		if (fabs(fc) < fabs(fb)) {
			a=b;
			b=c;
			c=a;
			fa=fb;
			fb=fc;
			fc=fa;
		}
		tol1=2.0*EPS*fabs(b)+0.5*tol;
		xm=0.5*(c-b);
		if (fabs(xm) <= tol1 || fb == 0.0) return b;
		if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
			s=fb/fa;
			if (a == c) {
				p=2.0*xm*s;
				q=1.0-s;
			} else {
				q=fa/fc;
				r=fb/fc;
				p=s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
				q=(q-1.0)*(r-1.0)*(s-1.0);
			}
			if (p > 0.0) q = -q;
			p=fabs(p);
			min1=3.0*xm*q-fabs(tol1*q);
			min2=fabs(e*q);
			if (2.0*p < (min1 < min2 ? min1 : min2)) {
				e=d;
				d=p/q;
			} else {
				d=xm;
				e=d;
			}
		} else {
			d=xm;
			e=d;
		}
		a=b;
		fa=fb;
		if (fabs(d) > tol1)
			b += d;
		else
			b += SIGN(tol1,xm);
		fb=(*func)(b);
	}
	printf("Maximum number of iterations exceeded in zbrent\n");
	return 0.0;
}
#undef SIGN
#undef ITMAX
#undef EPS

void merr(void) {
   printf("****Mem error\n");
   exit(-1);
}

void getpars(void) {
   int x,y;
   circ[0]=(int*)calloc(rphot*2+1,sizeof(int));
   circ[1]=(int*)calloc(rphot*2+1,sizeof(int));
   cwt=(float**)calloc(rphot*2+1,sizeof(float*));
   if (!circ[0] || !circ[1] || !cwt) merr();
   circ[0]+=rphot;
   circ[1]+=rphot;
   cwt+=rphot;
   for (x=-rphot;x<=rphot;x++) {
      circ[0][x]=50;
      circ[1][x]=-50;
      cwt[x]=(float*)calloc(rphot*2+1,sizeof(float));
      if (!cwt[x]) merr();
      cwt[x]+=rphot;
      for (y=-rphot;y<=rphot;y++) {
	 cwt[x][y]=rap+.465-sqrt(x*x+y*y);
	 if (cwt[x][y]<0) cwt[x][y]=0;
	 else if (cwt[x][y]>1) cwt[x][y]=1;
	 if (cwt[x][y]>0 && y<circ[0][x]) circ[0][x]=y;
	 if (cwt[x][y]>0 && y>circ[1][x]) circ[1][x]=y;
      }
   }
}

void getpsf(void) {
   FILE *pf;
   char fn[80];
   int tf,x,y,x1,y1,rad,px,py,N;
   float z;

   sprintf(fn,"%s/psfs/%s.6.5.psf",BASEDIR,filters[flt].suffix);
   if ((pf=fopen(fn,"rb"))==NULL) {
      printf("****File error: %s\n",fn);
      exit(-1);
   }
   for (tf=0;tf<fld;tf++) {
      float buffer[200];
      int r2;
      if (tf) r2=((10+2)/2)*2+1;
      else r2=((5+2)/2)*2+1;
      N=64*r2*r2;
      if (tf) r2=4*2+1;
      else r2=6*2+1;
      for (x=0;x<N;x++) fread(buffer,sizeof(float),r2*r2,pf);
   }
   opsf=(float**)calloc((2*rpsf+1),sizeof(float*));
   if (!opsf) merr();
   opsf+=rpsf;
   for (x1=-rpsf;x1<=rpsf;x1++) {
      opsf[x1]=(float*)calloc((2*rpsf+1),sizeof(float));
      if (!opsf[x1]) merr();
      opsf[x1]+=rpsf;
   }
   rad=(sub+2)/2;
   for (px=0;px<8;px++) for (py=0;py<8;py++) for (x=-rad;x<=rad;x++) for (y=-rad;y<=rad;y++) for (x1=-rpsf;x1<=rpsf;x1++) for (y1=-rpsf;y1<=rpsf;y1++) {
      ffread(&z,1,sizeof(float),pf);
      if (abs(x)!=rad && abs(y)!=rad) opsf[x1][y1]+=z;
   }
   fclose(pf);
   for (x1=-rpsf;x1<=rpsf;x1++) for (y1=-rpsf;y1<=rpsf;y1++) opsf[x1][y1]/=64*(2*rad-1)*(2*rad-1);
   sprintf(fn,"%s/psfs/%s.6.5.dpsf",BASEDIR,filters[flt].suffix);
   if ((pf=fopen(fn,"rb"))==NULL) {
      printf("No default PSF residual for %s\n",filters[flt].name);
      return;
   }
   else {
      int tf,r2;
      float z;
      for (tf=0;tf<fld;tf++) {
	 float buffer[15];
	 if (tf) r2=4;
	 else r2=6;
	 for (x=-r2;x<=r2;x++) fread(buffer,sizeof(float),2*r2+1,pf);
      }
      for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) {
	 ffread(&z,sizeof(float),1,pf);
	 opsf[x][y]+=z;
	 if (x==0 && y==0) printf("offset: %f\n",z);
      }
      fclose(pf);
   }
   printf("center: %f\n",opsf[0][0]);
}

void normpsf(void) {
   double av=0.,r2,r;
   int N=0,x,y;

   r=rpsf;
   if (r>rphot+2) r=rphot+2;
   for (x=-r;x<=r;x++) for (y=-r;y<=r;y++) {
      r2=sqrt(x*x+y*y);
      if (r2>=rphot+1 && r2<=r) {
	 av+=npsf[x][y];
	 N++;
      }
   }
   av/=N;
   for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) npsf[x][y]-=av;
}

inline double ngauss(double FWHM) {return 2.7725887/FWHM/FWHM;}
inline double gauss(double r2) {return exp(-r2);}
inline double nhubble(double FWHM) {return 4.0000000/FWHM/FWHM;}
inline double hubble(double r2) {return 1./(1.+r2);}
inline double nmoffat15(double FWHM) {return 2.3496042/FWHM/FWHM;}
inline double moffat15(double r2) {return pow(1.+r2,-1.5);}
inline double nmoffat20(double FWHM) {return 1.6568542/FWHM/FWHM;}
inline double moffat20(double r2) {double x=1.+r2; return 1./x/x;}
inline double nmoffat25(double FWHM) {return 1.2780316/FWHM/FWHM;}
inline double moffat25(double r2) {return pow(1.+r2,-2.5);}
double (*func)(double),(*nfunc)(double);

void makenpsf(float wtry) {
   int x,y;
   float r2,t,w;

   if (npsf==NULL) {
      npsf=(float**)calloc((2*rpsf+1),sizeof(float*));
      data=(float**)calloc((2*rpsf+1),sizeof(float*));
      isig2=(float**)calloc((2*rpsf+1),sizeof(float*));
      if (!npsf || !data || !isig2) merr();
      npsf+=rpsf;
      data+=rpsf;
      isig2+=rpsf;
      for (x=-rpsf;x<=rpsf;x++) {
	 npsf[x]=(float*)calloc((2*rpsf+1),sizeof(float));
	 data[x]=(float*)calloc((2*rpsf+1),sizeof(float));
	 isig2[x]=(float*)calloc((2*rpsf+1),sizeof(float));
	 if (!npsf[x] || !data[x] || !isig2[x]) merr();
	 npsf[x]+=rpsf;
	 data[x]+=rpsf;
	 isig2[x]+=rpsf;
      }
   }
   w=nfunc(wtry);
   if (sig>0) {
      float w0,t0;
      t=0;
      t0=0;
      w0=nfunc(sig);
      for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) {
	 r2=x*x+y*y;
	 t0+=func(r2*w0);
	 t+=func(r2*w);
      }
      for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) {
	 r2=x*x+y*y;
	 npsf[x][y]=opsf[x][y]-func(r2*w0)/t0+func(r2*w)/t;
      }
      if (ssky) normpsf();
      return;
   }
   t=0;
   for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) {
      r2=x*x+y*y;
      t+=func(r2*w);
   }
   for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) {
      r2=x*x+y*y;
      npsf[x][y]=func(r2*w)/t;
   }
   if (ssky) normpsf();
   //for (x=-rpsf;x<=rpsf;x++) {for (y=-rpsf;y<=rpsf;y++) printf(" %5d",(int)(100000*npsf[x][y])); printf("\n");}
}

float calcdc(float ct0) {
   int x,y,r3;
   float A=0.,B=0.,tx;

   srand(s_EXT+1);
   rand();
   for (x=-rpsf;x<=rpsf;x++) for (y=-rpsf;y<=rpsf;y++) {
      data[x][y]=npsf[x][y]*ct0;
      if (s_EXT>=0) {
	 if (data[x][y]>0) tx=sqrt(data[x][y]+bg);
	 else tx=sqrt(bg);
	 if ((r3=rand()%3)==0) data[x][y]+=tx;
	 else if (r3==2) data[x][y]-=tx;
      }
      if (data[x][y]>0) isig2[x][y]=1./(data[x][y]+bg);
      else isig2[x][y]=1./bg;
   }
   for (x=-rphot;x<=rphot;x++) for (y=circ[0][x];y<=circ[1][x];y++) {
      A+=data[x][y]*opsf[x][y]*isig2[x][y]*cwt[x][y];
      B+=opsf[x][y]*opsf[x][y]*isig2[x][y]*cwt[x][y];
   }
   return A/B-ct;
}

float calcsh1(float w,float*sh,float*ct0) {
   int x,y;
   float pm,A;

   makenpsf(w);
   //printf("p'-p=%f\n",npsf[0][0]-opsf[0][0]);
   *ct0=zbrent(&calcdc,ct*0.5,ct*1.5,0.0001);
   //printf("ct0=%e\n",(*ct0));
   pm=A=0;
   for (x=-rphot;x<=rphot;x++) for (y=circ[0][x];y<=circ[1][x];y++) {
      pm+=opsf[x][y]*opsf[x][y]*isig2[x][y]*cwt[x][y];
      A+=opsf[x][y]*isig2[x][y]*cwt[x][y];
   }
   pm/=A;
   //printf("<p>=%f\n",pm);
   *sh=A=0;
   for (x=-rphot;x<=rphot;x++) for (y=circ[0][x];y<=circ[1][x];y++) {
      (*sh)+=(data[x][y]/ct-opsf[x][y])*(opsf[x][y]-pm)*sqrt(isig2[x][y])*cwt[x][y];
      A+=(opsf[x][y]-pm)*(opsf[x][y]-pm)*sqrt(isig2[x][y])*cwt[x][y];
   }
   (*sh)/=A;
   //printf("sh=%f\n",(*sh));
   //printf("%f %f %f\n",w,*ct0,*sh);
   return ssh-(*sh);
}

#define NIT 100
#define INIT 0.01
float calcsh(float w,float*av,float*sd,float*ctav,float*ctsd) {
   int i;
   float tsh,tct,y,z,rval=0.;

   *av=*sd=*ctav=*ctsd=0.;
   s_EXT=-1;
   calcsh1(w,&tsh,&tct);
   for (i=0;i<NIT;i++) {
      s_EXT=i;
      rval+=calcsh1(w,&y,&z)*INIT;
      (*av)+=y*INIT;
      (*sd)+=(y-tsh)*(y-tsh)*INIT;
      (*ctav)+=z*INIT;
      (*ctsd)+=(z-tct)*(z-tct)*INIT;
   }
   *sd=sqrt(*sd);
   *ctsd=sqrt(*ctsd);
   //printf("%7.5f %7.4f %7.4f\n",w,*av,*sd);
   //printf("%f %f\n",w,rval);
   return rval;
}

float calcshw(float w) {
   float a=0,b=0,c=0,d=0;
   return calcsh(w,&a,&b,&c,&d);
}

int main(int argc,char**argv) {
   float w0,w1,s2,a=0,b=0,c=0,d=0;
   int x,y;
   if (argc!=9) {
      printf("Usage: %s <filt> <chp> <sh> <cts(e)> <sky(e)> <RN(e)> <ssky> <fpsf>\n",*argv);
      return -1;
   }
   ssky=atoi(argv[7]);
   if (!strcasecmp(argv[8],"gauss")) {
      func=&gauss;
      nfunc=&ngauss;
   }
   else if (!strcasecmp(argv[8],"hubble")) {
      func=&hubble;
      nfunc=&nhubble;
   }
   else if (!strcasecmp(argv[8],"moffat15")) {
      func=&moffat15;
      nfunc=&nmoffat15;
   }
   else if (!strcasecmp(argv[8],"moffat20")) {
      func=&moffat20;
      nfunc=&nmoffat20;
   }
   else if (!strcasecmp(argv[8],"moffat25")) {
      func=&moffat25;
      nfunc=&nmoffat25;
   }
   else {
      printf("Unknown psf shape\n");
      return -1;
   }
   initfilters();
   flt=findfilt(argv[1]);
   fld=atoi(argv[2]);
   if (fld<0 || fld>3) {
      printf("Chip must be 0-3\n");
      return -1;
   }
   if (fld) {
      rpsf=(rpsf*2+1)/3;
      sub*=2;
      rap*=2./3.;
   }
   rphot=(int)(rap+0.9);
   getpars();
   getpsf();
   ct=atof(argv[4]);
   bg=atof(argv[5])+atof(argv[6])*atof(argv[6]);
   ssh=0;
   if (!fld) {
      w0=0.7;
      w1=1.3;
   }
   else {
      w0=0.5;
      w1=1.1;
   }
   printf(" sh=0 FWHM is %f\n",s2=zbrent(&calcshw,w0,w1,0.0001));
   makenpsf(s2);
   a=0;
   for (x=-rphot;x<=rphot;x++) for (y=-rphot;y<=rphot;y++) a+=(npsf[x][y]-opsf[x][y])*(npsf[x][y]-opsf[x][y]);
   printf("   scatter is %f\n",sqrt(a)/(2*rphot+1.));
   sig=s2;
   ssh=atof(argv[3]);
   w0=sig*0.7;
   w1=sig*1.5;
   printf(" Best FWHM is %f\n",s2=zbrent(&calcshw,w0,w1,0.0001));
   printf("delta FWHM is %f\n",s2-sig);
   calcsh(s2,&a,&b,&c,&d);
   printf("  sharp = %7.4f +/- %6.4f\n",a,b);
   printf("   dmag = %7.4f +/- %6.4f\n",-2.5*log10(c/ct),2.5*d/c/log(10));
   return 0;
}
