#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"

typedef float imtype[800][800];
typedef char fntype[41];
typedef double mtype[3];
typedef double double2[2];
typedef int int2[2];

int *DMAX,*DMIN,*TEMP,N,min,nrej[4]={0,0,0,0};
double *EXP,*EXP0,*GAIN,*RN,*EPOCH,TGAIN,c2,nsig;
imtype *data,*sky;
fntype *FILT,TFILT;
mtype *moff;
double2 *dpos;
int2 *tdpos;

int ferr(FILE *f,char *str) {
   printf("%s\n",str);
   fclose(f);
   exit(1);
}

FILE *readfits(char *fn,int n) {
   FILE *f;
   char str[82];
   int MODE=0,line=0,tmp,HPP=0;

   if ((f=fopen(fn,"rb"))==NULL) {
      printf("File \"%s\" not found\n",fn);
      exit(1);
   }
   while ((!feof(f)) && ((!MODE) || (line%36))) {
      fgets(str,81,f);
      line++;
      if (!MODE) {
	 if (!strncmp(str,"END     ",8)) MODE=1;
	 else if ((!strncmp(str,"SIMPLE  =",9)) && (str[29]!='T')) ferr(f,"Illegal format type");
	 else if ((!strncmp(str,"BITPIX  =",9)) && (strncmp(str+27,"-32",3))) ferr(f,"Illegal bits per pixel");
	 else if ((!strncmp(str,"NAXIS   =",9)) && (strncmp(str+27,"  3",3))) ferr(f,"Illegal number of axes");
	 else if (!strncmp(str,"NAXIS",5)) {
	    tmp=atoi(str+9);
	    if (str[5]=='1' && tmp!=800) ferr(f,"Illegal X axis size");
	    else if (str[5]=='2' && tmp!=800) ferr(f,"Illegal Y axis size");
	    else if (str[5]=='3' && tmp!=4) ferr(f,"Illegal Z axis size");
	 }
	 else if (!strncmp(str,"ATODGAIN=",9)) GAIN[n]=atof(str+9);
	 else if (!strncmp(str,"RNOISE  =",9)) RN[n]=atof(str+9);
	 else if (!strncmp(str,"EPOCH   =",9)) EPOCH[n]=atof(str+9);
	 else if (!strncmp(str,"EXPTIME =",9)) EXP[n]=atof(str+9);
	 else if (!strncmp(str,"EXP0    =",9)) EXP0[n]=atof(str+9);
	 else if (!strncmp(str,"RSDPFILL=",9)) DMIN[n]=atoi(str+9);
	 else if (!strncmp(str,"SATURATE=",9)) DMAX[n]=atoi(str+9);
	 else if (!strncmp(str,"TEMP    =",9)) TEMP[n]=atoi(str+9);
	 else if (!strncmp(str,"FILTNAM1=",9)) {
	    int i=0;
	    while ((i<19) && (str[i+11]!=32)) {
	       FILT[n][i]=str[i+11];
	       i++;
	    }
	    FILT[n][i]=0;
	    moff[n][0]=pow(10,0.4*(21-filters[findfilt(FILT[n])].zero));
	    moff[n][1]=moff[n][2]=1/moff[n][0];
	 }
	 else if (!strncmp(str,"HSTPHOT =",9)) HPP=1;
      }
   }
   if (!HPP) ferr(f,"Not pre-processed by mask");
   printf("Reading FITS file %s:\n",fn);
   printf("Exp time=%f s, Gain=%f e/ADU, Noise=%f ADU\n",EXP[n],GAIN[n],RN[n]);
   printf("Saturation at %d, bad value=%d\n\n",DMAX[n],DMIN[n]);
   return f;
}

FILE *writefits(char*fn,int n) {
   FILE *f;
   char str[81];
   int line;

   if ((f=fopen(fn,"wb"))==NULL) {
      printf("Error on output file\n");
      exit(1);
   }
   fputs("SIMPLE  =                    T / Standard FITS format                           ",f);
   fputs("BITPIX  =                  -32 / 32 bit IEEE floating point numbers             ",f);
   fputs("NAXIS   =                    3 / Number of axes                                 ",f);
   fputs("NAXIS1  =                  800 /                                                ",f);
   fputs("NAXIS2  =                  800 /                                                ",f);
   fputs("NAXIS3  =                    4 /                                                ",f);
   fputs("EXTEND  =                    T / There may be standard extensions               ",f);
   sprintf(str,"ATODGAIN=       %14f / Analog to Digital Gain (Electrons/DN)          ",GAIN[n]);
   fputs(str,f);
   sprintf(str,"RNOISE  =       %14f / Readout Noise (DN)                             ",RN[n]);
   fputs(str,f);
   sprintf(str,"EPOCH   =     %16.8f / exposure average time (Modified Julian Date)   ",EPOCH[n]);
   fputs(str,f);
   sprintf(str,"EXPTIME =       %14f / exposure duration (seconds)--calculated        ",EXP[n]);
   fputs(str,f);
   sprintf(str,"EXP0    =       %14f / weighted average initial exposure time         ",EXP0[n]);
   fputs(str,f);
   sprintf(str,"RSDPFILL=       %14d / bad data fill value for calibrated images      ",DMIN[n]);
   fputs(str,f);
   sprintf(str,"SATURATE=       %14d / Data value at which saturation occurs          ",DMAX[n]);
   fputs(str,f);
   sprintf(str,"TEMP    =       %14d / Temperature (0=cold, 1=warm)                   ",TEMP[n]);
   fputs(str,f);
   sprintf(str,"FILTNAM1= \'%-18s\' / first filter name                              ",FILT[n]);
   fputs(str,f);
   fputs("HSTPHOT =                    T / Preprocessed by HSTphot/mask                   ",f);
   fputs("END                                                                             ",f);
   for (line=18;line<36;line++) fputs("                                                                                ",f);
   return f;
}

void cleanit(int fld) {
   int i,j,k,l,n,cn,jj,kk,x0,x1,y0,y1,x,y;
   double hi,lo,mid,t,*list,*cmplist,*mult;
   double gr[4]={1.987,2.003,2.006,1.955};

   mult=(double*)calloc(N,sizeof(double));
   list=(double*)calloc(N,sizeof(double));
   cmplist=(double*)calloc(N,sizeof(double));
   if (!mult || !list || !cmplist) {
      printf("Memory allocation error\n");
      exit(-1);
   }
   x0=y0=0;
   x1=y1=800;
   for (i=0;i<N;i++) {
      double dx=0,dy=0;
      if (GAIN[i]==TGAIN) mult[i]=moff[i][0];
      else if (GAIN[i]>TGAIN) mult[i]=gr[fld]*moff[i][0];
      else mult[i]=moff[i][0]/gr[fld];
      switch(fld) {
      case 0:
	 dx=dpos[i][0];
	 dy=dpos[i][1];
	 break;
      case 1:
	 dx=-dpos[i][1]*0.457;
	 dy=dpos[i][0]*0.457;
	 break;
      case 2:
	 dx=-dpos[i][0]*0.457;
	 dy=-dpos[i][1]*0.457;
	 break;
      case 3:
	 dx=dpos[i][1]*0.457;
	 dy=-dpos[i][0]*0.457;
	 break;
      }
      tdpos[i][0]=(int)(dx+1000.5)-1000;
      tdpos[i][1]=(int)(dy+1000.5)-1000;
      if (-tdpos[i][0]<x0) x0=-tdpos[i][0];
      if (-tdpos[i][0]+800>x1) x1=-tdpos[i][0]+800;
      if (-tdpos[i][1]<y0) y0=-tdpos[i][1];
      if (-tdpos[i][1]+800>y1) y1=-tdpos[i][1]+800;
   }
   for (j=x0;j<x1;j++) for (k=y0;k<y1;k++) {
      if (k==y0 && ((j-x0)*10)%(x1-x0)<10) {
	 printf(".");
	 fflush(stdout);
      }
      n=0;
      for (i=0;i<N;i++) {
	 x=j+tdpos[i][0];
	 y=k+tdpos[i][1];
	 if (x>=0 && x<800 && y>=0 && y<800) {
	    if (data[i][x][y]>DMIN[i] && data[i][x][y]<DMAX[i]*0.855) {
	       t=(data[i][x][y]-sky[i][x][y])/EXP[i]*mult[i];
	       for (l=n;l>0 && list[l-1]>t;l--) list[l]=list[l-1];
	       list[l]=t;
	       n++;
	    }
	    else if (data[i][x][y]<=DMIN[i]) {
	       data[i][x][y]=DMIN[i];
	       nrej[3]++;
	    }
	    else {
	       data[i][x][y]=DMAX[i];
	       nrej[2]++;
	    }
	 }
      }
      if (n==0) mid=0;
      else if (min) mid=list[0];
      else mid=list[(n-1)/2];
      cn=0;
      for (i=0;i<N;i++) {
	 t=-2.e30;
	 for (jj=j-1+tdpos[i][0];jj<=j+1+tdpos[i][0];jj++) for (kk=k-1+tdpos[i][1];kk<=k+1+tdpos[i][1];kk++) if (jj>=0 && jj<800 && kk>=0 && kk<800 && data[i][jj][kk]>DMIN[i] && data[i][jj][kk]<DMAX[i]*0.855 && (data[i][jj][kk]-sky[i][jj][kk])/EXP[i]*mult[i]>t) t=(data[i][jj][kk]-sky[i][jj][kk])/EXP[i]*mult[i];
	 if (t>-1.e30) {
	    for (l=cn;l>0 && cmplist[l-1]>t;l--) cmplist[l]=cmplist[l-1];
	    cmplist[l]=t;
	    cn++;
	 }
      }
      if (cn==0) hi=mid;
      else if (min) hi=cmplist[0];
      else hi=cmplist[(cn-1)/2];
      if (hi<mid) hi=mid;
      cn=0;
      for (i=0;i<N;i++) {
	 t=2.e30;
	 for (jj=j-1+tdpos[i][0];jj<=j+1+tdpos[i][0];jj++) for (kk=k-1+tdpos[i][1];kk<=k+1+tdpos[i][1];kk++) if (jj>=0 && jj<800 && kk>=0 && kk<800 && data[i][jj][kk]>DMIN[i] && data[i][jj][kk]<DMAX[i]*0.855 && (data[i][jj][kk]-sky[i][jj][kk])/EXP[i]*mult[i]<t) t=(data[i][jj][kk]-sky[i][jj][kk])/EXP[i]*mult[i];
	 if (t<1.e30) {
	    for (l=cn;l>0 && cmplist[l-1]>t;l--) cmplist[l]=cmplist[l-1];
	    cmplist[l]=t;
	    cn++;
	 }
      }
      if (cn==0) lo=mid;
      else if (min) lo=cmplist[0];
      else lo=cmplist[(cn-1)/2];
      if (lo>mid) lo=mid;
      for (i=0;i<N;i++) {
	 x=j+tdpos[i][0];
	 y=k+tdpos[i][1];
	 if (x>=0 && x<800 && y>=0 && y<800 && data[i][x][y]>DMIN[i] && data[i][x][y]<DMAX[i]) {
	    double min,max,dmin,dmax;
	    min=lo*EXP[i]*moff[i][1]+sky[i][x][y];
	    if (min>0) dmin=nsig*nsig*min/GAIN[i]+c2*c2*min*min;
	    else dmin=0;
	    dmin=sqrt(dmin+RN[i]*RN[i]*nsig*nsig);
	    max=hi*EXP[i]*moff[i][2]+sky[i][x][y];
	    if (max>0) dmax=nsig*nsig*max/GAIN[i]+c2*c2*max*max;
	    else dmax=0;
	    dmax=sqrt(dmax+RN[i]*RN[i]*nsig*nsig);
	    if (data[i][x][y]<min-dmin) {
	       data[i][x][y]=DMIN[i];
	       if (x>=50 && x<795 && y>=50 && y<795) nrej[1]++;
	    }
	    else if (data[i][x][y]>max+dmax) {
	       data[i][x][y]=DMIN[i];
	       if (x>=50 && x<795 && y>=50 && y<795) nrej[0]++;
	    }
	 }
      }
   }
   free(list);
   free(cmplist);
   free(mult);
}

int main(int argc,char**argv) {
   int i,fld,x,y;
   FILE **fdata,**fsky,**fout;
   fntype *fns;

   if (argc<10 || (argc-4)%6) {
      printf("Usage: %s <reg factor> <#sigma> <min?>\n  (<fits base> <scale> <minscale> <maxscale> <dx> <dy>) * nimg\n",*argv);
      printf("Recommend reg=1, sig=3, min=0\n");
      return 1;
   }
   N=(argc-4)/6;
   fdata=(FILE**)calloc(N,sizeof(FILE*));
   fsky=(FILE**)calloc(N,sizeof(FILE*));
   fout=(FILE**)calloc(N,sizeof(FILE*));
   fns=(fntype*)calloc(N,sizeof(fntype));
   DMAX=(int*)calloc(N,sizeof(int));
   DMIN=(int*)calloc(N,sizeof(int));
   TEMP=(int*)calloc(N,sizeof(int));
   EXP=(double*)calloc(N,sizeof(double));
   EXP0=(double*)calloc(N,sizeof(double));
   GAIN=(double*)calloc(N,sizeof(double));
   RN=(double*)calloc(N,sizeof(double));
   EPOCH=(double*)calloc(N,sizeof(double));
   data=(imtype*)calloc(N,sizeof(imtype));
   sky=(imtype*)calloc(N,sizeof(imtype));
   FILT=(fntype*)calloc(N,sizeof(fntype));
   moff=(mtype*)calloc(N,sizeof(mtype));
   dpos=(double2*)calloc(N,sizeof(double2));
   tdpos=(int2*)calloc(N,sizeof(int2));
   if (!fdata || !fsky || !fout || !DMAX || !DMIN || !TEMP || !EXP || !EXP0 || !GAIN || !RN || !EPOCH || !data || !sky || !FILT || !moff || !dpos || !tdpos) {
      printf("Memory allocation error\n");
      return 1;
   }
   c2=atof(argv[1]);
   nsig=atof(argv[2]);
   min=atoi(argv[3]);
   initfilters();
   for (i=0;i<N;i++) {
      sprintf(fns[i],"%s.fits",argv[4+6*i]);
      fdata[i]=readfits(fns[i],i);
      sprintf(fns[i],"%s.sky",argv[4+6*i]);
      fsky[i]=fopen(fns[i],"rb");
      if (fsky[i]==NULL) {
	 printf("Cannot open %s; setting to zero\n",fns[i]);
	 for (x=0;x<800;x++) for (y=0;y<800;y++) sky[i][x][y]=0;
      }
      sprintf(fns[i],"cleansep%d_tmp.%d",(int)getpid(),i);
      fout[i]=writefits(fns[i],i);
      moff[i][0]*=pow(10,0.4*atof(argv[5+6*i]));
      moff[i][1]/=pow(10,0.4*atof(argv[7+6*i]));
      moff[i][2]/=pow(10,0.4*atof(argv[6+6*i]));
      dpos[i][0]=atof(argv[9+6*i]);
      dpos[i][1]=atof(argv[8+6*i]);
   }
   TGAIN=14;
   for (i=0;i<N;i++) if (GAIN[i]==7) TGAIN=7;
   for (fld=0;fld<4;fld++) {
      for (i=0;i<N;i++) ffread(data[i],800*800,4,fdata[i]);
      for (i=0;i<N;i++) if (fsky[i]) ffread(sky[i],800*800,4,fsky[i]);
      cleanit(fld);
      for (i=0;i<N;i++) ffwrite(data[i],800*800,4,fout[i]);
   }
   printf("\n");
   for (i=0;i<N;i++) {
      FILE *f;
      char str[401];
      fclose(fdata[i]);
      if (fsky[i]) fclose(fsky[i]);
      fclose(fout[i]);
      str[0]=0; f=NULL;
      sprintf(str,"mv %s.fits %s.old.fits ; mv %s %s.fits",argv[4+6*i],argv[4+6*i],fns[i],argv[4+6*i]);
      if ((f=popen(str,"w"))!=NULL) pclose(f);
      else printf("Cannot replace old fits file %s\n",argv[4+6*i]);
   }
   printf("%7d (%7.4f%%) rejected as CRs\n",nrej[0],(double)nrej[0]/22201./N);
   printf("%7d (%7.4f%%) rejected as low\n",nrej[1],(double)nrej[1]/22201./N);
   printf("%7d (%7.4f%%) rejected as sat\n",nrej[2],(double)nrej[2]/22201./N);
   printf("%7d (%7.4f%%) rejected as bad\n",nrej[3],(double)nrej[3]/22201./N);
   return 0;
}
