#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "filters.h"
#define BORDER 50
#define TSIZE 900

typedef double ddouble[2];
typedef struct {int f;double x,y; ddouble *m0,**m;} stype;
typedef struct {int f,tp;double x,y,mag,sh,chi;} ptype;

int Nfilt=0,N=0,*NT,*filt,reject,ref[4][TSIZE][TSIZE],pref[4][TSIZE][TSIZE];
double maxChi,minSh,maxSh;
double reftime,**dt;
stype *stars;
ptype *phot;

int readref(char *fn0) {
   FILE *f;
   char str[161],fn[41];
   int i,j,k;

   sprintf(fn,"%s.phot",fn0);
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open reference photometry %s\n",fn);
      return -1;
   }
   fgets(str,161,f);
   Nfilt=atoi(str);
   if (Nfilt<1) {
      printf("Illegal format for reference phot file\n");
      return -1;
   }
   filt=(int*)calloc(sizeof(int),Nfilt);
   NT=(int*)calloc(sizeof(int),Nfilt);
   dt=(double**)calloc(sizeof(double),Nfilt);
   if (!filt || !NT || !dt) {
      printf("Memory allocation error\n");
      return -1;
   }
   initfilters();
   for (i=0;i<Nfilt;i++) {
      fgets(str,161,f);
      str[strlen(str)-1]=0;
      filt[i]=findfilt(str);
   }
   while (fgets(str,161,f)) N++;
   fclose(f);
   stars=(stype*)calloc(sizeof(stype),N);
   if (!stars) {
      printf("Memory allocation error\n");
      return -1;
   }
   for (i=0;i<4;i++) for (j=0;j<TSIZE;j++) for (k=0;k<TSIZE;k++) ref[i][j][k]=-1;
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open reference photometry %s\n",fn);
      return -1;
   }
   for (i=0;i<=Nfilt;i++) fgets(str,161,f);
   for (i=0;i<N;i++) {
      stars[i].m0=(ddouble*)calloc(sizeof(ddouble),Nfilt);
      stars[i].m=(ddouble**)calloc(sizeof(ddouble*),Nfilt);
      if (!stars[i].m0 || !stars[i].m) {
	 printf("Memory allocation error\n");
	 return -1;
      }
      fscanf(f,"%d %lf %lf",&stars[i].f,&stars[i].x,&stars[i].y);
      for (j=0;j<Nfilt;j++) fscanf(f,"%lf %lf",stars[i].m0[j],stars[i].m0[j]+1);
      fscanf(f,"%d",&j);
      j=(int)(stars[i].x+BORDER);
      k=(int)(stars[i].y+BORDER);
      if (j<0 || k<0 || j>=TSIZE || k>=TSIZE) {
	 printf("%d %6.2f %6.2f out of range\n",stars[i].f,stars[i].x,stars[i].y);
	 return -1;
      }
      else if (ref[stars[i].f][j][k]!=-1) {
	 printf("Too many stars at %d %6.2f %6.2f\n",stars[i].f,stars[i].x,stars[i].y);
	 return -1;
      }
      ref[stars[i].f][j][k]=i;
   }
   fclose(f);
   return 0;
}

double parsetime(char *str) {
   int t[6]={0,0,0,0,0,0},i;
   char *ptr;
   struct tm ttime;

   ptr=str;
   for (i=0;i<6;i++) {
      while ((*ptr<'0' || *ptr>'9') && *ptr) ptr++;
      if (!*ptr) {
	 printf("Illegal time format \"%s\"\n",str);
	 exit(1);
      }
      while (*ptr>='0' && *ptr<='9') {
	 t[i]=t[i]*10+(*ptr)-'0';
	 ptr++;
      }
   }
   ttime.tm_sec=t[5];
   ttime.tm_min=t[4];
   ttime.tm_hour=t[3];
   ttime.tm_mday=t[2];
   ttime.tm_mon=t[1]-1;
   ttime.tm_year=t[0]-1900;
   ttime.tm_isdst=0;
   return mktime(&ttime)/3600.;
}

int checkphot(char *fn,int i,int indx) {
   int isep,x0,x1,y0,y1,x,y,j;
   double sep,r;
   char estr[81];

   if (phot[i].f) {
      sep=10;
      isep=4;
   }
   else {
      sep=20;
      isep=5;
   }
   x0=(int)(phot[i].x+BORDER-isep);
   if (x0<0) x0=0;
   x1=(int)(phot[i].x+BORDER+isep);
   if (x1>=TSIZE) x1=TSIZE-1;
   y0=(int)(phot[i].y+BORDER-isep);
   if (y0<0) y0=0;
   y1=(int)(phot[i].y+BORDER+isep);
   if (y1>=TSIZE) y1=TSIZE-1;
   for (x=x0;x<=x1;x++) for (y=y0;y<=y1;y++) if ((j=pref[phot[i].f][x][y])>0 && j!=i && (r=(phot[j].x-phot[i].x)*(phot[j].x-phot[i].x)+(phot[j].y-phot[i].y)*(phot[j].y-phot[i].y))<=sep) {
      r=sqrt(r);
      estr[0]=0;
      if (phot[j].mag<=phot[i].mag-.75) strcpy(estr,"brighter companion");
      else if (phot[j].mag<=phot[i].mag+.75) strcpy(estr,"comparable companion");
      else if (phot[j].sh<-.5 || phot[j].tp==5) strcpy(estr,"extended companion");
      else if (phot[j].sh>.5 || phot[j].tp==4) strcpy(estr,"CR companion");
      else if (phot[j].chi>3) strcpy(estr,"bad companion");
      if (estr[0]) {
	 if (!reject) printf("%d %6.2f %6.2f %6d %s: %s (%d %6.2f %6.2f)\n",phot[i].f,phot[i].x,phot[i].y,indx,fn,estr,phot[j].f,phot[j].x,phot[j].y);
	 return 1;
      }
      else {
	 if (phot[j].mag<=phot[i].mag+2.5) strcpy(estr,"fainter companion");
	 else if (phot[j].sh<-.25) strcpy(estr,"extended? companion");
	 else if (phot[j].sh>.25) strcpy(estr,"CR? companion");
	 else if (phot[j].chi>2) strcpy(estr,"bad? companion");
	 if (estr[0]) printf("%d %6.2f %6.2f %6d %s: %s (%d %6.2f %6.2f)\n",phot[i].f,phot[i].x,phot[i].y,indx,fn,estr,phot[j].f,phot[j].x,phot[j].y);
      }
   }
   return 0;
}

void readphot(char *fn,int i) {
   FILE *f;
   int sf,xx,yy,j,k,xb,yb,tp,N=0;
   char mark[4][TSIZE][TSIZE];
   double x,y,m,dm,q,r,rb,chi,sh;

   printf("Matching %s\n",fn);
   for (sf=0;sf<4;sf++) for (j=0;j<TSIZE;j++) for (k=0;k<TSIZE;k++) pref[sf][j][k]=0;
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open photometry file %s\n",fn);
      exit(0);
   }
   while (!feof(f)) {
      fscanf(f,"%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %d",&sf,&x,&y,&q,&q,&m,&q,&chi,&q,&sh,&tp);
      N++;
   }
   fclose(f);
   N--;
   phot=(ptype*)calloc(sizeof(ptype),N);
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open photometry file %s\n",fn);
      exit(0);
   }
   for (j=0;j<N;j++) {
      fscanf(f,"%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %d",&phot[j].f,&phot[j].x,&phot[j].y,&q,&q,&phot[j].mag,&q,&phot[j].chi,&q,&phot[j].sh,&phot[j].tp);
      pref[phot[j].f][(int)(phot[j].x+BORDER)][(int)(phot[j].y+BORDER)]=j;
   }
   fclose(f);
   for (sf=0;sf<4;sf++) for (j=0;j<TSIZE;j++) for (k=0;k<TSIZE;k++)
      if (ref[sf][j][k]>=0) mark[sf][j][k]=0;
      else mark[sf][j][k]=1;
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open photometry file %s\n",fn);
      exit(0);
   }
   N=0;
   fscanf(f,"%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %d",&sf,&x,&y,&q,&q,&m,&dm,&chi,&q,&sh,&tp);
   while (!feof(f)) {
      if (chi<=maxChi && sh>=minSh && sh<=maxSh && tp<4) {
	 j=(int)(x+BORDER);
	 k=(int)(y+BORDER);
	 xb=yb=-1;
	 rb=1000;
	 for (xx=j-2;xx<=j+2;xx++) for (yy=k-2;yy<=k+2;yy++) if (xx>=0 && xx<TSIZE && yy>=0 && yy<TSIZE && !mark[sf][xx][yy]) {
	    r=(x-stars[ref[sf][xx][yy]].x)*(x-stars[ref[sf][xx][yy]].x)+(y-stars[ref[sf][xx][yy]].y)*(y-stars[ref[sf][xx][yy]].y);
	    if (r<rb) {
	       rb=r;
	       xb=xx;
	       yb=yy;
	    }
	 }
	 if (xb>=0) {
	    mark[sf][xb][yb]=1;
	    j=ref[sf][xb][yb];
	    if (!checkphot(fn,N,j) || !reject) {
	       stars[j].m[i][NT[i]][0]=m-stars[j].m0[i][0];
	       stars[j].m[i][NT[i]][1]=dm;
	    }
	 }
      }
      N++;
      fscanf(f,"%d %lf %lf %lf %lf %lf %lf %lf %lf %lf %d",&sf,&x,&y,&q,&q,&m,&dm,&chi,&q,&sh,&tp);
   }
   fclose(f);
}

int readtimes(char *fn) {
   FILE *f;
   char str[161],*ptr,*suff;
   int i,j,k;

   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open time file %s\n",fn);
      return -1;
   }
   fgets(str,161,f);
   reftime=parsetime(str);
   while (fgets(str,161,f)) {
      suff=NULL;
      ptr=str;
      while (*ptr && *ptr!=' ') {
	 if (*ptr=='.') suff=ptr+1;
	 ptr++;
      }
      if (!*ptr) {
	 printf("Illegal file + time \"%s\"\n",str);
	 return -1;
      }
      *ptr=0;
      ptr++;
      if (!suff) {
	 printf("Illegal filename %s\n",str);
	 return -1;
      }
      for (i=0;i<Nfilt && strcmp(suff,filters[filt[i]].suffix);i++);
      if (i==Nfilt) {
	 printf("Illegal filename suffix %s\n",suff);
	 return -1;
      }
      NT[i]++;
   }
   fclose(f);
   for (i=0;i<Nfilt;i++) {
      dt[i]=(double*)calloc(sizeof(double),NT[i]);
      if (!dt[i]) {
	 printf("Memory allocation error\n");
	 exit(0);
      }
      for (j=0;j<N;j++) {
	 stars[j].m[i]=(ddouble*)calloc(sizeof(ddouble),NT[i]);
	 if (!stars[j].m[i]) {
	    printf("Memory allocation error\n");
	    exit(0);
	 }
	 for (k=0;k<NT[i];k++) stars[j].m[i][k][0]=99.0;
      }
      NT[i]=0;
   }
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Could not open time file %s\n",fn);
      return -1;
   }
   fgets(str,161,f);
   while (fgets(str,161,f)) {
      suff=NULL;
      ptr=str;
      while (*ptr && *ptr!=' ') {
	 if (*ptr=='.') suff=ptr+1;
	 ptr++;
      }
      if (!*ptr) {
	 printf("Illegal file + file + time \"%s\"\n",str);
	 return -1;
      }
      *ptr=0;
      ptr++;
      if (!suff) {
	 printf("Illegal filename %s\n",str);
	 return -1;
      }
      for (i=0;i<Nfilt && strcmp(suff,filters[filt[i]].suffix);i++);
      if (i==Nfilt) {
	 printf("Illegal filename suffix %s\n",suff);
	 return -1;
      }
      dt[i][NT[i]]=parsetime(ptr)-reftime;
      readphot(str,i);
      NT[i]++;
   }
   fclose(f);
   return 0;
}

int output(char *fn0) {
   FILE *f;
   char fn[41];
   int i,j,k;

   sprintf(fn,"%s.epochs",fn0);
   if ((f=fopen(fn,"w"))==NULL) {
      printf("Could not open output file %s\n",fn);
      return -1;
   }
   fprintf(f,"%d filters\n",Nfilt);
   for (i=0;i<Nfilt;i++) {
      fprintf(f,"%s\n%d epochs\n",filters[filt[i]].name,NT[i]);
      for (j=0;j<NT[i];j++) fprintf(f,"%10.4f hours offset\n",dt[i][j]);
   }
   for (i=0;i<N;i++) {
      fprintf(f,"%d %6.2f %6.2f",stars[i].f,stars[i].x,stars[i].y);
      for (j=0;j<Nfilt;j++) {
	 fprintf(f," %6.3f %6.3f",stars[i].m0[j][0],stars[i].m0[j][1]);
	 for (k=0;k<NT[j];k++) fprintf(f," %7.3f %6.3f",stars[i].m[j][k][0],stars[i].m[j][k][1]);
      }
      fprintf(f,"\n");
   }
   fclose(f);
   return 0;
}

int main(int argc,char**argv) {
   if (argc!=7) {
      printf("Usage: %s <phot base> <times> <chi> <min sh> <max sh> <reject>\n",*argv);
      return -1;
   }
   maxChi=atof(argv[3]);
   minSh=atof(argv[4]);
   maxSh=atof(argv[5]);
   reject=atoi(argv[6]);
   if (readref(argv[1])) return -1;
   if (readtimes(argv[2])) return -1;
   if (output(argv[1])) return -1;
   return 0;
}
