#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"

typedef float dtype[4][800][800];
typedef struct{int x,y,fx,fy;double s,ffx,ffy;} stype;
typedef float ****psftype;

dtype data;
psftype psf[4][8][8];
float GAIN=7.,RN=0.;
int rpsf[4],sub[4],fld,filt;
long seed=-1;
stype stars[1];

void merr(int i) {
   printf("Mem error %d\n",i);
   exit(0);
}

void getpars(void) {
   rpsf[1]=rpsf[2]=rpsf[3]=(rpsf[0]*2+1)/3;
   sub[1]=sub[2]=sub[3]=sub[0]*2;
}

void getpsf(void) {
   int f,x,y,z,rad,px,py;
   FILE *pf;
   int x1;
   char fn[80];

   for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
      rad=(sub[f]+2)/2;
      psf[f][px][py]=(float****)calloc(rad*2+1,sizeof(float***));
      if (psf[f][px][py]==NULL) merr(1);
      psf[f][px][py]+=rad;
      for (x=-rad;x<=rad;x++) {
	 psf[f][px][py][x]=(float***)calloc(rad*2+1,sizeof(float**));
	 if (psf[f][px][py][x]==NULL) merr(2);
	 psf[f][px][py][x]+=rad;
	 for (y=-rad;y<=rad;y++) {
	    psf[f][px][py][x][y]=(float**)calloc(2*rpsf[f]+1,sizeof(float*));
	    if (psf[f][px][py][x][y]==NULL) merr(3);
	    psf[f][px][py][x][y]+=rpsf[f];
	    for (z=-rpsf[f];z<=rpsf[f];z++) {
	       psf[f][px][py][x][y][z]=(float*)calloc(2*rpsf[f]+1,sizeof(float));
	       if (psf[f][px][py][x][y][z]==NULL) merr(4);
	       psf[f][px][py][x][y][z]+=rpsf[f];
	    }
	 }
      }
   }
   sprintf(fn,"%s/psfs/%s.%d.%d.psf",BASEDIR,filters[filt].suffix,rpsf[0],sub[0]);
   if ((pf=fopen(fn,"rb"))==NULL) {
      printf("File error: %s\n",fn);
      exit(0);
   }
   for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
      for (x=-(sub[f]+2)/2;x<=(sub[f]+2)/2;x++) for (y=-(sub[f]+2)/2;y<=(sub[f]+2)/2;y++) for (x1=-rpsf[f];x1<=rpsf[f];x1++) ffread(psf[f][px][py][x][y][x1]-rpsf[f],2*rpsf[f]+1,sizeof(float),pf);
   }
   fclose(pf);
}

/*
void psfcorr(char *str) {
   char fn[81];
   FILE *fpsf;
   int f,x,y,x2,y2,x3,y3,rad;
   float c[101];

   sprintf(fn,"%s.psf",str);
   if ((fpsf=fopen(fn,"rb"))==NULL) return;
   for (f=0;f<4;f++) for (x=-rpsf[f];x<=rpsf[f];x++) {
      rad=(sub[f]+2)/2;
      ffread(c,sizeof(float),2*rpsf[f]+1,fpsf);
      for (y=-rpsf[f];y<=rpsf[f];y++) for (x2=0;x2<8;x2++) for (y2=0;y2<8;y2++) for (x3=-rad;x3<=rad;x3++) for (y3=-rad;y3<=rad;y3++) psf[f][x2][y2][x3][y3][x][y]+=c[y+rpsf[f]];
   }
   fclose(fpsf);
}
*/

void imgsub(void) {
   int x1,y1,x0,y0;
   double q,mx,my;

   if (stars[0].ffx<0) {
      x0=stars[0].fx-1;
      mx=stars[0].ffx+1;
   }
   else {
      x0=stars[0].fx;
      mx=stars[0].ffx;
   }
   if (stars[0].ffy<0) {
      y0=stars[0].fy-1;
      my=stars[0].ffy+1;
   }
   else {
      y0=stars[0].fy;
      my=stars[0].ffy;
   }
   for (x1=stars[0].x-rpsf[fld];x1<=stars[0].x+rpsf[fld];x1++) for (y1=stars[0].y-rpsf[fld];y1<=stars[0].y+rpsf[fld];y1++) if (x1>=50 && x1<800 && y1>=50 && y1<800) {
      q=psf[fld][stars[0].y/100][stars[0].x/100][x0][y0][x1-stars[0].x][y1-stars[0].y]*(1-mx)*(1-my)+psf[fld][stars[0].y/100][stars[0].x/100][x0+1][y0][x1-stars[0].x][y1-stars[0].y]*mx*(1-my)+psf[fld][stars[0].y/100][stars[0].x/100][x0][y0+1][x1-stars[0].x][y1-stars[0].y]*(1-mx)*my+psf[fld][stars[0].y/100][stars[0].x/100][x0+1][y0+1][x1-stars[0].x][y1-stars[0].y]*mx*my;
      data[fld][x1][y1]+=stars[0].s*q;
   }
}

// from NR;
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran2(void)
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (seed <= 0) {
		if (-(seed) < 1) seed=1;
		else seed = -(seed);
		idum2=(seed);
		for (j=NTAB+7;j>=0;j--) {
			k=(seed)/IQ1;
			seed=IA1*(seed-k*IQ1)-k*IR1;
			if (seed < 0) seed += IM1;
			if (j < NTAB) iv[j] = seed;
		}
		iy=iv[0];
	}
	k=(seed)/IQ1;
	seed=IA1*(seed-k*IQ1)-k*IR1;
	if (seed < 0) seed += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0) idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = seed;
	if (iy < 1) iy += IMM1;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

double gammln(double xx) {
   double x,y,tmp,ser;
   static double cof[6]={76.18009172947146,-86.50532032941677,
			 24.01409824083091,-1.231739572450155,
			 0.1208650973866179e-2,-0.5395239384953e-5};
   int j;

   y=x=xx;
   tmp=x+5.5;
   tmp -= (x+0.5)*log(tmp);
   ser=1.000000000190015;
   for (j=0;j<=5;j++) ser += cof[j]/++y;
   return -tmp+log(2.5066282746310005*ser/x);
}

#define Nmax 32
double fact(int n) {
   static int ntop=4;
   static double a[Nmax+1]={1.0,1.0,2.0,6.0,24.0};
   int j;

   if (n<0) return 1;
   if (n>Nmax) return exp(gammln(n+1.0));
   while (ntop<n) {
      j=ntop++;
      a[ntop]=a[j]*ntop;
   }
   return a[n];
}
#undef Nmax

double gauss(void) {
   return (ran2()+ran2()-ran2()-ran2())/0.57735027;
}

int poisson(float m) {
   int n;
   double nmax;

   if (m>100) return (int)(m+gauss()*sqrt(m)+ran2());
   nmax=(int)(10+m+5*sqrt(m));
   while (1) {
      n=(int)(ran2()*nmax);
      if (ran2()<exp(-m)/fact(n)*pow(m,n)) return n;
   }
}

void readstars(char *fn) {
   FILE *f;
   double x,y;

   if ((f=fopen(fn,"r"))==NULL) {
      printf("Cannot find %s\n",fn);
      exit(0);
   }
   fscanf(f,"%d %lf %lf %lf",&fld,&y,&x,&stars[0].s);
   while (!feof(f)) {
      if (x<0) x=50.+750.*ran2();
      if (y<0) y=50.+750.*ran2();
      printf("%d %f %f %f\n",fld,y,x,stars[0].s);
      stars[0].x=(int)x;
      stars[0].fx=(int)((x-stars[0].x-.5)*sub[fld]+20.5)-20;
      stars[0].ffx=(x-stars[0].x-.5)*sub[fld]-stars[0].fx;
      stars[0].y=(int)y;
      stars[0].fy=(int)((y-stars[0].y-.5)*sub[fld]+20.5)-20;
      stars[0].ffy=(y-stars[0].y-.5)*sub[fld]-stars[0].fy;
      imgsub();
      fscanf(f,"%d %lf %lf %lf",&fld,&y,&x,&stars[0].s);
   }
   fclose(f);
}

int main(int argc,char**argv) {
   FILE *fout;
   char str[82];
   int x,y,line,RAND=0;
   float sky;

   if (argc!=7 && argc!=8) {
      printf("Usage: %s <filter> <stars> <PC background> <gain> <rd-noise> <output> <<noise>>\n",*argv);
      printf("stars: chip, X, Y, electrons\n");
      return 1;
   }
   rpsf[0]=6;
   sub[0]=5;
   getpars();
   GAIN=atof(argv[4]);
   RN=atof(argv[5]);
   if (argc==8 && atoi(argv[7])) RAND=1;
   initfilters();
   filt=findfilt(argv[1]);
   getpsf();
   //psfcorr(argv[1]);
   seed=time(NULL);
   if ((fout=fopen(argv[6],"wb"))==NULL) {
      printf("Cannot open %s\n",argv[6]);
      return 1;
   }
   fputs("SIMPLE  =                    T / Standard FITS format                           ",fout);
   fputs("BITPIX  =                  -32 / 32 bit IEEE floating point numbers             ",fout);
   fputs("NAXIS   =                    3 / Number of axes                                 ",fout);
   fputs("NAXIS1  =                  800 /                                                ",fout);
   fputs("NAXIS2  =                  800 /                                                ",fout);
   fputs("NAXIS3  =                    4 /                                                ",fout);
   fputs("EXTEND  =                    T / There may be standard extensions               ",fout);
   sprintf(str,"ATODGAIN= %20f / Analog to Digital Gain (Electrons/DN)          ",GAIN);
   fputs(str,fout);
   sprintf(str,"RNOISE  = %20f / Readout Noise (DN)                             ",RN/GAIN);
   fputs(str,fout);
   fputs("EPOCH   =                51000 / exposure average time (MJD)                    ",fout);
   fputs("EXPTIME =                 1200 / exposure duration (seconds)                    ",fout);
   fputs("EXP0    =                 1200 / weighted average initial exposure time (sec)   ",fout);
   fputs("RSDPFILL=                -1000 / bad data value for calibrated image            ",fout);
   fputs("SATURATE=              1000000 / data value at which saturation occurs          ",fout);
   fputs("TEMP    =                    0 / Temperature (0=cold, 1=warm)                   ",fout);
   sprintf(str,"FILTNAM1= \'%-18s\' /                                                ",argv[1]);
   fputs(str,fout);
   fputs("HSTPHOT =                    T / Preprocessed by HSTphot/mask                   ",fout);
   fputs("END                                                                             ",fout);
   for (line=18;line<36;line++) fputs("                                                                                ",fout);
   for (fld=0;fld<4;fld++) {
      if (fld==0) sky=atof(argv[3]);
      else sky=atof(argv[3])*4.7843157;
      for (x=0;x<800;x++) for (y=0;y<800;y++) data[fld][x][y]=-1100*GAIN;
      for (x=50;x<800;x++) for (y=50;y<800;y++) data[fld][x][y]=sky;
   }
   readstars(argv[2]);
   for (fld=0;fld<4;fld++) for (x=0;x<800;x++) for (y=0;y<800;y++) {
      if (data[fld][x][y]<=-1000*GAIN) data[fld][x][y]=-1001;
      else if (data[fld][x][y]>=1000000.*GAIN) data[fld][x][y]=1000001.;
      else {
	 if (RAND) data[fld][x][y]=poisson(data[fld][x][y])+RN*gauss();
	 data[fld][x][y]=(int)(data[fld][x][y]/GAIN+ran2());
	 if (data[fld][x][y]<=-1000) data[fld][x][y]=-1001;
	 else if (data[fld][x][y]>=1000000.) data[fld][x][y]=1000001.;
      }
   }
   ffwrite(data,sizeof(dtype),1,fout);
   fclose(fout);
   return 0;
}
