#include <string.h>

#define NFILTERS 20
typedef struct {
   char name[11],suffix[6],color;
   double zero;
   char xorder[8];
   double xform[8][5];
} filttype;

filttype filters[NFILTERS];

int findfilt(char*str) {
   int i;
   for (i=0;i<NFILTERS;i++) if (!strcmp(filters[i].name,str)) return i;
   printf("Illegal filter %s\n",str);
   exit(0);
}

void initfilters(void) {
   int i,j,n;
   FILE *f;
   char str[81];

   sprintf(str,"%s/filters.dat",BASEDIR);
   if ((f=fopen(str,"r"))==NULL) {
      printf("Could not find %s\n",str);
      exit(0);
   }
   for (i=0;i<NFILTERS;i++) {
      fgets(filters[i].name,11,f); filters[i].name[strlen(filters[i].name)-1]=0;
      filters[i].color=fgetc(f);
      fgets(filters[i].suffix,11,f); filters[i].suffix[strlen(filters[i].suffix)-1]=0;
      fscanf(f,"%lf",&filters[i].zero);
      fgets(str,11,f);
   }
   for (i=0;i<NFILTERS;i++) {
      for (j=0;j<8;j++) filters[i].xorder[j]='X';
      fscanf(f,"%d",&n);
      fgets(str,11,f);
      for (j=0;j<n;j++) {
	 filters[i].xorder[j]=fgetc(f);
	 fscanf(f,"%lf %lf %lf %lf %lf",filters[i].xform[j],filters[i].xform[j]+1,filters[i].xform[j]+2,filters[i].xform[j]+3,filters[i].xform[j]+4);
	 filters[i].xform[j][2]-=filters[i].zero;
	 fgets(str,11,f);
      }
   }
}

double findcolor(double ci[NFILTERS],char color,int *n) {
   int i;
   for (i=0;i<NFILTERS;i++) if (filters[i].color==color && ci[i]<99) {
      if (n!=NULL) *n=i;
      return ci[i];
   }
   return 99.999;
}

void transform(double ci0[NFILTERS],double dci[NFILTERS],double ci[NFILTERS]) {
   int f,i,j,CONT=1,it=0;
   double c,tm,tn,ciold[NFILTERS];

   memcpy(ci,ci0,8*NFILTERS);
   while (CONT) {
      CONT=0;
      it++;
      memcpy(ciold,ci,8*NFILTERS);
      for (f=0;f<NFILTERS;f++) ci[f]=99.999;
      for (f=0;f<NFILTERS;f++) if (ciold[f]<99) {
	 tm=tn=0;
	 for (i=0;i<8;i++) if (filters[f].xorder[i]!='X' && (c=findcolor(ciold,filters[f].xorder[i],&j))<99) {
	    if (j<f) c=c-ciold[f];
	    else c=ciold[f]-c;
	    if ((filters[f].xform[i][3]==-99 || c>filters[f].xform[i][3]) && (filters[f].xform[i][4]==99 || c<=filters[f].xform[i][4])) {
	       tm+=(filters[f].xform[i][0]*c+filters[f].xform[i][1]*c*c+filters[f].xform[i][2])/(dci[j]+0.01);
	       tn+=1./(dci[j]+0.01);
	    }
	 }
	 if (tn>0) {
	    ci[f]=ci0[f]+tm/tn;
	    if (it>5) ci[f]=(ci[f]*5+ciold[f]*(it-5))/it;
	    if (fabs(ci[f]-ciold[f])>0.001) CONT=1;
	 }
      }
   }
}

double untransform(int f,double ci[5]) {
   int i,j,f0=-1;
   double c,tm=0,tn=0;
   char UBVRI[6]="UBVRI";
   for (i=0;i<5;i++) if (filters[f].color==UBVRI[i]) f0=i;
   if (f0<0 || ci[f0]>90) return 99999.;
   for (i=0;i<8;i++) if (filters[f].xorder[i]!='X') {
      for (j=0;j<5;j++) if (filters[f].xorder[i]==UBVRI[j] && ci[j]<99) {
	 if (j<f0) c=ci[j]-ci[f0];
	 else c=ci[f0]-ci[j];
	 if ((filters[f].xform[i][3]==-99 || c>filters[f].xform[i][3]) && (filters[f].xform[i][4]==99 || c<=filters[f].xform[i][4])) {
	    tm-=filters[f].xform[i][0]*c+filters[f].xform[i][1]*c*c+filters[f].xform[i][2];
	    tn++;
	 }
      }
   }
   if (tn>0) return tm/tn+ci[f0];
   return 99999.;
}
