#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"
#define XMIN 50
#define XMAX 800
#define YMIN 50
#define YMAX 800
#define MMIN 0

#define NSIG0 1.5
#define NSIG1 1.6
#define NSIG 1.7

typedef float dtype[800][800];
typedef struct{int x,y,fx,fy,use,type;double s,aps,sig,ffx,ffy;} stype;
typedef float ****psftype;

stype *stars;
int X,Y,Z,DMAX,DMIN,MMAX=4;
float ran[800][800],cran[800][800],MIN=-1.e25,MAX=1.e25,Fmin;
float GAIN,RN,apcs[4];
dtype data,sky,res,noise;
psftype psf[4][8][8];
char FILT[20]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int rpsf[4],sub[4],Nstars,Napcor,fld,Gauss_PSF=0,filt;
double apsize[4]={10.979,5.020,5.021,5.018};
float dnoise[4]={0.0361,0.0625,0.0625,0.0625};

void merr(int i) {
   printf("Mem error %d\n",i);
   exit(0);
}

void getpars(void) {
   int x,y;
   for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) cran[x][y]=-100;
   rpsf[1]=rpsf[2]=rpsf[3]=(rpsf[0]*2+1)/3;
   sub[1]=sub[2]=sub[3]=sub[0]*2;
   stars=(stype*)calloc(Napcor+1,sizeof(stype));
   if (!stars) merr(0);
}

void readsky(char *fn) {
   int f,x,y;
   FILE *sf;

   if ((sf=fopen(fn,"rb"))==NULL) {
      if (fld==0) printf("Cannot open file %s\n",fn);
      for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) {
	 sky[x][y]=0;
	 noise[x][y]=RN*RN;
	 if (res[x][y]>0) noise[x][y]+=res[x][y]/GAIN;
      }
      return;
   }
   for (f=0;f<=fld;f++) ffread(sky,800*800,4,sf);
   for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) {
      res[x][y]-=sky[x][y];
      noise[x][y]=RN*RN;
      if (res[x][y]>0) noise[x][y]+=res[x][y]/GAIN;
      if (sky[x][y]>0) noise[x][y]+=sky[x][y]/GAIN;
   }
   fclose(sf);
}

void getpsf(void) {
   int f,x,y,z,rad,px,py;

   for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
      rad=(sub[f]+2)/2;
      psf[f][px][py]=(float****)calloc(rad*2+1,sizeof(float***));
      if (psf[f][px][py]==NULL) merr(1);
      psf[f][px][py]+=rad;
      for (x=-rad;x<=rad;x++) {
	 psf[f][px][py][x]=(float***)calloc(rad*2+1,sizeof(float**));
	 if (psf[f][px][py][x]==NULL) merr(2);
	 psf[f][px][py][x]+=rad;
	 for (y=-rad;y<=rad;y++) {
	    psf[f][px][py][x][y]=(float**)calloc(2*rpsf[f]+1,sizeof(float*));
	    if (psf[f][px][py][x][y]==NULL) merr(3);
	    psf[f][px][py][x][y]+=rpsf[f];
	    for (z=-rpsf[f];z<=rpsf[f];z++) {
	       psf[f][px][py][x][y][z]=(float*)calloc(2*rpsf[f]+1,sizeof(float));
	       if (psf[f][px][py][x][y][z]==NULL) merr(4);
	       psf[f][px][py][x][y][z]+=rpsf[f];
	    }
	 }
      }
   }
   if (Gauss_PSF) {
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 int x1,y1;
	 float sig,n;

	 rad=(sub[f]+2)/2;
	 if (f) sig=0.54;
	 else sig=0.84;
	 for (x=-rad;x<=rad;x++) for (y=-rad;y<=rad;y++) {
	    n=0;
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) n+=(psf[f][px][py][x][y][x1][y1]=exp(-.5*((x1-x/(float)sub[f])*(x1-x/(float)sub[f])+(y1-y/(float)sub[f])*(y1-y/(float)sub[f]))/sig/sig));
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) psf[f][px][py][x][y][x1][y1]/=n;
	 }
      }
   }
   else {
      FILE *pf;
      int x1;
      char fn[80];

      sprintf(fn,"%s/psfs/%s.%d.%d.psf",BASEDIR,filters[filt].suffix,rpsf[0],sub[0]);
      if ((pf=fopen(fn,"rb"))==NULL) {
	 printf("File error: %s\n",fn);
	 exit(0);
      }
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 for (x=-(sub[f]+2)/2;x<=(sub[f]+2)/2;x++) for (y=-(sub[f]+2)/2;y<=(sub[f]+2)/2;y++) for (x1=-rpsf[f];x1<=rpsf[f];x1++) ffread(psf[f][px][py][x][y][x1]-rpsf[f],2*rpsf[f]+1,sizeof(float),pf);
      }
      fclose(pf);
   }
}

void psfcorr(char *str) {
   char fn[81];
   FILE *fpsf;
   int f,x,y,x2,y2,x3,y3,rad;
   float c[101];

   sprintf(fn,"%s.psf",str);
   if ((fpsf=fopen(fn,"rb"))==NULL) return;
   for (f=0;f<4;f++) for (x=-rpsf[f];x<=rpsf[f];x++) {
      rad=(sub[f]+2)/2;
      ffread(c,sizeof(float),2*rpsf[f]+1,fpsf);
      for (y=-rpsf[f];y<=rpsf[f];y++) for (x2=0;x2<8;x2++) for (y2=0;y2<8;y2++) for (x3=-rad;x3<=rad;x3++) for (y3=-rad;y3<=rad;y3++) psf[f][x2][y2][x3][y3][x][y]+=c[y+rpsf[f]];
   }
   fclose(fpsf);
}

double CC[10][4]=
{{3.54356E+02,-8.12003E+02,-8.07068E+02,7.72904E+02},
 {1.00021E+00,1.95172E-02,-2.18757E+00,1.37619E-02},
 {9.79758E-04,-2.18805E+00,-8.51284E-03,2.18899E+00},
 {9.84222E-08,-1.51015E-06,-3.21458E-07,1.33326E-06},
 {-6.31327E-09,5.06693E-06,2.77459E-06,-4.00828E-06},
 {-7.19983E-07,1.04063E-06,-2.06874E-06,-5.71282E-07},
 {-3.73922E-08,-4.55094E-10,7.39223E-08,-2.00520E-09},
 {6.65101E-10,7.50249E-08,-1.18833E-09,-7.86673E-08},
 {-3.51470E-08,-1.51652E-09,7.73056E-08,-3.37491E-09},
 {-2.55003E-09,7.34568E-08,-3.72971E-10,-7.81490E-08}};
double DD[10][4]=
{{3.43646E+02,7.66592E+02,-7.71489E+02,-7.74638E+02},
 {1.00544E-03,2.18607E+00,7.40319E-03,-2.18688E+00},
 {9.99786E-01,2.02006E-02,-2.18554E+00,1.29831E-02},
 {-5.80870E-07,-3.79469E-06,-1.49841E-06,1.30814E-06},
 {-5.07221E-07,-2.99518E-06,2.96650E-06,2.01760E-06},
 {3.41574E-07,1.05789E-07,-2.02756E-07,-8.47656E-07},
 {-1.08091E-09,-7.47857E-08,-1.45720E-09,7.70408E-08},
 {-3.42070E-08,-1.67666E-09,7.72421E-08,-1.68514E-09},
 {1.17819E-09,-7.55302E-08,8.52633E-10,7.75257E-08},
 {-4.48966E-08,3.02579E-10,7.68023E-08,-3.69210E-10}};

void gtransform(int c,double x,double y,double*xc,double*yc) {
   *xc=CC[1][c]*x+CC[2][c]*y+CC[3][c]*x*x+CC[4][c]*x*y+CC[5][c]*y*y+CC[6][c]*x*x*x+CC[7][c]*x*x*y+CC[8][c]*x*y*y+CC[9][c]*y*y*y;
   *yc=DD[1][c]*x+DD[2][c]*y+DD[3][c]*x*x+DD[4][c]*x*y+DD[5][c]*y*y+DD[6][c]*x*x*x+DD[7][c]*x*x*y+DD[8][c]*x*y*y+DD[9][c]*y*y*y;
}

double pixsize(int c,double x,double y) {
   double tx,ty,dx=0,dy=0;
   if (c%2) {
      gtransform(c,x+.5,y+.5,&ty,&tx); dx+=tx; dy+=ty;
      gtransform(c,x-.5,y+.5,&ty,&tx); dx-=tx; dy+=ty;
      gtransform(c,x-.5,y-.5,&ty,&tx); dx-=tx; dy-=ty;
      gtransform(c,x+.5,y-.5,&ty,&tx); dx+=tx; dy-=ty;
      dx=-dx;
   }
   else {
      gtransform(c,x+.5,y+.5,&tx,&ty); dx+=tx; dy+=ty;
      gtransform(c,x-.5,y+.5,&tx,&ty); dx-=tx; dy+=ty;
      gtransform(c,x-.5,y-.5,&tx,&ty); dx-=tx; dy-=ty;
      gtransform(c,x+.5,y-.5,&tx,&ty); dx+=tx; dy-=ty;
   }
   return dx*dy*0.25;
}

double rowsize(int x) {
   //Ron Gilliland, private communication for raw data
   int Nrows=32,rows[32]={64,98,132,166,200,201,234,235,268,269,302,303,337,371,405,439,440,473,474,507,508,541,542,576,610,644,678,712,746,747,780,781};
   double col0[32]={0.9789,0.9878,0.9675,0.9696,0.9829,0.9985,0.9806,0.9796,0.989,0.9803,0.9985,0.9855,0.9683,0.9849,0.9648,0.982,0.9978,0.9813,0.9851,0.9819,0.977,0.9979,0.9834,0.9674,0.9842,0.9667,0.9697,0.9801,0.9807,0.9799,0.9872,0.98};
   double dcol[32]={-3.42557e-06,-3.83428e-05,-3.92468e-06,-5.96548e-06,-3.72032e-05,-1.58276e-05,9.18357e-06,7.45086e-06,-2.89753e-07,-5.09053e-06,-1.21806e-05,-3.36296e-05,-2.00527e-06,-4.03303e-05,1.91426e-06,-2.69071e-05,-1.06506e-05,1.08298e-05,4.22073e-06,9.60101e-06,1.04708e-05,-1.01982e-05,-3.10483e-05,-1.51857e-06,-3.55187e-05,-4.69483e-06,-5.38823e-07,-3.14124e-05,1.22734e-05,3.24158e-06,1.91252e-07,-1.48993e-06};
   double lam0=568.104;
   int i;

   for (i=0;i<Nrows;i++) if (x==rows[i]) {
      double p;
      p=col0[i]+(atoi(FILT+1)-lam0)*dcol[i];
      if (p>1.) return 1.;
      if (p<0.) return 0.;
      return p;
   }
   return 1.;
}

void imgsub(long i) {
   int x1,y1,x0,y0;
   double mx,my;

   if (stars[i].type==4) {
      res[stars[i].x][stars[i].y]-=stars[i].s;
      return;
   }
   if (stars[i].ffx<0) {
      x0=stars[i].fx-1;
      mx=stars[i].ffx+1;
   }
   else {
      x0=stars[i].fx;
      mx=stars[i].ffx;
   }
   if (stars[i].ffy<0) {
      y0=stars[i].fy-1;
      my=stars[i].ffy+1;
   }
   else {
      y0=stars[i].fy;
      my=stars[i].ffy;
   }
   for (x1=stars[i].x-rpsf[fld];x1<=stars[i].x+rpsf[fld];x1++) for (y1=stars[i].y-rpsf[fld];y1<=stars[i].y+rpsf[fld];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) if (data[x1][y1]>DMIN && data[x1][y1]<DMAX) res[x1][y1]-=(psf[fld][stars[i].y/100][stars[i].x/100][x0][y0][x1-stars[i].x][y1-stars[i].y]*(1-mx)*(1-my)+psf[fld][stars[i].y/100][stars[i].x/100][x0+1][y0][x1-stars[i].x][y1-stars[i].y]*mx*(1-my)+psf[fld][stars[i].y/100][stars[i].x/100][x0][y0+1][x1-stars[i].x][y1-stars[i].y]*(1-mx)*my+psf[fld][stars[i].y/100][stars[i].x/100][x0+1][y0+1][x1-stars[i].x][y1-stars[i].y]*mx*my)*stars[i].s;
}

int goodstar(void) {
   int x,y;
   float r;

   if (stars[Nstars].s<Fmin) return 0;
   if (stars[Nstars].x<XMIN+apsize[fld]*1.5+1 || stars[Nstars].x>XMAX-apsize[fld]*1.5-1 || stars[Nstars].y<YMIN+apsize[fld]*1.5+1 || stars[Nstars].y>YMAX-apsize[fld]*1.5-1) return 0;
   for (x=stars[Nstars].x-(int)(apsize[fld])-1;x<=stars[Nstars].x+(int)(apsize[fld])+1;x++) for (y=stars[Nstars].y-(int)(apsize[fld])-1;y<=stars[Nstars].y+(int)(apsize[fld])+1;y++) {
      r=.5+apsize[fld]-sqrt((x-stars[Nstars].x)*(x-stars[Nstars].x)+(y-stars[Nstars].y)*(y-stars[Nstars].y));
      if (r>0 && (data[x][y]<=DMIN || data[x][y]>=DMAX)) return 0;
   }
   for (x=0;x<Nstars;x++) if ((stars[Nstars].x-stars[x].x)*(stars[Nstars].x-stars[x].x)+(stars[Nstars].y-stars[x].y)*(stars[Nstars].y-stars[x].y)<2.00*apsize[fld]*apsize[fld]) return 0;
   return 1;
}

void readstars(char *fn) {
   FILE *f;
   int i,j,tp;
   double x,y,chi,sh,q;
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Cannot find %s\n",fn);
      exit(0);
   }
   fscanf(f,"%d",&i);
   while (i!=fld && !feof(f)) {
      for (j=0;j<10;j++) fscanf(f,"%lf",&x);
      fscanf(f,"%d",&i);
   }
   Nstars=0;
   while (i==fld && !feof(f)) {
      fscanf(f,"%lf %lf %lf",&y,&x,&stars[Nstars].s);
      fscanf(f,"%lf %lf %lf %lf %lf %lf %d",&q,&q,&q,&chi,&q,&sh,&tp);
      stars[Nstars].x=(int)x;
      stars[Nstars].fx=(int)((x-stars[Nstars].x-.5)*sub[fld]+20.5)-20;
      stars[Nstars].ffx=(x-stars[Nstars].x-.5)*sub[fld]-stars[Nstars].fx;
      stars[Nstars].y=(int)y;
      stars[Nstars].fy=(int)((y-stars[Nstars].y-.5)*sub[fld]+20.5)-20;
      stars[Nstars].ffy=(y-stars[Nstars].y-.5)*sub[fld]-stars[Nstars].fy;
      stars[Nstars].type=tp;
      if (Nstars<Napcor && chi<=2.5 && fabs(sh)<=0.5 && tp<4 && goodstar()) Nstars++;
      else imgsub(Nstars);
      fscanf(f,"%d",&i);
   }
   fclose(f);
}

void runapcor(FILE *of) {
   int i,x,y,Nsky;
   double r,mn=1.e30,sd,sky,*slist,NS;

   slist=(double*)calloc((int)(3.14*apsize[fld]*apsize[fld]*2.25+1),sizeof(double));
   for (i=0;i<Nstars;i++) {
      Nsky=0;
      stars[i].aps=stars[i].sig=0;
      for (x=stars[i].x-(int)(apsize[fld]*1.5)-1;x<=stars[i].x+(int)(apsize[fld]*1.5)+1;x++) for (y=stars[i].y-(int)(apsize[fld]*1.5)-1;y<=stars[i].y+(int)(apsize[fld]*1.5)+1;y++) {
	 r=sqrt((x-stars[i].x)*(x-stars[i].x)+(y-stars[i].y)*(y-stars[i].y));
	 if (r>apsize[fld]+.5 && r<=apsize[fld]*1.5) slist[Nsky++]=res[x][y];
      }
      y=1;
      sky=0;
      while (y && Nsky) {
	 sky=sd=0;
	 for (x=0;x<Nsky;x++) sky+=slist[x];
	 sky/=Nsky;
	 for (x=0;x<Nsky;x++) sd+=(slist[x]-sky)*(slist[x]-sky);
	 sd=sqrt(sd/Nsky);
	 y=x=0;
	 while (x<Nsky) if (slist[x]<sky-sd*1.75 || slist[x]>sky+sd*1.75) {
	    Nsky--;
	    slist[x]=slist[Nsky];
	    y=1;
	 }
	 else x++;
      }
      for (x=stars[i].x-(int)(apsize[fld])-1;x<=stars[i].x+(int)(apsize[fld])+1;x++) for (y=stars[i].y-(int)(apsize[fld])-1;y<=stars[i].y+(int)(apsize[fld])+1;y++) {
	 r=.5+apsize[fld]-sqrt((x-stars[i].x)*(x-stars[i].x)+(y-stars[i].y)*(y-stars[i].y));
	 if (r>1) r=1;
	 if (r>0) {
	    stars[i].aps+=(res[x][y]-sky)*r*pixsize(fld,stars[i].y-400,stars[i].x-400)/pixsize(fld,0,0)*rowsize(stars[i].x);
	    stars[i].sig+=noise[x][y]*r;
	 }
      }
      stars[i].sig=sqrt(stars[i].sig);
      stars[i].use=0;
      if (stars[i].sig>0 && stars[i].aps>0 && stars[i].s>0) {
	 fprintf(of,"%d %6.2f %6.2f %8.1f %8.1f %8.3f %8.4f\n",fld,stars[i].y+.5+stars[i].fy/(float)sub[fld],stars[i].x+.5+stars[i].fx/(float)sub[fld],stars[i].s,stars[i].aps,stars[i].sig,sky);
	 if (-2.5*log10(stars[i].aps/stars[i].s)>=MIN && -2.5*log10(stars[i].aps/stars[i].s)<=MAX && stars[i].aps>=7*stars[i].sig) stars[i].use=1;
	 stars[i].sig=stars[i].sig/stars[i].aps*1.0857;
	 stars[i].aps=-2.5*log10(stars[i].aps/stars[i].s);
	 if (fabs(stars[i].aps)>0.5) stars[i].use=0;
      }
   }
   free(slist);
   i=1;
   NS=NSIG0;
   while (i) {
      mn=r=0;
      for (x=0;x<Nstars;x++) if (stars[x].use) {
	 mn+=stars[x].aps/stars[x].sig;
	 r+=1./stars[x].sig;
      }
      mn/=r;
      sd=0;
      y=0;
      r=0;
      for (x=0;x<Nstars;x++) if (stars[x].use) {
	 sd+=(stars[x].aps-mn)*(stars[x].aps-mn)/stars[x].sig/stars[x].sig;
	 y++;
	 r+=1./stars[x].sig/stars[x].sig;
      }
      sd=sqrt(sd/r);
      i=0;
      for (x=0;x<Nstars;x++) {
	 if (stars[x].use && (stars[x].aps-mn>NS*sd || stars[x].aps-mn<-NS*sd)) {
	    stars[x].use=0;
	    i=1;
	 }
      }
      if (NS==NSIG0) NS=NSIG1;
      else NS=NSIG;
   }
   if (y>0) {
      printf(" %d %9.6f +/- %8.6f %3d\n",fld,mn,sd/sqrt(y),y);
      apcs[fld]=mn;
   }
   else {
      printf(" %d %9.6f +/- %8.6f %3d\n",fld,99.999,0.0,y);
      apcs[fld]=99.999;
   }
}

int ferr(FILE *f,char *str) {
   printf("%s\n",str);
   fclose(f);
   return 1;
}

int main(int argc,char**argv) {
   FILE *f,*of;
   char str[401];
   int MODE=0,line=0,tmp,HPP=0;

   if (argc<6 || argc>9) {
      printf("Usage: %s <fits base> <photometry> <Nstars/chip> <min cts> <correction file> <<output>> <<min apcor> <max apcor>>\n",*argv);
      return 1;
   }
   if (argc>7) {
      MIN=atof(argv[argc-2]);
      MAX=atof(argv[argc-1]);
   }
   rpsf[0]=6;
   sub[0]=5;
   Napcor=atoi(argv[3]);
   if (Napcor<1 || Napcor>10000) {
      printf("Illegal # of stars, must be 1-10000\n");
      return 1;
   }
   Fmin=atof(argv[4]);
   getpars();
   if ((of=fopen(argv[5],"w"))==NULL) {
      printf("Cannot write ouput file\n");
      return 1;
   }
   sprintf(str,"%s.fits",argv[1]);
   if ((f=fopen(str,"rb"))==NULL) {
      printf("%s not found\n",str);
      return 1;
   }
   while ((!feof(f)) && ((!MODE) || (line%36))) {
      fgets(str,81,f);
      line++;
      if (!MODE) {
	 if (!strncmp(str,"END     ",8)) MODE=1;
	 else if ((!strncmp(str,"SIMPLE  =",9)) && (str[29]!='T')) return ferr(f,"Illegal format type");
	 else if ((!strncmp(str,"BITPIX  =",9)) && (strncmp(str+27,"-32",3))) return ferr(f,"Illegal bits per pixel");
	 else if ((!strncmp(str,"NAXIS   =",9)) && (strncmp(str+27,"  3",3))) return ferr(f,"Illegal number of axes");
	 else if (!strncmp(str,"NAXIS",5)) {
	    tmp=atoi(str+9);
	    if (str[5]=='1') X=tmp;
	    else if (str[5]=='2') Y=tmp;
	    else if (str[5]=='3') Z=tmp;
	 }
	 else if (!strncmp(str,"ATODGAIN=",9)) GAIN=atof(str+9);
	 else if (!strncmp(str,"RNOISE  =",9)) RN=atof(str+9);
	 else if (!strncmp(str,"RSDPFILL=",9)) DMIN=atoi(str+9);
	 else if (!strncmp(str,"SATURATE=",9)) DMAX=atoi(str+9);
	 else if (!strncmp(str,"FILTNAM1=",9)) {
	    int i=0;
	    while ((i<19) && (str[i+11]!=32)) {
	       FILT[i]=str[i+11];
	       i++;
	    }
	 }
	 else if (!strncmp(str,"HSTPHOT =",9)) HPP=1;
      }
   }
   if (!HPP) return ferr(f,"Not pre-processed by mask");
   if ((X!=800) || (Y!=800) || (Z!=4)) return ferr(f,"Illegal axis size");
   printf("Reading FITS file:\n");
   printf("Saturation at %d, bad value=%d\n",DMAX,DMIN);
   printf("Filter=\"%s\"\n",FILT);
   initfilters();
   filt=findfilt(FILT);
   if (GAIN!=7 && GAIN!=14) {
      printf("Illegal Gain\n");
      exit(1);
   }
   getpsf();
   psfcorr(argv[2]);
   for (fld=0;fld<4;fld++) {
      ffread(data,sizeof(dtype),1,f);
      if (fld>=MMIN && fld<MMAX) {
	 memcpy(res,data,sizeof(dtype));
	 sprintf(str,"%s.sky",argv[1]);
	 readsky(str);
	 readstars(argv[2]);
	 runapcor(of);
      }
   }
   fclose(of);
   fclose(f);
   if (argc%2) {
      sprintf(str,"%s/makeapcor %s %s %f %f %f %f",BASEDIR,argv[2],argv[6],apcs[0],apcs[1],apcs[2],apcs[3]);
      if ((f=popen(str,"w"))!=NULL) pclose(f);
   }
   return 0;
}
