#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif

typedef float dtype[4][800][800];

int X,Y,Z,DMAX,DMIN;
dtype data,sky,sky2;
int rsky[4],rin[4],rpsf[4];

void merr(void) {
   printf("Mem error\n");
   exit(0);
}

void getsky(char *fn) {
   double *list,s,sig;
   int i,n,x1,y1,*mask,f,x,y;
   FILE *sf;

   list=(double*)calloc((2*rsky[0]+1)*(2*rsky[0]+1),sizeof(double));
   mask=(int*)calloc((2*rsky[0]+1)*(2*rsky[0]+1),sizeof(int));
   if (!list || !mask) merr();
   for (f=0;f<4;f++) for (x=0;x<800;x++) for (y=0;y<800;y++) {
      if (y==0 && x*5%800<5) {
	 printf(".");
	 fflush(stdout);
      }
      n=0;
      for (x1=x-rsky[f];x1<=x+rsky[f];x1+=4) for (y1=y-rsky[f];y1<=y+rsky[f];y1+=4) if (x1>=0 && x1<800 && y1>=0 && y1<800 && abs(x-x1)>rin[f] && abs(y-y1)>rin[f] && data[f][x1][y1]>DMIN && data[f][x1][y1]<DMAX*0.855) {
	 mask[n]=1;
	 list[n++]=data[f][x1][y1];
      }
      x1=1;
      while (x1) {
	 s=0;
	 y1=0;
	 for (i=0;i<n;i++) if (mask[i]) {
	    s+=list[i];
	    y1++;
	 }
	 if (y1==0) s=x1=0;
	 else {
	    s/=y1;
	    sig=0;
	    for (i=0;i<n;i++) if (mask[i]) sig+=(list[i]-s)*(list[i]-s);
	    if (y1>1) sig=sqrt(sig/(y1-1));
	    else sig=sqrt(sig);
	    x1=0;
	    for (i=0;i<n;i++) if (mask[i] && (list[i]<s-2.25*sig || list[i]>s+1.75*sig)) {
	       mask[i]=0;
	       x1=1;
	    }
	 }
      }
      sky[f][x][y]=s;
   }
   free(list);
   free(mask);
   for (f=0;f<4;f++) for (x=0;x<800;x++) for (y=0;y<800;y++) {
      if (y==0 && x*5%800<5) {
	 printf(".");
	 fflush(stdout);
      }
      sky2[f][x][y]=0;
      n=0;
      for (x1=x-rpsf[f];x1<=x+rpsf[f];x1++) for (y1=y-rpsf[f];y1<=y+rpsf[f];y1++) if (x1>=0 && x1<800 && y1>=0 && y1<800) {
	 n++;
	 sky2[f][x][y]+=sky[f][x1][y1];
      }
      if (n) sky2[f][x][y]/=n;
      else sky2[f][x][y]=sky[f][x][y];
   }
   if ((sf=fopen(fn,"wb"))!=NULL) ffwrite(sky2,4*800*800,4,sf);
   else {
      printf("Cannot open file %s\n",fn);
      exit(0);
   }
   fclose(sf);
}

void ferr(FILE *f,char *str) {
   printf("%s\n",str);
   fclose(f);
   exit(1);
}

int main(int argc,char**argv) {
   FILE *f;
   char str[82];
   int MODE=0,line=0,tmp,HPP=0;

   if (argc!=2) {
      printf("Usage: %s <fits base>\n",*argv);
      return 1;
   }
   sprintf(str,"%s.fits",argv[1]);
   if ((f=fopen(str,"rb"))==NULL) {
      printf("%s not found\n",str);
      return 1;
   }
   while ((!feof(f)) && ((!MODE) || (line%36))) {
      fgets(str,81,f);
      line++;
      if (!MODE) {
	 if (!strncmp(str,"END     ",8)) MODE=1;
	 else if ((!strncmp(str,"SIMPLE  =",9)) && (str[29]!='T')) ferr(f,"Illegal format type");
	 else if ((!strncmp(str,"BITPIX  =",9)) && (strncmp(str+27,"-32",3))) ferr(f,"Illegal bits per pixel");
	 else if ((!strncmp(str,"NAXIS   =",9)) && (strncmp(str+27,"  3",3))) ferr(f,"Illegal number of axes");
	 else if (!strncmp(str,"NAXIS",5)) {
	    tmp=atoi(str+9);
	    if (str[5]=='1') X=tmp;
	    else if (str[5]=='2') Y=tmp;
	    else if (str[5]=='3') Z=tmp;
	 }
	 else if (!strncmp(str,"RSDPFILL=",9)) DMIN=atoi(str+9);
	 else if (!strncmp(str,"SATURATE=",9)) DMAX=atoi(str+9);
	 else if (!strncmp(str,"HSTPHOT =",9)) HPP=1;
      }
   }
   if (!HPP) ferr(f,"Not pre-processed by mask");
   if ((X!=800) || (Y!=800) || (Z!=4)) ferr(f,"Illegal axis size");
   ffread(data,4*800*800,4,f);
   fclose(f);
   rsky[0]=36;
   rsky[1]=rsky[2]=rsky[3]=20;
   rin[0]=15;
   rin[1]=rin[2]=rin[3]=8;
   rpsf[0]=6;
   rpsf[1]=rpsf[2]=rpsf[3]=4;
   sprintf(str,"%s.sky",argv[1]);
   getsky(str);
   printf("\n");
   return 0;
}
