#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"
#define XMIN 50
#define XMAX 800
#define YMIN 50
#define YMAX 800
#define MMIN 0
#define MMAX 4

typedef float dtype[800][800];
typedef struct{int f,x,y,fx,fy;double s,ss,chi,sh;} stype;
typedef float ****psftype;

stype stars[2][640001],pstar;
int X,Y,Z,DMAX[2],DMIN[2],Nimg,Nmin,SOLTYPE=1,cmdsz[2],TEMP[2];
float ran[800][800],cran[800][800],rap[4],rmin[4],****cwt[4],limits[2][2];
double EXP[2],EXP0[2],YEAR[2],GAIN[2],RN[2],corr[2][4]={{1.0066,1.005,1,1.0013},{1.0066,1.005,1,1.0013}},fsig,maxchi,Shmin,Shmax;
dtype data[2],sky[2],*res,res1[2],res2[2],noise1[2],noise2[2],*noise;
psftype psf[2][4][8][8];
char FILT[2][20]={{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}};
int rsky[4],rphot[4],rpsf[4],sub[4],***circ[2][4],Nstars[2]={0,0},fld,Gauss_PSF=0,filt[2];
float dnoise[4]={0.0361,0.0625,0.0625,0.0625};

void merr(void) {
   printf("Mem error\n");
   exit(0);
}

void getpars(void) {
   int mx,my,x,y,f,rad;
   for (x=0;x<800;x++) for (y=0;y<800;y++) cran[x][y]=-10000;
   rpsf[1]=rpsf[2]=rpsf[3]=(rpsf[0]*2+1)/3;
   sub[1]=sub[2]=sub[3]=sub[0]*2;
   rap[1]=rap[2]=rap[3]=rap[0]*2./3.;
   for (f=0;f<4;f++) rphot[f]=(int)(rap[f]+.9);
   rmin[1]=rmin[2]=rmin[3]=rmin[0]*2./3.;
   for (f=0;f<4;f++) {
      rad=(sub[f]+2)/2;
      circ[0][f]=(int***)calloc(rad*2+1,sizeof(int**));
      circ[1][f]=(int***)calloc(rad*2+1,sizeof(int**));
      cwt[f]=(float****)calloc(rad*2+1,sizeof(float***));
      if (!circ[0][f] || !circ[1][f] || !cwt[f]) merr();
      circ[0][f]+=rad;
      circ[1][f]+=rad;
      cwt[f]+=rad;
      for (mx=-rad;mx<=rad;mx++) {
	 circ[0][f][mx]=(int**)calloc(rad*2+1,sizeof(int*));
	 circ[1][f][mx]=(int**)calloc(rad*2+1,sizeof(int*));
	 cwt[f][mx]=(float***)calloc(rad*2+1,sizeof(float**));
	 if (!circ[0][f][mx] || !circ[1][f][mx] || !cwt[f][mx]) merr();
	 circ[0][f][mx]+=rad;
	 circ[1][f][mx]+=rad;
	 cwt[f][mx]+=rad;
	 for (my=-rad;my<=rad;my++) {
	    circ[0][f][mx][my]=(int*)calloc(rphot[f]*2+1,sizeof(int));
	    circ[1][f][mx][my]=(int*)calloc(rphot[f]*2+1,sizeof(int));
	    cwt[f][mx][my]=(float**)calloc(rphot[f]*2+1,sizeof(float*));
	    if (!circ[0][f][mx][my] || !circ[1][f][mx][my] || !cwt[f][mx][my]) merr();
	    circ[0][f][mx][my]+=rphot[f];
	    circ[1][f][mx][my]+=rphot[f];
	    cwt[f][mx][my]+=rphot[f];
	    for (x=-rphot[f];x<=rphot[f];x++) {
	       circ[0][f][mx][my][x]=50;
	       circ[1][f][mx][my][x]=-50;
	       cwt[f][mx][my][x]=(float*)calloc(rphot[f]*2+1,sizeof(float));
	       if (!cwt[f][mx][my][x]) merr();
	       cwt[f][mx][my][x]+=rphot[f];
	       for (y=-rphot[f];y<=rphot[f];y++) {
		  cwt[f][mx][my][x][y]=rap[f]+.5-sqrt((x-mx/(float)sub[f])*(x-mx/(float)sub[f])+(y-my/(float)sub[f])*(y-my/(float)sub[f]));
		  if (cwt[f][mx][my][x][y]<0) cwt[f][mx][my][x][y]=0;
		  else if (cwt[f][mx][my][x][y]>1) cwt[f][mx][my][x][y]=1;
		  if (cwt[f][mx][my][x][y]>0 && y<circ[0][f][mx][my][x]) circ[0][f][mx][my][x]=y;
		  if (cwt[f][mx][my][x][y]>0 && y>circ[1][f][mx][my][x]) circ[1][f][mx][my][x]=y;
	       }
	    }
	 }
      }
   }
}

void readsky(int flt,char *fn) {
   int f,x,y;
   FILE *sf;

   if ((sf=fopen(fn,"rb"))==NULL) {
      printf("Cannot open file %s\n",fn);
      exit(0);
   }
   for (f=0;f<=fld;f++) ffread(sky[flt],800*800,4,sf);
   for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) {
      res1[flt][x][y]-=sky[flt][x][y];
      noise1[flt][x][y]=RN[flt]*RN[flt];
      if (res[flt][x][y]>0) noise1[flt][x][y]+=res1[flt][x][y]/GAIN[flt];
      if (sky[flt][x][y]>0) noise1[flt][x][y]+=sky[flt][x][y]/GAIN[flt];
   }
   memcpy(res2+flt,res1+flt,800*800*4);
   memcpy(noise2+flt,noise1+flt,800*800*4);
   fclose(sf);
}

void getpsf(void) {
   int flt,f,x,y,z,rad,px,py;

   for (flt=0;flt<2;flt++) for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
      rad=(sub[f]+2)/2;
      psf[flt][f][px][py]=(float****)calloc(rad*2+1,sizeof(float***));
      if (!psf[flt][f][px][py]) merr();
      psf[flt][f][px][py]+=rad;
      for (x=-rad;x<=rad;x++) {
	 psf[flt][f][px][py][x]=(float***)calloc(rad*2+1,sizeof(float**));
	 if (!psf[flt][f][px][py][x]) merr();
	 psf[flt][f][px][py][x]+=rad;
	 for (y=-rad;y<=rad;y++) {
	    psf[flt][f][px][py][x][y]=(float**)calloc(2*rpsf[f]+1,sizeof(float*));
	    if (!psf[flt][f][px][py][x][y]) merr();
	    psf[flt][f][px][py][x][y]+=rpsf[f];
	    for (z=-rpsf[f];z<=rpsf[f];z++) {
	       psf[flt][f][px][py][x][y][z]=(float*)calloc(2*rpsf[f]+1,sizeof(float));
	       if (!psf[flt][f][px][py][x][y][z]) merr();
	       psf[flt][f][px][py][x][y][z]+=rpsf[f];
	    }
	 }
      }
   }
   if (Gauss_PSF) for (flt=0;flt<2;flt++) {
      int x1,y1;
      float sig,n;
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 rad=(sub[f]+2)/2;
	 if (f) sig=0.54;
	 else sig=0.84;
	 for (x=-rad;x<=rad;x++) for (y=-rad;y<=rad;y++) {
	    n=0;
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) n+=(psf[flt][f][px][py][x][y][x1][y1]=exp(-.5*((x1-x/(float)sub[f])*(x1-x/(float)sub[f])+(y1-y/(float)sub[f])*(y1-y/(float)sub[f]))/sig/sig));
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) psf[flt][f][px][py][x][y][x1][y1]/=n;
	 }
      }
   }
   else for (flt=0;flt<2;flt++) {
      FILE *pf;
      int x1;
      char fn[80];

      sprintf(fn,"%s/psfs/%s.%d.%d.psf",BASEDIR,filters[filt[flt]].suffix,rpsf[0],sub[0]);
      if ((pf=fopen(fn,"rb"))==NULL) {
	 printf("File error: %s\n",fn);
	 exit(0);
      }
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) for (x=-(sub[f]+2)/2;x<=(sub[f]+2)/2;x++) for (y=-(sub[f]+2)/2;y<=(sub[f]+2)/2;y++) for (x1=-rpsf[f];x1<=rpsf[f];x1++) ffread(psf[flt][f][px][py][x][y][x1]-rpsf[f],2*rpsf[f]+1,sizeof(float),pf);
      fclose(pf);
   }
}

void psfcorr(int flt,char *str) {
   char fn[81];
   FILE *fpsf;
   int f,x,y,x2,y2,x3,y3,rad;
   float c[101];

   sprintf(fn,"%s.psf",str);
   if ((fpsf=fopen(fn,"rb"))==NULL) return;
   for (f=0;f<4;f++) for (x=-rpsf[f];x<=rpsf[f];x++) {
      rad=(sub[f]+2)/2;
      ffread(c,sizeof(float),2*rpsf[f]+1,fpsf);
      for (y=-rpsf[f];y<=rpsf[f];y++) for (x2=0;x2<8;x2++) for (y2=0;y2<8;y2++) for (x3=-rad;x3<=rad;x3++) for (y3=-rad;y3<=rad;y3++) psf[flt][f][x2][y2][x3][y3][x][y]+=c[y+rpsf[f]];
   }
   fclose(fpsf);
}

void photsearch(int flt,int x,int y,int fx,int fy,int*first,int*sx,int*sy,int*bx,int*by,double*bs,double*bss,double*bm,double*bc,double*bsh) {
   int xbx=0,xby=0,rad,x1,y1,x2,y2,xfirst=1,set=0,DONE=0;
   double xbm=0,sh,m,cs,cc,s,ss,a,b,c,A,np;
   static double **result=NULL;
   static int lastf=-1,lastrad=0;

   if (x<=XMIN || x>=XMAX-1 || y<=YMIN || y>=YMAX-1) return;
   if (ran[x][y]>-10000) return;
   rad=(sub[fld]+2)/2;
   if (fld!=lastf) {
      int i;
      if (lastf>=0) {
	 for (i=-lastrad;i<=lastrad;i++) free(result[i]-lastrad);
	 free(result-lastrad);
      }
      result=(double**)calloc(sizeof(double*),rad*2+1);
      if (!result) merr();
      result+=rad;
      for (i=-rad;i<=rad;i++) {
	 result[i]=(double*)calloc(sizeof(double),rad*2+1);
	 if (!result[i]) merr();
	 result[i]+=rad;
      }
      lastf=fld;
      lastrad=rad;
   }
   for (x2=-rad;x2<=rad;x2++) for (y2=-rad;y2<=rad;y2++) result[x2][y2]=-1;
   while (!DONE) {
      for (x2=fx-1;x2<=fx+1;x2++) for (y2=fy-1;y2<=fy+1;y2++) if (result[x2][y2]<0) {
	 cs=s=ss=c=cc=0;
	 for (x1=x-rphot[fld];x1<=x+rphot[fld];x1++) for (y1=y+circ[0][fld][x2][y2][x1-x];y1<=y+circ[1][fld][x2][y2][x1-x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) if (data[flt][x1][y1]>DMIN[flt] && data[flt][x1][y1]<DMAX[flt]) {
	    if (SOLTYPE==1) {
	       a=psf[flt][fld][y/100][x/100][x2][y2][x1-x][y1-y];
	       b=a*cwt[fld][x2][y2][x1-x][y1-y]/noise[flt][x1][y1];
	       cs+=b;
	       if (a>0) {
		  s+=res[flt][x1][y1]*b;
		  a*=b;
		  c+=a;
		  cc+=a;
		  ss+=a*cwt[fld][x2][y2][x1-x][y1-y];
	       }
	       else cc+=a*b;
	    }
	    else {
	       cs+=cwt[fld][x2][y2][x1-x][y1-y];
	       c+=psf[flt][fld][y/100][x/100][x2][y2][x1-x][y1-y]*cwt[fld][x2][y2][x1-x][y1-y];
	       s+=res[flt][x1][y1]*cwt[fld][x2][y2][x1-x][y1-y];
	       ss+=noise[flt][x1][y1]*cwt[fld][x2][y2][x1-x][y1-y];
	    }
	 }
	 if (ss>0 && c>0) {
	    cs=cc/cs;
	    s/=c;
	    ss=sqrt(ss)/c;
	    c=sh=0;
	    A=np=0;
	    for (x1=x-rphot[fld];x1<=x+rphot[fld];x1++) for (y1=y+circ[0][fld][x2][y2][x1-x];y1<=y+circ[1][fld][x2][y2][x1-x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) if (data[flt][x1][y1]>DMIN[flt] && data[flt][x1][y1]<DMAX[flt]) {
	       a=psf[flt][fld][y/100][x/100][x2][y2][x1-x][y1-y]*s;
	       b=1/(noise[flt][x1][y1]+dnoise[fld]*a*a);
	       a=res[flt][x1][y1]-a;
	       c+=a*a*b*cwt[fld][x2][y2][x1-x][y1-y];
	       np+=cwt[fld][x2][y2][x1-x][y1-y];
	       sh+=a/sqrt(noise[flt][x1][y1])*(psf[flt][fld][y/100][x/100][x2][y2][x1-x][y1-y]-cs)*cwt[fld][x2][y2][x1-x][y1-y];
	       A+=(psf[flt][fld][y/100][x/100][x2][y2][x1-x][y1-y]-cs)*(psf[flt][fld][y/100][x/100][x2][y2][x1-x][y1-y]-cs)*s/sqrt(noise[flt][x1][y1])*cwt[fld][x2][y2][x1-x][y1-y];
	    }
	    if (A>0) sh/=A;
	    else sh=0;
	    c=c/np;
	    m=s/(ss*sqrt(c+.25));
	    if (c>1) ss*=sqrt(c);
	    if (m>0) result[x2][y2]=m;
	    else result[x2][y2]=0;
	    if (m>xbm || xfirst) {
	       xfirst=0;
	       xbm=m;
	       xbx=x2;
	       xby=y2;
	    }
	    if (abs(x2)!=rad && abs(y2)!=rad && (m>*bm || *first) && s>0 && s>=ss*fsig) {
	       set=1;
	       *first=0;
	       *bm=m;
	       *bs=s;
	       *bss=ss;
	       *bc=c;
	       *bsh=sh;
	       *bx=x2;
	       *by=y2;
	       *sx=x;
	       *sy=y;
	    }
	 }
	 else result[x2][y2]=0;
      }
      DONE=1;
      if (set && (result[*bx][*by+1]<0 || result[*bx+1][*by+1]<0 || result[*bx+1][*by]<0 || result[*bx+1][*by-1]<0 || result[*bx][*by-1]<0 || result[*bx-1][*by-1]<0 || result[*bx-1][*by]<0 || result[*bx-1][*by+1]<0)) DONE=0;
      fx=*bx;
      fy=*by;
   }
   if (xfirst) {
      ran[x][y]=-99;
      return;
   }
   else if (*first) ran[x][y]=-9999;
   else ran[x][y]=*bm;
   if (xbx==-rad && xby==-rad) photsearch(flt,x-1,y-1,rad-1,rad-1,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xbx==-rad && xby==rad) photsearch(flt,x-1,y+1,rad-1,1-rad,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xbx==rad && xby==rad) photsearch(flt,x+1,y+1,1-rad,1-rad,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xbx==rad && xby==-rad) photsearch(flt,x+1,y-1,1-rad,rad-1,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xbx==-rad) photsearch(flt,x-1,y,rad-1,xby,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xbx==rad) photsearch(flt,x+1,y,1-rad,xby,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xby==-rad) photsearch(flt,x,y-1,xbx,rad-1,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   else if (xby==rad) photsearch(flt,x,y+1,xbx,1-rad,first,sx,sy,bx,by,bs,bss,bm,bc,bsh);
   if (*first && *bm<-9998) ran[x][y]=-9999;
   else ran[x][y]=*bm;
}

void imgsub(int it,int i) {
   int x1,y1;
   float q;
   for (x1=stars[it][i].x-rpsf[fld];x1<=stars[it][i].x+rpsf[fld];x1++) for (y1=stars[it][i].y-rpsf[fld];y1<=stars[it][i].y+rpsf[fld];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) if (data[it][x1][y1]>DMIN[it] && data[it][x1][y1]<DMAX[it]) {
      res2[it][x1][y1]-=(q=stars[it][i].s*psf[it][fld][stars[it][i].y/100][stars[it][i].x/100][stars[it][i].fx][stars[it][i].fy][x1-stars[it][i].x][y1-stars[it][i].y]);
      noise2[it][x1][y1]+=dnoise[fld]*q*q;
   }
}

inline double frand(void) {
   return (rand()/10%100)*.01+(rand()/10%100)*.0001+(rand()/10%100+.5)*.000001;
}

inline int randomize(int flt,float r) {
   int i=0,j,g;

   //r+=(frand()+frand()-frand()-frand())/0.577562*r*sqrt(dnoise[fld]);
   if (r*GAIN[flt]>400) return (int)(r+(frand()+frand()-frand()-frand())/0.577562*sqrt(r/GAIN[flt])+0.5);
   g=(int)(GAIN[flt]+.5);
   for (j=((int)r)*g*2;j>0;j--) if ((rand()/256)%2) i++;
   i=(i+g/2)/g;
   r-=(int)r;
   if (r>=.1 && (rand()%200)*.005<r) i++;
   else if (r<.1 && r>=.0001 && rand()%(int)(1/r+.5)==0) i++;
   return i;
}

int addstar(int flt,int x,int y,int fx,int fy,float s) {
   int x1,y1;
   double z;
   if (data[flt][x][y]<=DMIN[flt] || data[flt][x][y]>=DMAX[flt]) return 0;
   s*=1+(frand()+frand()-frand()-frand())/0.577562*0.015;
   if (s<0) s=0;
   for (x1=x-rpsf[fld];x1<=x+rpsf[fld];x1++) for (y1=y-rpsf[fld];y1<=y+rpsf[fld];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) if (data[flt][x1][y1]>DMIN[flt] && data[flt][x1][y1]<DMAX[flt]) {
      res1[flt][x1][y1]+=(z=randomize(flt,s*psf[flt][fld][y/100][x/100][fx][fy][x1-x][y1-y]));
      res2[flt][x1][y1]+=z;
      data[flt][x1][y1]+=z;
      if (data[flt][x][y]>=DMAX[flt]*0.855 && fabs(x1-x)+fabs(y1-y)<2) return 0;
   }
   return 1;
}

int ferr(FILE *f,char *str) {
   printf("%s\n",str);
   fclose(f);
   return 1;
}

inline float getflux(float m,int x,int y,int flt) {
   double bgcorr[NFILTERS][2][2]={
      {{0.000000e-06,2.869069e-04},{0.000000e-06,2.292851e-04}}, //F170W
      {{8.540038e-07,1.010904e-04},{2.347928e-06,8.078760e-05}}, //F300W
      {{8.540038e-07,1.171397e-04},{2.347928e-06,1.000053e-04}}, //F336W
      {{8.540038e-07,9.556493e-05},{2.347928e-06,1.000053e-04}}, //F380W
      {{1.663013e-07,8.647951e-05},{0.000000e-06,5.432855e-05}}, //F410M
      {{1.663013e-07,8.770231e-05},{0.000000e-06,5.432855e-05}}, //F439W
      {{1.663013e-07,7.891117e-05},{0.000000e-06,5.432855e-05}}, //F450W
      {{1.663013e-07,6.920918e-05},{0.000000e-06,5.432855e-05}}, //F467M
      {{5.764012e-06,7.091177e-05},{5.497646e-06,8.393854e-05}}, //F547M
      {{5.764012e-06,8.348157e-05},{5.497646e-06,7.363844e-05}}, //F555W
      {{2.288654e-05,8.749157e-05},{5.497646e-06,7.331492e-05}}, //F569W
      {{2.288654e-05,8.625639e-05},{5.497646e-06,6.200388e-05}}, //F606W
      {{3.245634e-05,9.317634e-05},{1.270904e-06,1.109481e-04}}, //F622W
      {{3.245634e-05,9.317634e-05},{1.270904e-06,1.109481e-04}}, //F675W
      {{3.245634e-05,8.589005e-05},{1.270904e-06,7.000852e-05}}, //F702W
      {{2.466686e-05,8.710683e-05},{1.781239e-05,6.953957e-05}}, //F785LP
      {{2.466686e-05,8.533950e-05},{1.781239e-05,9.479239e-05}}, //F791W
      {{2.466686e-05,8.533950e-05},{1.781239e-05,9.479239e-05}}, //F814W
      {{2.466686e-05,8.533950e-05},{1.781239e-05,9.479239e-05}}, //F850LP
      {{2.466686e-05,8.533950e-05},{1.781239e-05,9.479239e-05}}  //F1042M
   };
   double gcorr[4]={3.719247e-02,-1.203985e-02,6.861146e-03,3.992702e-03};
   double fcorr[2][NFILTERS]={
      {0.000000e+00,0.000000e+00,-2.304644e-02,-9.948792e-02,-7.099982e-02,-1.567542e-02,-9.765530e-03,4.599253e-02,-3.731368e-03,-9.215336e-03,7.585554e-03,9.387174e-03,-1.170134e-03,-6.744189e-03,-1.713857e-02,4.278063e-03,3.849086e-02,1.230014e-02,1.362391e-02,1.583694e-02},
      {0.000000e+00,0.000000e+00,4.467948e-02,-5.021531e-02,-2.172721e-02,3.359719e-02,3.950708e-02,9.526514e-02,1.863842e-02,1.315445e-02,2.995534e-02,3.175696e-02,1.809453e-02,1.252047e-02,2.126090e-03,-2.218946e-02,1.202334e-02,-1.416738e-02,-1.284361e-02,-1.063058e-02}};
   float m0,mult,XCTE,YCTE,ct,bg,lct,lbg,yr;

   mult=EXP0[flt]/EXP[flt];
   if (GAIN[flt]!=7) mult*=14;
   else mult*=7;
   ct=pow(10,-(m-filters[filt[flt]].zero)*0.4)*EXP[flt]/corr[flt][fld]*mult;
   bg=(sky[flt][x][y]*mult)-ct*bgcorr[filt[flt]][TEMP[flt]][(fld==0)?0:1];
   //Soften background and signal;
   if (bg<0) bg=1.;
   else bg=sqrt(1.+bg*bg);
   if (ct<0) ct=1.;
   else ct=sqrt(1.+ct*ct);
   /*
   //CTE from Whitmore, Heyer, & Casertano 1999;
   YCTE=0.023*pow(10,-0.256*log10(bg/7.));
   XCTE=0.025;
   if (s0*mult<28000) {
      YCTE*=(1+0.245*(-0.0087*log10(ct/28000.))*(YEAR[flt]-0.04)*365.25);
      XCTE*=(1+0.341*(-0.0020*log10(ct/28000.))*(YEAR[flt]-0.04)*365.25);
   }
   m0=m+2.5*log10(1.+YCTE*x/80000.+XCTE*y/80000.);
   */
   //HSTphot CTE determination;
   lct=log(ct)-7;
   lbg=log(bg)-1;
   yr=YEAR[flt]-2;
   if (TEMP[flt]==0) YCTE=0.018009+(0.087972+exp(-0.506635*lct))*exp(-0.042368*bg-0.035149*lbg)*(0.096532+0.041442*yr);
   else YCTE=0.102822+0.027988*exp(-0.958904*lct);
   if (TEMP[flt]==0) XCTE=exp(-0.195966*lct)*exp(-0.126435*lbg)*(0.023881+0.002327*yr);
   else XCTE=0.;
   m0=m+fcorr[TEMP[flt]][filt[flt]]+gcorr[fld]+YCTE*x/800.+XCTE*y/800.;
   return pow(10,-(m0-filters[filt[flt]].zero)*0.4)*EXP[flt]/corr[flt][fld];
}

inline float rate(int flt,stype*s) {
   float dx,dy;
   if (s->f!=fld) return -1;
   dx=s->x-pstar.x+(s->fx-pstar.fx)/(float)sub[fld];
   if (fld && fabs(dx)>1.5) return -1;
   if (!fld && fabs(dx)>2.25) return -1;
   dy=s->y-pstar.y+(s->fy-pstar.fx)/(float)sub[fld];
   if (fld && fabs(dy)>1.5) return -1;
   if (!fld && fabs(dy)>2.25) return -1;
   if (dx>0) dx+=.5;
   else dx=-dx+.5;
   if (dx>rpsf[fld]) dx=rpsf[fld];
   if (dy>0) dy+=.5;
   else dy=-dy+.5;
   if (dy>rpsf[fld]) dy=rpsf[fld];
   return s->s*psf[flt][fld][s->x/100][s->y/100][0][0][(int)dx][(int)dy];
}

float runstar(int flt,stype*instar) {
   int Nrec=0,it,i,j,pfirst,fx=0,fy=0;
   float diffmag=0;
   double r,bm;
   stype p2star;
   static float **buff1,**buff2,**dbuff;
   static int first=1;

   if (first) {
      first=0;
      buff1=(float**)calloc(rpsf[0]*2+1,sizeof(float*));
      buff2=(float**)calloc(rpsf[0]*2+1,sizeof(float*));
      dbuff=(float**)calloc(rpsf[0]*2+1,sizeof(float*));
      if (!buff1 || !buff2 || !dbuff) merr();
      buff1+=rpsf[0];
      buff2+=rpsf[0];
      dbuff+=rpsf[0];
      for (i=-rpsf[0];i<=rpsf[0];i++) {
	 buff1[i]=(float*)calloc(rpsf[0]*2+1,sizeof(float));
	 buff2[i]=(float*)calloc(rpsf[0]*2+1,sizeof(float));
	 dbuff[i]=(float*)calloc(rpsf[0]*2+1,sizeof(float));
	 if (!buff1[i] || !buff2[i] || !dbuff[i]) merr();
	 buff1[i]+=rpsf[0];
	 buff2[i]+=rpsf[0];
	 dbuff[i]+=rpsf[0];
      }
   }
   for (it=0;it<Nimg;it++) {
      if (it%2) fx=instar->fx;
      else if (instar->fx<=0) fx=instar->fx+sub[fld]/2;
      else fx=instar->fx-(sub[fld]+1)/2;
      if (it/2%2) fy=instar->fy;
      else if (instar->fy<=0) fy=instar->fy+sub[fld]/2;
      else fy=instar->fy-(sub[fld]+1)/2;
      j=sizeof(float)*(2*rpsf[fld]+1);
      for (i=-rpsf[fld];i<=rpsf[fld];i++) memcpy(buff1[i]-rpsf[fld],res1[flt][instar->x+i]+instar->y-rpsf[fld],j);
      for (i=-rpsf[fld];i<=rpsf[fld];i++) memcpy(buff2[i]-rpsf[fld],res2[flt][instar->x+i]+instar->y-rpsf[fld],j);
      for (i=-rpsf[fld];i<=rpsf[fld];i++) memcpy(dbuff[i]-rpsf[fld],data[flt][instar->x+i]+instar->y-rpsf[fld],j);
      if (addstar(flt,instar->x,instar->y,fx,fy,instar->s)) {
	 pstar.s=0;
	 pfirst=1;
	 bm=-100;
	 memcpy(ran,cran,sizeof(float[800][800]));
	 res=res2;
	 noise=noise2;
	 photsearch(flt,instar->x,instar->y,fx,fy,&pfirst,&pstar.x,&pstar.y,&pstar.fx,&pstar.fy,&pstar.s,&pstar.ss,&bm,&pstar.chi,&p2star.sh);
	 p2star.s=0;
	 pfirst=1;
	 bm=-100;
	 memcpy(ran,cran,sizeof(float[800][800]));
	 res=res1;
	 noise=noise1;
	 photsearch(flt,instar->x,instar->y,fx,fy,&pfirst,&p2star.x,&p2star.y,&p2star.fx,&p2star.fy,&p2star.s,&p2star.ss,&p2star.chi,&bm,&p2star.sh);
	 if (pstar.s>0 && (pstar.s>fsig*pstar.ss || p2star.s>fsig*p2star.ss) && pstar.chi<=maxchi && pstar.sh>=Shmin && pstar.sh<=Shmax) {
	    r=rate(flt,instar);
	    if (r>-1) {
	       for (i=0;i<Nstars[flt] && r>rate(flt,stars[flt]+i);i++);
	       if (i<Nstars[flt]) pstar.s=0;
	    }
	    else pstar.s=0;
	 }
	 else pstar.s=0;
      }
      else pstar.s=0;
      for (i=-rpsf[fld];i<=rpsf[fld];i++) memcpy(res1[flt][instar->x+i]+instar->y-rpsf[fld],buff1[i]-rpsf[fld],j);
      for (i=-rpsf[fld];i<=rpsf[fld];i++) memcpy(res2[flt][instar->x+i]+instar->y-rpsf[fld],buff2[i]-rpsf[fld],j);
      for (i=-rpsf[fld];i<=rpsf[fld];i++) memcpy(data[flt][instar->x+i]+instar->y-rpsf[fld],dbuff[i]-rpsf[fld],j);
      if (pstar.s>0) {
	 diffmag-=2.5*log10(pstar.s/instar->s);
	 Nrec++;
      }
   }
   if (Nrec>=Nmin) return diffmag/Nrec;
   else return 99;
}

void onestar(FILE *of,int x,int y) {
   int i,j;
   stype instar;
   float m,dm[2];
   static float *rec[2];
   static int first=1;

   if (first) {
      rec[0]=(float*)calloc(cmdsz[0],sizeof(float));
      rec[1]=(float*)calloc(cmdsz[1]+(cmdsz[0]-1)*2,sizeof(float));
      first=0;
   }
   instar.f=fld;
   instar.x=x;
   instar.y=y;
   i=sub[fld]/2;
   instar.fx=rand()%(i*2+1)-i;
   instar.fy=rand()%(i*2+1)-i;
   dm[0]=((rand()%100)*.01)*0.5;
   dm[1]=((rand()%100)*.01)*0.25;
   for (i=0;i<cmdsz[0];i++) {
      m=i*0.5+dm[0]+limits[0][0];
      instar.s=getflux(m,x,y,0);
      rec[0][i]=runstar(0,&instar);
   }
   for (i=0;i<(cmdsz[0]-1)*2+cmdsz[1];i++) {
      m=(i+1-cmdsz[1])*0.25+dm[0]-dm[1]+limits[0][0]-limits[1][0];
      instar.s=getflux(m,x,y,1);
      rec[1][i]=runstar(1,&instar);
   }
   for (i=0;i<cmdsz[0];i++) for (j=0;j<cmdsz[1];j++) {
      fprintf(of,"%d %6.2f %6.2f",fld,instar.y+.5+instar.fy/(float)sub[fld],instar.x+.5+instar.fx/(float)sub[fld]);
      fprintf(of," %6.3f %6.3f %6.3f %6.3f\n",i*0.5+limits[0][0]+dm[0],rec[0][i],i*0.5+limits[0][0]+dm[0]-j*0.25-limits[1][0]-dm[1],rec[1][i*2+cmdsz[1]-1-j]);
   }
   for (j=0;j<cmdsz[1];j++) {
      fprintf(of,"%d %6.2f %6.2f",fld,instar.y+.5+instar.fy/(float)sub[fld],instar.x+.5+instar.fx/(float)sub[fld]);
      fprintf(of," %6.3f 99.000 %6.3f 99.000\n",i*0.5+limits[0][0]+dm[0],i*0.5+limits[0][0]+dm[0]-j*0.25-limits[1][0]-dm[1]);
   }
   fflush(of);
}

int main(int argc,char**argv) {
   FILE *f[2],*of;
   char str[82];
   int MODE,line,tmp,it,picker[1000],MULT[4],HPP;
   float LUM[16][16],x;
   int i,j,k,ct,px,py;

   if (argc!=18 && argc!=19) {
      printf("Usage: %s <fits base1> <fits base2> <phot base> <Mmin> <Mmax> <Cmin> <Cmax> <N img> <N min> <max chi> <min S/N> <min sh> <max sh> <PC mult> <WF2 mult> <WF3 mult> <WF4 mult> <<aperture>>\n",*argv);
      return 1;
   }
   Nimg=atoi(argv[8]);
   Nmin=atoi(argv[9]);
   maxchi=atof(argv[10]);
   fsig=atof(argv[11]);
   Shmin=atof(argv[12]);
   Shmax=atof(argv[13]);
   MULT[0]=atoi(argv[14]);
   MULT[1]=atoi(argv[15]);
   MULT[2]=atoi(argv[16]);
   MULT[3]=atoi(argv[17]);
  if (argc==19 && atoi(argv[18])) SOLTYPE=0;
   limits[0][0]=atof(argv[4]);
   limits[0][1]=atof(argv[5]);
   if (limits[0][0]>=limits[0][1]) {
      printf("Illegal limits\n");
      exit(0);
   }
   cmdsz[0]=(int)((limits[0][1]-limits[0][0])*2+.5);
   if (cmdsz[0]<1) cmdsz[0]=1;
   limits[0][1]=limits[0][0]+cmdsz[0]*.5;
   limits[1][0]=atof(argv[6]);
   limits[1][1]=atof(argv[7]);
   if (limits[1][0]>=limits[1][1]) {
      printf("Illegal limits\n");
      exit(0);
   }
   cmdsz[1]=(int)((limits[1][1]-limits[1][0])*4+.5);
   if (cmdsz[1]<1) cmdsz[1]=1;
   limits[1][1]=limits[1][0]+cmdsz[1]*.25;
   srand(time(NULL));
   rap[0]=2.;
   rpsf[0]=6;
   sub[0]=5;
   rmin[0]=3.;
   res=res1;
   noise=noise1;
   getpars();
   initfilters();
   for (it=0;it<2;it++) {
      MODE=line=HPP=0;
      sprintf(str,"%s.fits",argv[1+it]);
      if (!(f[it]=fopen(str,"rb"))) {
	 printf("%s not found\n",str);
	 return 1;
      }
      while ((!feof(f[it])) && ((!MODE) || (line%36))) {
	 fgets(str,81,f[it]);
	 line++;
	 if (!MODE) {
	    if (!strncmp(str,"END     ",8)) MODE=1;
	    else if ((!strncmp(str,"SIMPLE  =",9)) && (str[29]!='T')) return ferr(f[it],"Illegal format type");
	    else if ((!strncmp(str,"BITPIX  =",9)) && (strncmp(str+27,"-32",3))) return ferr(f[it],"Illegal bits per pixel");
	    else if ((!strncmp(str,"NAXIS   =",9)) && (strncmp(str+27,"  3",3))) return ferr(f[it],"Illegal number of axes");
	    else if (!strncmp(str,"NAXIS",5)) {
	       tmp=atoi(str+9);
	       if (str[5]=='1') X=tmp;
	       else if (str[5]=='2') Y=tmp;
	       else if (str[5]=='3') Z=tmp;
	    }
	    else if (!strncmp(str,"ATODGAIN=",9)) GAIN[it]=atof(str+9);
	    else if (!strncmp(str,"RNOISE  =",9)) RN[it]=atof(str+9);
	    else if (!strncmp(str,"EPOCH   =",9)) {
	       YEAR[it]=(atof(str+9)-49462.5)/365.25;
	       if (YEAR[it]<0) YEAR[it]=0;
	    }
	    else if (!strncmp(str,"EXPTIME =",9)) EXP[it]=atof(str+9);
	    else if (!strncmp(str,"EXP0    =",9)) EXP0[it]=atof(str+9);
	    else if (!strncmp(str,"RSDPFILL=",9)) DMIN[it]=atoi(str+9);
	    else if (!strncmp(str,"SATURATE=",9)) DMAX[it]=atoi(str+9);
            else if (!strncmp(str,"TEMP    =",9)) TEMP[it]=atoi(str+9);
	    else if (!strncmp(str,"FILTNAM1=",9)) {
	       int i=0;
	       while ((i<19) && (str[i+11]!=32)) {
		  FILT[it][i]=str[i+11];
		  i++;
	       }
	    }
	    else if (!strncmp(str,"HSTPHOT =",9)) HPP=1;
	 }
      }
      if (!HPP) return ferr(f[it],"Not pre-processed by mask");
      if ((X!=800) || (Y!=800) || (Z!=4)) return ferr(f[it],"Illegal axis size");
      printf("Reading FITS file:\n");
      printf("Exp time=%fs, Gain=%f e/ADU, Noise=%f ADU\n",EXP[it],GAIN[it],RN[it]);
      printf("Saturation at %d, bad value=%d\n",DMAX[it],DMIN[it]);
      printf("Filter=\"%s\"\n",FILT[it]);
      filt[it]=findfilt(FILT[it]);
      if (GAIN[it]==7) {
	 corr[it][0]/=1.987;
	 corr[it][1]/=2.003;
	 corr[it][2]/=2.006;
	 corr[it][3]/=1.955;
      }
      else if (GAIN[it]!=14) {
	 printf("Illegal Gain\n");
	 exit(1);
      }
   }
   getpsf();
   for (it=0;it<2;it++) {
      int fld=-1;
      sprintf(str,"%s.%s",argv[3],filters[filt[it]].suffix);
      psfcorr(it,str);
      if ((of=fopen(str,"r"))==NULL) {
	 printf("Error opening %s\n",str);
	 exit(0);
      }
      fscanf(of,"%d",&stars[it][Nstars[it]].f);
      while (!feof(of)) {
	 float m,x,y,z;
	 if (fld!=stars[it][Nstars[it]].f) {
	    fld=stars[it][Nstars[it]].f;
	    sprintf(str,"%s.sky",argv[1]);
	    readsky(0,str);
	    sprintf(str,"%s.sky",argv[2]);
	    readsky(1,str);
	 }
	 fscanf(of,"%f %f %f %f %f %f %f %f %f %f",&y,&x,&z,&z,&m,&z,&z,&z,&z,&z);
	 stars[it][Nstars[it]].x=(int)(x+.5);
	 stars[it][Nstars[it]].fx=(int)((x-stars[it][Nstars[it]].x)*sub[stars[it][Nstars[it]].f]+50.5)-50;
	 stars[it][Nstars[it]].y=(int)(y+.5);
	 stars[it][Nstars[it]].fy=(int)((y-stars[it][Nstars[it]].y)*sub[stars[it][Nstars[it]].f]+50.5)-50;
	 stars[it][Nstars[it]].s=getflux(m,stars[it][Nstars[it]].x,stars[it][Nstars[it]].y,it);
	 Nstars[it]++;
	 if (Nstars[it]>=640001) {
	    printf("Too many stars\n");
	    exit(0);
	 }
	 fscanf(of,"%d",&stars[it][Nstars[it]].f);
      }
      fclose(of);
   }
   sprintf(str,"%s.fake",argv[3]);
   if (!(of=fopen(str,"w"))) {
      printf("Error opening %s\n",str);
      exit(0);
   }
   for (fld=0;fld<4;fld++) {
      if (!fld) printf("** PC **\n");
      else printf("** WFC%d **\n",fld+1);
      fflush(stdout);
      ffread(data[0],sizeof(dtype),1,f[0]);
      ffread(data[1],sizeof(dtype),1,f[1]);
      if (fld>=MMIN && fld<MMAX) {
	 memcpy(res[0],data[0],sizeof(dtype));
	 memcpy(res[1],data[1],sizeof(dtype));
	 sprintf(str,"%s.sky",argv[1]);
	 readsky(0,str);
	 sprintf(str,"%s.sky",argv[2]);
	 readsky(1,str);
	 for (it=0;it<2;it++) for (i=0;i<Nstars[it];i++) if (stars[it][i].f==fld) imgsub(it,i);
	 for (i=1;i<16;i++) for (j=1;j<16;j++) LUM[i][j]=0;
	 for (i=50;i<800;i++) for (j=50;j<800;j++) {
	    if (data[0][i][j]>DMIN[0] && data[0][i][j]<DMAX[0] && res1[0][i][j]>0) LUM[i/50][j/50]+=res[0][i][j];
	    if (data[1][i][j]>DMIN[1] && data[1][i][j]<DMAX[1] && res1[1][i][j]>0) LUM[i/50][j/50]+=res[1][i][j];
	 }
	 x=0;
	 for (i=1;i<16;i++) for (j=1;j<16;j++) {
	    if (LUM[i][j]<0) LUM[i][j]=0;
	    x+=LUM[i][j];
	 }
	 for (i=1;i<16;i++) for (j=1;j<16;j++) if (LUM[i][j]<.002*x) LUM[i][j]=.002*x;
	 x=0;
	 for (i=1;i<16;i++) for (j=1;j<16;j++) x+=LUM[i][j];
	 for (i=1;i<16;i++) for (j=1;j<16;j++) if (LUM[i][j]<.002*x) LUM[i][j]=.002*x;
	 x=0;
	 for (i=1;i<16;i++) for (j=1;j<16;j++) x+=LUM[i][j];
	 *picker=0;
	 for (i=1;i<16;i++) for (j=1;j<16;j++) {
	    it=(int)(800*LUM[i][j]/x+.5);
	    if (it<1) it=1;
	    while (it>0) {
	       picker[++(*picker)]=i*16+j;
	       it--;
	    }
	 }
	 for (i=1;i<*picker;i++) {
	    j=rand()/4%(*picker-i+1)+i;
	    k=picker[j];
	    picker[j]=picker[i];
	    picker[i]=k;
	 }
	 for (ct=0;ct<MULT[fld]*(*picker);ct++) {
	    k=picker[ct%(*picker)+1];
	    do {
	       px=(k/16)*50+rand()/16%50;
	       py=(k%16)*50+rand()/16%50;
	    } while (px<50+rpsf[fld] || px>=800-rpsf[fld] || py<50+rpsf[fld] || py>=800-rpsf[fld]);
	    onestar(of,px,py);
	 }
      }
   }
   fclose(f[0]);
   fclose(f[1]);
   fclose(of);
   return 0;
}
