#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"

typedef float dtype[800][800];
typedef double ddtype[800][800];
typedef struct{int x,y,fx,fy,flag,type;double s,ss,chi,sh,sky,ffx,ffy;} stype;
typedef float ****psftype;
typedef float **ppsftype;
typedef double real;

stype *stars;
int X,Y,Z,DMAX,DMIN,IGN,TEMP=0,XMIN,XMAX,YMIN,YMAX,XMIN2=50,XMAX2=800,YMIN2=50,YMAX2=800,MMIN=0,MMAX=3,indx[800][800][5],OPTS,NO_SKY=0;
float ran[800][800],cran[800][800],rmin[4],rap[4],****cwt[4],**poff,pixmap[80][80][2],exmap[80][80];
real EXP,GAIN,RN,YEAR,EXP0,corr[4]={1.0066,1.005,1,1.0013},fsig;
dtype data,sky;
ddtype noise,res;
psftype psf[4][8][8];
ppsftype ppsf[4][8][8];
double **psf1,**tpsf1;
char FILT[20]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int rphot[4],rpsf[4],sub[4],***circ[2][4],Nstars,fld,Gauss_PSF=0,MaxIT=25,filt;
float dnoise[4]={0.0361,0.0625,0.0625,0.0625};
float rows[800];

void merr(void) {
   printf("****Mem error\n");
   exit(-1);
}

void getpars(void) {
   int mx,my,x,y,f,rad;
   for (x=0;x<800;x++) for (y=0;y<800;y++) cran[x][y]=-10000;
   rpsf[1]=rpsf[2]=rpsf[3]=(rpsf[0]*2+1)/3;
   sub[1]=sub[2]=sub[3]=sub[0]*2;
   rap[1]=rap[2]=rap[3]=rap[0]*2./3.;
   for (f=0;f<4;f++) rphot[f]=(int)(rap[f]+.9);
   rmin[1]=rmin[2]=rmin[3]=rmin[0]*2./3.;
   for (f=0;f<4;f++) {
      rad=(sub[f]+2)/2;
      circ[0][f]=(int***)calloc(rad*2+1,sizeof(int**));
      circ[1][f]=(int***)calloc(rad*2+1,sizeof(int**));
      cwt[f]=(float****)calloc(rad*2+1,sizeof(float***));
      if (!circ[0][f] || !circ[1][f] || !cwt[f]) merr();
      circ[0][f]+=rad;
      circ[1][f]+=rad;
      cwt[f]+=rad;
      for (mx=-rad;mx<=rad;mx++) {
	 circ[0][f][mx]=(int**)calloc(rad*2+1,sizeof(int*));
	 circ[1][f][mx]=(int**)calloc(rad*2+1,sizeof(int*));
	 cwt[f][mx]=(float***)calloc(rad*2+1,sizeof(float**));
	 if (!circ[0][f][mx] || !circ[1][f][mx] || !cwt[f][mx]) merr();
	 circ[0][f][mx]+=rad;
	 circ[1][f][mx]+=rad;
	 cwt[f][mx]+=rad;
	 for (my=-rad;my<=rad;my++) {
	    circ[0][f][mx][my]=(int*)calloc(rphot[f]*2+1,sizeof(int));
	    circ[1][f][mx][my]=(int*)calloc(rphot[f]*2+1,sizeof(int));
	    cwt[f][mx][my]=(float**)calloc(rphot[f]*2+1,sizeof(float*));
	    if (!circ[0][f][mx][my] || !circ[1][f][mx][my] || !cwt[f][mx][my]) merr();
	    circ[0][f][mx][my]+=rphot[f];
	    circ[1][f][mx][my]+=rphot[f];
	    cwt[f][mx][my]+=rphot[f];
	    for (x=-rphot[f];x<=rphot[f];x++) {
	       circ[0][f][mx][my][x]=50;
	       circ[1][f][mx][my][x]=-50;
	       cwt[f][mx][my][x]=(float*)calloc(rphot[f]*2+1,sizeof(float));
	       if (!cwt[f][mx][my][x]) merr();
	       cwt[f][mx][my][x]+=rphot[f];
	       for (y=-rphot[f];y<=rphot[f];y++) {
		  cwt[f][mx][my][x][y]=rap[f]+.465-sqrt((x-mx/(float)sub[f])*(x-mx/(float)sub[f])+(y-my/(float)sub[f])*(y-my/(float)sub[f]));
		  if (cwt[f][mx][my][x][y]<0) cwt[f][mx][my][x][y]=0;
		  else if (cwt[f][mx][my][x][y]>1) cwt[f][mx][my][x][y]=1;
		  if (cwt[f][mx][my][x][y]>0 && y<circ[0][f][mx][my][x]) circ[0][f][mx][my][x]=y;
		  if (cwt[f][mx][my][x][y]>0 && y>circ[1][f][mx][my][x]) circ[1][f][mx][my][x]=y;
	       }
	    }
	 }
      }
   }
   stars=(stype*)calloc(160000,sizeof(stype));
   if (!stars) merr();
}

//uses Holtzman et at (1995), PASP 107, 156
double CC[9][4]=
{{1.00021E+00,1.95172E-02,-2.18757E+00,1.37619E-02},
 {9.79758E-04,-2.18805E+00,-8.51284E-03,2.18899E+00},
 {9.84222E-08,-1.51015E-06,-3.21458E-07,1.33326E-06},
 {-6.31327E-09,5.06693E-06,2.77459E-06,-4.00828E-06},
 {-7.19983E-07,1.04063E-06,-2.06874E-06,-5.71282E-07},
 {-3.73922E-08,-4.55094E-10,7.39223E-08,-2.00520E-09},
 {6.65101E-10,7.50249E-08,-1.18833E-09,-7.86673E-08},
 {-3.51470E-08,-1.51652E-09,7.73056E-08,-3.37491E-09},
 {-2.55003E-09,7.34568E-08,-3.72971E-10,-7.81490E-08}};
double DD[9][4]=
{{1.00544E-03,2.18607E+00,7.40319E-03,-2.18688E+00},
 {9.99786E-01,2.02006E-02,-2.18554E+00,1.29831E-02},
 {-5.80870E-07,-3.79469E-06,-1.49841E-06,1.30814E-06},
 {-5.07221E-07,-2.99518E-06,2.96650E-06,2.01760E-06},
 {3.41574E-07,1.05789E-07,-2.02756E-07,-8.47656E-07},
 {-1.08091E-09,-7.47857E-08,-1.45720E-09,7.70408E-08},
 {-3.42070E-08,-1.67666E-09,7.72421E-08,-1.68514E-09},
 {1.17819E-09,-7.55302E-08,8.52633E-10,7.75257E-08},
 {-4.48966E-08,3.02579E-10,7.68023E-08,-3.69210E-10}};

void transformpix(int c,double x,double y,double*xc,double*yc) {
   double a,b;
   a=CC[0][c]*x+CC[1][c]*y+CC[2][c]*x*x+CC[3][c]*x*y+CC[4][c]*y*y+CC[5][c]*x*x*x+CC[6][c]*x*x*y+CC[7][c]*x*y*y+CC[8][c]*y*y*y;
   b=DD[0][c]*x+DD[1][c]*y+DD[2][c]*x*x+DD[3][c]*x*y+DD[4][c]*y*y+DD[5][c]*x*x*x+DD[6][c]*x*x*y+DD[7][c]*x*y*y+DD[8][c]*y*y*y;
   switch(c) {
   case 0:
      *xc=a/CC[0][0];
      *yc=b/DD[1][0];
      break;
   case 1:
      *xc=b/DD[0][1];
      *yc=a/CC[1][1];
      break;
   case 2:
      *xc=a/CC[0][2];
      *yc=b/DD[1][2];
      break;
   case 3:
      *xc=b/DD[0][3];
      *yc=a/CC[1][3];
      break;
   }
}

double pixdx(int c,int x,int y) {
   double dx=0,t,q;
   transformpix(c,x+.5,y,&dx,&q);
   transformpix(c,x-.5,y,&t,&q);
   return dx-t;
}

double pixdy(int c,int x,int y) {
   double dy=0,t,q;
   transformpix(c,x,y+.5,&q,&dy);
   transformpix(c,x,y-.5,&q,&t);
   return dy-t;
}

void initpixmap(void) {
   int x,y;
   double xx,yy;
   double ex[4][6]={
      {0.011812,0.022028,0.018977,-0.052647,-0.009532,-0.039493},
      {0.019266,-0.006057,-0.006740,-0.042960,0.016111,-0.044705},
      {0.035159,-0.007716,0.038275,-0.027663,-0.006167,0.017977},
      {0.019981,0.031440,0.018518,-0.005264,0.019414,-0.066667}
   };

   for (x=5;x<800;x+=10) for (y=5;y<800;y+=10) {
      pixmap[y/10][x/10][0]=pixdy(fld,x-400,y-400);
      pixmap[y/10][x/10][1]=pixdx(fld,x-400,y-400);
      xx=x/400.-1;
      yy=y/400.-1;
      exmap[y/10][x/10]=ex[fld][0]+ex[fld][1]*xx+ex[fld][2]*yy+ex[fld][3]*xx*xx+ex[fld][4]*xx*yy+ex[fld][5]*yy*yy;
   }
}

double rowsize(int x) {
   //Ron Gilliland, private communication for raw data
   int Nrows=32,rows[32]={64,98,132,166,200,201,234,235,268,269,302,303,337,371,405,439,440,473,474,507,508,541,542,576,610,644,678,712,746,747,780,781};
   double col0[32]={0.9789,0.9878,0.9675,0.9696,0.9829,0.9985,0.9806,0.9796,0.989,0.9803,0.9985,0.9855,0.9683,0.9849,0.9648,0.982,0.9978,0.9813,0.9851,0.9819,0.977,0.9979,0.9834,0.9674,0.9842,0.9667,0.9697,0.9801,0.9807,0.9799,0.9872,0.98};
   double dcol[32]={-3.42557e-06,-3.83428e-05,-3.92468e-06,-5.96548e-06,-3.72032e-05,-1.58276e-05,9.18357e-06,7.45086e-06,-2.89753e-07,-5.09053e-06,-1.21806e-05,-3.36296e-05,-2.00527e-06,-4.03303e-05,1.91426e-06,-2.69071e-05,-1.06506e-05,1.08298e-05,4.22073e-06,9.60101e-06,1.04708e-05,-1.01982e-05,-3.10483e-05,-1.51857e-06,-3.55187e-05,-4.69483e-06,-5.38823e-07,-3.14124e-05,1.22734e-05,3.24158e-06,1.91252e-07,-1.48993e-06};
   double lam0=568.104;
   int i;

   for (i=0;i<Nrows;i++) if (x==rows[i]) {
      double p;
      p=col0[i]+(atoi(FILT+1)-lam0)*dcol[i];
      if (p>1.) return 1.;
      if (p<0.) return 0.;
      return p;
   }
   return 1.;
}

void initrows(void) {
   int y;
   for (y=0;y<800;y++) rows[y]=rowsize(y);
}

int peak2(int x,int y) {
   if (x>0 && data[x-1][y]>data[x][y] && data[x-1][y]<DMAX) return 0;
   if (x<XMAX-1 && data[x+1][y]>data[x][y] && data[x+1][y]<DMAX) return 0;
   if (y>0 && data[x][y-1]>data[x][y] && data[x][y-1]<DMAX) return 0;
   if (y<YMAX-1 && data[x][y+1]>data[x][y] && data[x][y+1]<DMAX) return 0;
   return 1;
}

void cleansat(void) {
   int x,y,c=1;
   float DDMAX=DMAX*0.855;
   for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) {
      if (data[x][y]>=DDMAX) {
	 cran[x][y]=1;
	 data[x][y]=DMAX*0.99;
      }
      else cran[x][y]=0;
   }
   while (c) {
      c=0;
      memcpy(ran,cran,sizeof(float[800][800]));
      for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) cran[x][y]=0;
      for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) if (data[x][y]<DMAX && data[x][y]>DMIN && ran[x][y] && peak2(x,y)) {
	 data[x][y]=DMAX;
	 if (x>XMIN) cran[x-1][y]=1;
	 if (x<XMAX-1) cran[x+1][y]=1;
	 if (y>YMIN) cran[x][y-1]=1;
	 if (y<YMAX-1) cran[x][y+1]=1;
	 c=1;
      }
   }
   for (x=0;x<800;x++) for (y=0;y<800;y++) cran[x][y]=-10000;
}

void readsky(char *fn) {
   int f,x,y;
   double z;
   FILE *sf;

   if ((sf=fopen(fn,"rb"))==NULL) {
      printf("****Cannot open file %s\n",fn);
      exit(-1);
   }
   for (f=0;f<=fld;f++) ffread(sky,800*800,4,sf);
   for (x=0;x<800;x++) for (y=0;y<800;y++) {
      z=pixmap[x/10][y/10][0]*pixmap[x/10][y/10][1]*rows[x];
      if (data[x][y]>DMIN && data[x][y]<DMAX) data[x][y]*=z;
      sky[x][y]*=z;
      res[x][y]=res[x][y]*z-sky[x][y];
      noise[x][y]=RN*RN;
      if (res[x][y]>0) noise[x][y]+=res[x][y]/GAIN;
      if (sky[x][y]>0) noise[x][y]+=sky[x][y]/GAIN;
   }
   fclose(sf);
}

void getspsf(int f) {
   int x,y,x2,y2,x1,y1,r,rad,N;
   float r2,av;

   if (!(OPTS&2)) return;
   r=rpsf[f];
   if (r>rphot[f]+2) r=rphot[f]+2;
   rad=(sub[f]+2)/2;
   for (x=0;x<8;x++) for (y=0;y<8;y++) for (x2=-rad;x2<=rad;x2++) for (y2=-rad;y2<=rad;y2++) {
      av=0;
      N=0;
      for (x1=-r;x1<=r;x1++) for (y1=-r;y1<=r;y1++) {
	 r2=sqrt(x1*x1+y1*y1);
	 if (r2>=rphot[f]+1 && r2<=r) {
	    av+=psf[f][x][y][x2][y2][x1][y1];
	    N++;
	 }
      }
      av/=N;
      for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) psf[f][x][y][x2][y2][x1][y1]-=av;
      ppsf[f][x][y][x2][y2]+=av;
   }
}

void getpsf(void) {
   int f,x,y,z,rad,px,py;

   for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
      rad=(sub[f]+2)/2;
      psf[f][px][py]=(float****)calloc(rad*2+1,sizeof(float***));
      ppsf[f][px][py]=(float**)calloc(rad*2+1,sizeof(float*));
      if (!psf[f][px][py] || !ppsf[f][px][py]) merr();
      psf[f][px][py]+=rad;
      ppsf[f][px][py]+=rad;
      for (x=-rad;x<=rad;x++) {
	 psf[f][px][py][x]=(float***)calloc(rad*2+1,sizeof(float**));
	 ppsf[f][px][py][x]=(float*)calloc(rad*2+1,sizeof(float));
	 if (!psf[f][px][py][x] || !ppsf[f][px][py][x]) merr();
	 psf[f][px][py][x]+=rad;
	 ppsf[f][px][py][x]+=rad;
	 for (y=-rad;y<=rad;y++) {
	    psf[f][px][py][x][y]=(float**)calloc(2*rpsf[f]+1,sizeof(float*));
	    if (!psf[f][px][py][x][y]) merr();
	    psf[f][px][py][x][y]+=rpsf[f];
	    for (z=-rpsf[f];z<=rpsf[f];z++) {
	       psf[f][px][py][x][y][z]=(float*)calloc(2*rpsf[f]+1,sizeof(float));
	       if (!psf[f][px][py][x][y][z]) merr();
	       psf[f][px][py][x][y][z]+=rpsf[f];
	    }
	 }
      }
   }
   if (Gauss_PSF) {
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 int x1,y1;
	 float sig,n;

	 rad=(sub[f]+2)/2;
	 if (f) sig=0.54;
	 else sig=0.84;
	 for (x=-rad;x<=rad;x++) for (y=-rad;y<=rad;y++) {
	    n=0;
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) n+=(psf[f][px][py][x][y][x1][y1]=exp(-.5*((x1-x/(float)sub[f])*(x1-x/(float)sub[f])+(y1-y/(float)sub[f])*(y1-y/(float)sub[f]))/sig/sig));
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) psf[f][px][py][x][y][x1][y1]/=n;
	 }
      }
   }
   else {
      FILE *pf;
      int x1;
      char fn[80];

      sprintf(fn,"%s/psfs/%s.%d.%d.psf",BASEDIR,filters[filt].suffix,rpsf[0],sub[0]);
      if ((pf=fopen(fn,"rb"))==NULL) {
	 printf("****File error: %s\n",fn);
	 exit(-1);
      }
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 for (x=-(sub[f]+2)/2;x<=(sub[f]+2)/2;x++) for (y=-(sub[f]+2)/2;y<=(sub[f]+2)/2;y++) {
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) ffread(psf[f][px][py][x][y][x1]-rpsf[f],2*rpsf[f]+1,sizeof(float),pf);
	 }
      }
      fclose(pf);
   }
   for (f=0;f<4;f++) getspsf(f);
}

void addindx(int x,int y,int i) {
   if (indx[x][y][0]>=4) {
      printf("****Index overflow\n");
      exit(-1);
   }
   indx[x][y][++indx[x][y][0]]=i;
}

void subindx(int x,int y,int i) {
   int j;
   if (indx[x][y][0]==0) {
      printf("****Index underflow\n");
      exit(-1);
   }
   for (j=1;j<=indx[x][y][0];j++) if (indx[x][y][j]==i) {
      indx[x][y][j]=indx[x][y][indx[x][y][0]--];
      return;
   }
   printf("****Subindx not found\n");
   exit(-1);
}

void repindx(int x,int y,int i,int ii) {
   int j;
   for (j=1;j<=indx[x][y][0];j++) if (indx[x][y][j]==i) {
      indx[x][y][j]=ii;
      return;
   }
   printf("****Repindx not found\n");
   exit(-1);
}

void calc1psf(int r,int x,int y,int fx,int fy,double ffx,double ffy) {
   int xx,yy;
   double z,zz;
   static int lastr=-1;
   static double *xoff;

   if (r>lastr) {
      if (lastr>=0) {
	 for (xx=-lastr;xx<=lastr;xx++) {
	    free(psf1[xx]-lastr);
	    free(tpsf1[xx]-lastr);
	 }
	 free(psf1-lastr);
	 free(tpsf1-lastr);
	 free(xoff-lastr);
      }
      psf1=(double**)calloc(sizeof(double*),r*2+1);
      tpsf1=(double**)calloc(sizeof(double*),r*2+1);
      xoff=(double*)calloc(sizeof(double),r*2+2);
      if (!psf1 || !tpsf1 || !xoff) merr();
      psf1+=r;
      tpsf1+=r;
      xoff+=r;
      for (xx=-r;xx<=r;xx++) {
	 psf1[xx]=(double*)calloc(sizeof(double),r*2+1);
	 tpsf1[xx]=(double*)calloc(sizeof(double),r*2+1);
	 if (!psf1[xx] || !tpsf1[xx]) merr();
	 psf1[xx]+=r;
	 tpsf1[xx]+=r;
      }
      lastr=r;
   }
   if (ffx==0 && ffy==0) {
      for (xx=-r;xx<=r;xx++) for (yy=-r;yy<=r;yy++) psf1[xx][yy]=psf[fld][y/100][x/100][fx][fy][xx][yy];
   }
   else {
      int x0,y0;
      double mx,my;
      if (ffx<0) {
	 x0=fx-1;
	 mx=ffx+1;
      }
      else {
	 x0=fx;
	 mx=ffx;
      }
      if (ffy<0) {
	 y0=fy-1;
	 my=ffy+1;
      }
      else {
	 y0=fy;
	 my=ffy;
      }
      for (xx=-r;xx<=r;xx++) for (yy=-r;yy<=r;yy++) psf1[xx][yy]=psf[fld][y/100][x/100][x0][y0][xx][yy]*(1-mx)*(1-my)+psf[fld][y/100][x/100][x0+1][y0][xx][yy]*mx*(1-my)+psf[fld][y/100][x/100][x0][y0+1][xx][yy]*(1-mx)*my+psf[fld][y/100][x/100][x0+1][y0+1][xx][yy]*mx*my;
   }
   zz=y-(int)y-0.5;
   z=pixmap[x/10][y/10][1]+exmap[x/10][y/10];
   xoff[1]=(1-z)*(0.5-zz);
   xoff[0]=-(1-z)*(0.5+zz);
   for (xx=2;xx<=r+1;xx++) xoff[xx]=xoff[xx-1]+(1-z);
   for (xx=-1;xx>=-r;xx--) xoff[xx]=xoff[xx+1]-(1-z);
   for (xx=-r;xx<=r;xx++) for (yy=-r;yy<=r;yy++) tpsf1[xx][yy]=0;
   for (xx=-r;xx<=r;xx++) for (yy=-r;yy<=r;yy++) {
      float q;
      tpsf1[xx][yy]+=(q=psf1[xx][yy]);
      if (xoff[yy]<0) {
	 if (yy>-r) {
	    tpsf1[xx][yy-1]+=0.5*(q+psf1[xx][yy-1])*(-xoff[yy]);
	    tpsf1[xx][yy]-=0.5*(q+psf1[xx][yy-1])*(-xoff[yy]);
	 }
	 else tpsf1[xx][yy]-=0.75*q*(-xoff[yy]);
      }
      if (xoff[yy+1]>0) {
	 if (yy<r) {
	    tpsf1[xx][yy+1]+=0.5*(q+psf1[xx][yy+1])*xoff[yy+1];
	    tpsf1[xx][yy]-=0.5*(q+psf1[xx][yy+1])*xoff[yy+1];
	 }
	 else tpsf1[xx][yy]-=0.75*q*xoff[yy+1];
      }
      if (xoff[yy]>0 && yy==-r) tpsf1[xx][yy]+=0.75*q*xoff[yy];
      if (xoff[yy+1]<0 && yy==r) tpsf1[xx][yy]+=0.75*q*(-xoff[yy+1]);
   }
   zz=x-(int)x-0.5;
   z=pixmap[x/10][y/10][0]+exmap[x/10][y/10];
   xoff[1]=(1-z*(1-(1-rows[x])*0.5))*(0.5-zz);
   xoff[0]=-(1-z*(1-(1-rows[x])*0.5))*(0.5+zz);
   for (yy=2;yy<=r+1;yy++) if (x+yy-1>=0 && x+yy-1<800) xoff[yy]=xoff[yy-1]+(1-z*(1-(1-rows[x+yy-1])*0.5));
   else xoff[yy]=xoff[yy-1]+(1-z);
   for (yy=-1;yy>=-r;yy--) if (x+yy+1>=0 && x+yy+1<800) xoff[yy]=xoff[yy+1]-(1-z*(1-(1-rows[x+yy+1])*0.5));
   else xoff[yy]=xoff[yy+1]-(1-z);
   for (xx=-r;xx<=r;xx++) for (yy=-r;yy<=r;yy++) psf1[xx][yy]=0;
   for (yy=-r;yy<=r;yy++) for (xx=-r;xx<=r;xx++) {
      float q;
      psf1[xx][yy]+=(q=tpsf1[xx][yy]);
      if (xoff[xx]<0) {
	 if (xx>-r) {
	    psf1[xx-1][yy]+=0.5*(q+tpsf1[xx-1][yy])*(-xoff[xx]);
	    psf1[xx][yy]-=0.5*(q+tpsf1[xx-1][yy])*(-xoff[xx]);
	 }
	 else psf1[xx][yy]-=0.75*q*(-xoff[xx]);
      }
      if (xoff[xx+1]>0) {
	 if (xx<r) {
	    psf1[xx+1][yy]+=0.5*(q+tpsf1[xx+1][yy])*xoff[xx+1];
	    psf1[xx][yy]-=0.5*(q+tpsf1[xx+1][yy])*xoff[xx+1];
	 }
	 else psf1[xx][yy]-=0.75*q*xoff[xx+1];
      }
      if (xoff[xx]>0 && xx==-r) psf1[xx][yy]+=0.75*q*xoff[xx];
      if (xoff[xx+1]<0 && xx==r) psf1[xx][yy]+=0.75*q*(-xoff[xx+1]);
   }
}

void mark1pos(int x,int y) {
   int r,xx,yy,i;
   r=rphot[fld]+rpsf[fld];
   for (xx=-r;xx<=r;xx++) if (x+xx>=XMIN && x+xx<XMAX) for (yy=-r;yy<=r;yy++) if (y+yy>=YMIN && y+yy<YMAX) for (i=1;i<=indx[x+xx][y+yy][0];i++) stars[indx[x+xx][y+yy][i]].flag=1;
}

void mark2pos(int x1,int y1,int x2,int y2) {
   int r,xx,yy,i;
   r=rphot[fld]+rpsf[fld];
   for (xx=-r;xx<=r;xx++) if (x1+xx>=XMIN && x1+xx<XMAX) for (yy=-r;yy<=r;yy++) if (y1+yy>=YMIN && y1+yy<YMAX) for (i=1;i<=indx[x1+xx][y1+yy][0];i++) stars[indx[x1+xx][y1+yy][i]].flag=1;
   for (xx=-r;xx<=r;xx++) if (x2+xx>=XMIN && x2+xx<XMAX) for (yy=-r;yy<=r;yy++) if (y2+yy>=YMIN && y2+yy<YMAX) for (i=1;i<=indx[x2+xx][y2+yy][0];i++) stars[indx[x2+xx][y2+yy][i]].flag=1;
}

float getsky_norm(int x,int y) {
   static float *list=NULL;
   double s,sig;
   int i,n=0,x1,y1,rsky=20,rin=8,ss=1;

   if (NO_SKY) return 0;
   if (!(OPTS&512)) return 0;
   if (fld==0) {rsky=36; rin=15; ss=2;}
   if (list==NULL) {
      list=(float*)calloc((2*36+1)*(2*36+1),sizeof(float));
      if (!list) merr();
   }
   for (x1=x-rsky;x1<=x+rsky;x1+=ss) for (y1=y-rsky;y1<=y+rsky;y1+=ss) if (x1>=0 && x1<800 && y1>=0 && y1<800 && abs(x-x1)>rin && abs(y-y1)>rin && data[x1][y1]>DMIN && data[x1][y1]<DMAX*0.855) list[n++]=res[x1][y1];
   x1=1;
   while (x1) {
      s=0;
      y1=0;
      for (i=0;i<n;i++) {
	 s+=list[i];
	 y1++;
      }
      if (y1==0) s=x1=0;
      else {
	 s/=y1;
	 sig=0;
	 for (i=0;i<n;i++) sig+=(list[i]-s)*(list[i]-s);
	 if (y1>1) sig=sqrt(sig/(y1-1))*2.25;
	 else sig=sqrt(sig)*2.25;
	 x1=0;
	 for (i=0;i<n;i++) if (fabs(list[i]-s)>sig) {
	    list[i--]=list[--n];
	    x1=1;
	 }
      }
   }
   return s;
}

float getsky(int x,int y) {
   int x1,y1,r,i,N=0;
   float r2,av,sd,z;
   static float *list;
   static int first=1;

   if (!(OPTS&2)) return getsky_norm(x,y);
   z=-RN*RN-1;
   r=rpsf[fld];
   if (r>rphot[fld]+2) r=rphot[fld]+2;
   if (first) {
      list=(float*)calloc(sizeof(float),4*r*r);
      if (!list) merr();
      first=0;
   }
   for (x1=x-r;x1<=x+r;x1++) for (y1=y-r;y1<=y+r;y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      r2=sqrt((x1-x)*(x1-x)+(y1-y)*(y1-y));
      if (r2>=rphot[fld]+1 && r2<=r && (res[x1][y1]>=z && res[x1][y1]>=data[x1][y1]-sky[x1][y1]-1)) list[N++]=res[x1][y1];
   }
   r=1;
   av=sd=0;
   if (N==1) av=list[0];
   while (r && N>1) {
      av=sd=0;
      for (i=0;i<N;i++) av+=list[i];
      av/=N;
      for (i=0;i<N;i++) sd+=(list[i]-av)*(list[i]-av);
      sd=sqrt(sd/(N-1))*2.5;
      r=0;
      for (i=0;i<N;)
	 if (fabs(list[i]-av)>sd) {
	    list[i]=list[--N];
	    r=1;
	 }
	 else i++;
   }
   return av;
}

void centroid(int x,int y,int*xc,int*yc,int*fx,int*fy) {
   int xx,yy;
   float X=0,Y=0,n=0,min=0;
   for (xx=-2;xx<=2;xx++) for (yy=-2;yy<=2;yy++) if (abs(xx)+abs(yy)<4 && x+xx>=XMIN && x+xx<XMAX && y+yy>=YMIN && y+yy<YMAX && data[x+xx][y+yy]>DMIN && data[x+xx][y+yy]<DMAX && res[x+xx][y+yy]<min) min=res[x+xx][y+yy];
   for (xx=-2;xx<=2;xx++) for (yy=-2;yy<=2;yy++) if (abs(xx)+abs(yy)<4) {
      if (x+xx>=XMIN && x+xx<XMAX && y+yy>=YMIN && y+yy<YMAX && data[x+xx][y+yy]>DMIN && data[x+xx][y+yy]<DMAX) {
	 if (res[x+xx][y+yy]>min) {
	    X+=xx*(res[x+xx][y+yy]-min);
	    Y+=yy*(res[x+xx][y+yy]-min);
	    n+=(res[x+xx][y+yy]-min);
	 }
      }
      else if (x-xx>=XMIN && x-xx<XMAX && y-yy>=YMIN && y-yy<YMAX && data[x-xx][y-yy]>DMIN && data[x-xx][y-yy]<DMAX && res[x-xx][y-yy]>min) {
	 X+=xx*(res[x-xx][y-yy]-min);
	 Y+=yy*(res[x-xx][y-yy]-min);
	 n+=(res[x-xx][y-yy]-min);
      }
   }
   if (n>0) {
      X/=n;
      Y/=n;
   }
   X+=x;
   Y+=y;
   *xc=(int)(X+.5);
   *fx=(int)((X-(*xc))*sub[fld]+100.5)-100.5;
   if (*xc<XMIN) {
      *xc=XMIN;
      *fx=-(sub[fld]/2);
   }
   if (*xc>=XMAX) {
      *xc=XMAX;
      *fx=sub[fld]/2;
   }
   *yc=(int)(Y+.5);
   *fy=(int)((Y-(*yc))*sub[fld]+100.5)-100.5;
   if (*yc<YMIN) {
      *yc=YMIN;
      *fy=-(sub[fld]/2);
   }
   if (*yc>=YMAX) {
      *yc=YMAX;
      *fy=sub[fld]/2;
   }
}

void photsearch(int x,int y,int fx,int fy,int*first,int*sx,int*sy,int*bx,int*by,real*bfx,real*bfy,real*bs,real*bss,real*bsky,real*bm,real*bc,real*bsh,int finding) {
   int xbx=0,xby=0,rad,x1,y1,x2,y2,xfirst=1,set=0,DONE=0;
   real xbm=0,c,sh,m,cs,cc,s,ss,ssky,a,b,A,np;
   static real **result=NULL;
   static int lastf=-1,lastrad=0;

   if (x<=XMIN || x>=XMAX-1 || y<=YMIN || y>=YMAX-1) return;
   if (finding) {
      if (ran[x][y]==-9999) return;
      if (ran[x][y]>-10000) {
	 if (ran[x][y]>*bm) {
	    *first=1;
	    *bm=ran[x][y];
	 }
	 return;
      }
   }
   else if (ran[x][y]>-10000) return;
   rad=(sub[fld]+2)/2;
   if (fld!=lastf) {
      int i;
      if (lastf>=0) {
	 for (i=-lastrad;i<=lastrad;i++) free(result[i]-lastrad);
	 free(result-lastrad);
      }
      result=(real**)calloc(sizeof(real*),rad*2+1);
      if (!result) merr();
      result+=rad;
      for (i=-rad;i<=rad;i++) {
	 result[i]=(real*)calloc(sizeof(real),rad*2+1);
	 if (!result[i]) merr();
	 result[i]+=rad;
      }
      lastf=fld;
      lastrad=rad;
   }
   ssky=getsky(x,y);
   for (x2=-rad;x2<=rad;x2++) for (y2=-rad;y2<=rad;y2++) result[x2][y2]=-1;
   while (!DONE) {
      for (x2=fx-1;x2<=fx+1;x2++) for (y2=fy-1;y2<=fy+1;y2++) if (result[x2][y2]<0) {
	 calc1psf(rphot[fld],x,y,x2,y2,0,0);
	 cs=s=ss=c=cc=a=b=0;
	 for (x1=x-rphot[fld];x1<=x+rphot[fld];x1++) for (y1=y+circ[0][fld][x2][y2][x1-x];y1<=y+circ[1][fld][x2][y2][x1-x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
	    if (OPTS&1) {
	       cc+=cwt[fld][x2][y2][x1-x][y1-y];
	       c+=psf1[x1-x][y1-y]*cwt[fld][x2][y2][x1-x][y1-y];
	       s+=(res[x1][y1]-ssky)*cwt[fld][x2][y2][x1-x][y1-y];
	       ss+=noise[x1][y1]*cwt[fld][x2][y2][x1-x][y1-y];
	    }
	    else {
	       a=psf1[x1-x][y1-y];
	       b=a*cwt[fld][x2][y2][x1-x][y1-y]/noise[x1][y1];
	       cc+=b;
	       if (a>0) {
		  s+=(res[x1][y1]-ssky)*b;
		  a*=b;
		  c+=a;
		  cs+=a;
		  ss+=a*cwt[fld][x2][y2][x1-x][y1-y];
	       }
	       else cs+=a*b;
	    }
	 }
	 if (ss>0 && c!=0) {
	    if (OPTS&1) cs=c;
	    cs/=cc;
	    s/=c;
	    ss=sqrt(ss)/c;
	    c=sh=0;
	    A=np=0;
	    for (x1=x-rphot[fld];x1<=x+rphot[fld];x1++) for (y1=y+circ[0][fld][x2][y2][x1-x];y1<=y+circ[1][fld][x2][y2][x1-x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
	       a=psf1[x1-x][y1-y]*s;
	       b=1/(noise[x1][y1]+dnoise[fld]*(a+ssky)*(a+ssky));
	       a=res[x1][y1]-ssky-a;
	       c+=a*a*b*cwt[fld][x2][y2][x1-x][y1-y];
	       np+=cwt[fld][x2][y2][x1-x][y1-y];
	       sh+=a/sqrt(noise[x1][y1])*(psf1[x1-x][y1-y]-cs)*cwt[fld][x2][y2][x1-x][y1-y];
	       A+=(psf1[x1-x][y1-y]-cs)*(psf1[x1-x][y1-y]-cs)*s/sqrt(noise[x1][y1])*cwt[fld][x2][y2][x1-x][y1-y];
	    }
	    if (A>0) sh/=A;
	    else sh=0;
	    c=c/np;
	    m=s/(ss*sqrt(c+0.1));
	    if (c>1) ss*=sqrt(c);
	    if (m>0) result[x2][y2]=m;
	    else result[x2][y2]=0;
	    if (m>xbm || xfirst) {
	       xfirst=0;
	       xbm=m;
	       xbx=x2;
	       xby=y2;
	    }
	    if (abs(x2)<rad && abs(y2)<rad && (m>*bm || *first)) {
	       set=1;
	       *first=0;
	       *bc=c;
	       *bsh=sh;
	       *bm=m;
	       *bs=s;
	       *bx=x2;
	       *by=y2;
	       *bfx=0;
	       *bfy=0;
	       *sx=x;
	       *sy=y;
	       *bss=ss;
	       *bsky=ssky;
	    }
	 }
	 else result[x2][y2]=0;
      }
      DONE=1;
      if (set && (result[*bx][*by+1]<0 || result[*bx+1][*by+1]<0 || result[*bx+1][*by]<0 || result[*bx+1][*by-1]<0 || result[*bx][*by-1]<0 || result[*bx-1][*by-1]<0 || result[*bx-1][*by]<0 || result[*bx-1][*by+1]<0)) DONE=0;
      fx=*bx;
      fy=*by;
   }
   if (set && result[*bx][*by]>0) {
      if (result[*bx+1][*by]>=result[*bx][*by]) *bfx=0.5;
      else if (result[*bx-1][*by]>=result[*bx][*by]) *bfx=-0.5;
      else *bfx=(result[*bx+1][*by]-result[*bx-1][*by])/(4*result[*bx][*by]-2*result[*bx+1][*by]-2*result[*bx-1][*by]);
      if (*bfx>0.5) *bfx=0.5;
      else if (*bfx<-0.5) *bfx=-0.5;
      if (result[*bx][*by+1]>=result[*bx][*by]) *bfy=0.5;
      else if (result[*bx][*by-1]>=result[*bx][*by]) *bfy=-0.5;
      else *bfy=(result[*bx][*by+1]-result[*bx][*by-1])/(4*result[*bx][*by]-2*result[*bx][*by+1]-2*result[*bx][*by-1]);
      if (*bfy>0.5) *bfy=0.5;
      else if (*bfy<-0.5) *bfy=-0.5;
   }
   if (xfirst) {
      ran[x][y]=-9999;
      return;
   }
   else if (*first) ran[x][y]=-9999;
   else ran[x][y]=*bm;
   if (xbx==-rad && xby==-rad) photsearch(x-1,y-1,rad-1,rad-1,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xbx==-rad && xby==rad) photsearch(x-1,y+1,rad-1,1-rad,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xbx==rad && xby==rad) photsearch(x+1,y+1,1-rad,1-rad,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xbx==rad && xby==-rad) photsearch(x+1,y-1,1-rad,rad-1,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xbx==-rad) photsearch(x-1,y,rad-1,xby,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xbx==rad) photsearch(x+1,y,1-rad,xby,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xby==-rad) photsearch(x,y-1,xbx,rad-1,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   else if (xby==rad) photsearch(x,y+1,xbx,1-rad,first,sx,sy,bx,by,bfx,bfy,bs,bss,bsky,bm,bc,bsh,finding);
   if (*first && *bm<-9998) ran[x][y]=-9999;
   else ran[x][y]=*bm;
}

void avgstar(int i,stype *s) {
   real d;
   stars[i].s=(stars[i].s+s->s)*0.5;
   stars[i].ss=sqrt((stars[i].ss*stars[i].ss+s->ss*s->ss)*0.5);
   stars[i].sky=(s->sky+stars[i].sky)*0.5;
   d=(((s->x+stars[i].x)%2)*sub[fld]+s->fx+s->ffx+stars[i].fx+stars[i].ffx)*0.5;
   stars[i].x=(s->x+stars[i].x)/2;
   stars[i].fx=(int)(d+100.5)-100;
   stars[i].ffx=d-stars[i].fx;
   while (stars[i].fx>sub[fld]/2) {
      stars[i].x++;
      stars[i].fx-=sub[fld];
   }
   while (stars[i].fx<-(sub[fld]/2)) {
      stars[i].x--;
      stars[i].fx+=sub[fld];
   }
   d=(((s->y+stars[i].y)%2)*sub[fld]+s->fy+s->ffy+stars[i].fy+stars[i].ffy)*0.5;
   stars[i].y=(s->y+stars[i].y)/2;
   stars[i].fy=(int)(d+100.5)-100;
   stars[i].ffy=d-stars[i].fy;
   while (stars[i].fy>sub[fld]/2) {
      stars[i].y++;
      stars[i].fy-=sub[fld];
   }
   while (stars[i].fy<-(sub[fld]/2)) {
      stars[i].y--;
      stars[i].fy+=sub[fld];
   }
}

int getclass(int i,double*s,double*ss,double*chi) {
   double st=0,cr=0,fl=0,np=0,a=0,b=0,c=0,d=0;
   int x1,y1;

   np=0;
   calc1psf(rphot[fld],stars[i].x,stars[i].y,stars[i].fx,stars[i].fy,stars[i].ffx,stars[i].ffy);
   for (x1=stars[i].x-rphot[fld];x1<=stars[i].x+rphot[fld];x1++) for (y1=stars[i].y+circ[0][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1<=stars[i].y+circ[1][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      a=res[x1][y1]-psf1[x1-stars[i].x][y1-stars[i].y]*stars[i].s;
      st+=a*a/noise[x1][y1]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
      np+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
   }
   if (np>0) st=sqrt(st/np);
   else st=-1;
   a=b=np=0;
   for (x1=stars[i].x-rphot[fld];x1<=stars[i].x+rphot[fld];x1++) for (y1=stars[i].y+circ[0][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1<=stars[i].y+circ[1][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1++) if ((x1!=stars[i].x || y1!=stars[i].y) && x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      cr+=res[x1][y1]*res[x1][y1]/noise[x1][y1]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
      np+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
   }
   if (np>0) cr=sqrt(cr/np);
   else cr=-1;
   a=b=c=d=np=0;
   for (x1=stars[i].x-rphot[fld];x1<=stars[i].x+rphot[fld];x1++) for (y1=stars[i].y+circ[0][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1<=stars[i].y+circ[1][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      a+=res[x1][y1]*res[x1][y1]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y]/noise[x1][y1];
      b+=res[x1][y1]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y]/noise[x1][y1];
      d+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y]/noise[x1][y1];
      c+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y]/sqrt(noise[x1][y1]);
      np+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
   }
   if (d>0 && np>0) fl=sqrt((a-b*b/d)/np);
   else fl=-1;
   if (cr>=0 && cr<st && stars[i].sh>0 && res[stars[i].x][stars[i].y]>=3.*sqrt(noise[stars[i].x][stars[i].y]+res[stars[i].x][stars[i].y]/GAIN)) {
      if (s) *s=res[stars[i].x][stars[i].y];
      if (ss) *ss=sqrt(noise[stars[i].x][stars[i].y]+res[stars[i].x][stars[i].y]/GAIN);
      if (chi) *chi=cr;
      return 4;
   }
   if (fl>=0 && fl<st && stars[i].sh<0) {
      if (s) *s=b/d*np;
      if (ss) *ss=sqrt(c)/d;
      if (chi) *chi=fl;
      return 5;
   }
   return 0;
}

void phot(int x,int y,int mark,int iter,int nit) {
   int sx,sy,bx,by,first=1,cx,cy,fx,fy;
   real bm=-100,bs,bss,bsky,bc,bsh,bfx,bfy;
   int st;
   stype old;

   if (mark<0) {
      centroid(x,y,&cx,&cy,&fx,&fy);
      photsearch(cx,cy,fx,fy,&first,&sx,&sy,&bx,&by,&bfx,&bfy,&bs,&bss,&bsky,&bm,&bc,&bsh,1);
   }
   else {
      ran[x][y]=-10000;
      if (y>0) ran[x][y-1]=-10000;
      if (y<799) ran[x][y+1]=-10000;
      if (x>0) {
	 ran[x-1][y]=-10000;
	 if (y>0) ran[x-1][y-1]=-10000;
	 if (y<799) ran[x-1][y+1]=-10000;
      }
      if (x<799) {
	 ran[x+1][y]=-10000;
	 if (y>0) ran[x+1][y-1]=-10000;
	 if (y<799) ran[x+1][y+1]=-10000;
      }
      memcpy(&old,stars+mark,sizeof(stype));
      photsearch(old.x,old.y,old.fx,old.fy,&first,&sx,&sy,&bx,&by,&bfx,&bfy,&bs,&bss,&bsky,&bm,&bc,&bsh,0);
   }
   if (first) bs=bss=0;
   if (bs<=0) bs=0;
   if ((bs>=bss*fsig && bs>0 && (!first)) || mark>=0) {
      if (mark<0) {
	 if (Nstars>=160000) {
	    printf("****Too many stars\n");
	    exit(-1);
	 }
	 st=Nstars;
	 Nstars++;
	 stars[st].type=1+iter;
      }
      else {
	 st=mark;
	 if (first || bs<=0) {
	    if (old.s!=0) mark1pos(old.x,old.y);
	 }
	 else if (sx!=old.x || sy!=old.y) mark2pos(sx,sy,old.x,old.y);
	 else if (abs(bx-old.fx)>1 || abs(by-old.fy)>1 || (fabs(bs-old.s)>1 && fabs(bs-old.s)>(bs+old.s)*0.0005)) mark1pos(sx,sy);
      }
      stars[st].flag=0;
      if (first || bs<=0) {
	 stars[st].s=0;
	 stars[st].ss=0;
	 stars[st].sky=0;
	 stars[st].chi=1000.;
	 stars[st].sh=0;
      }
      else {
	 stars[st].x=sx;
	 stars[st].y=sy;
	 stars[st].fx=bx;
	 stars[st].fy=by;
	 stars[st].ffx=bfx;
	 stars[st].ffy=bfy;
	 stars[st].s=bs;
	 stars[st].ss=bss;
	 stars[st].sky=bsky;
	 stars[st].chi=sqrt(bc);
	 stars[st].sh=bsh;
	 if (nit>MaxIT/2 && nit%3==MaxIT%3) avgstar(st,&old);
	 addindx(stars[st].x,stars[st].y,st);
      }
   }
}

void imgadd(int i) {
   int x1,y1;
   double q;

   calc1psf(rpsf[fld],stars[i].x,stars[i].y,stars[i].fx,stars[i].fy,stars[i].ffx,stars[i].ffy);
   for (x1=stars[i].x-rpsf[fld];x1<=stars[i].x+rpsf[fld];x1++) for (y1=stars[i].y-rpsf[fld];y1<=stars[i].y+rpsf[fld];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      res[x1][y1]+=(q=stars[i].s*(psf1[x1-stars[i].x][y1-stars[i].y]+ppsf[fld][stars[i].y/100][stars[i].x/100][stars[i].fx][stars[i].fy]));
      noise[x1][y1]-=dnoise[fld]*q*q;
      if (noise[x1][y1]<0) {
	 noise[x1][y1]=RN*RN;
	 if (res[x1][y1]>0) noise[x1][y1]+=res[x1][y1]/GAIN;
	 if (sky[x1][y1]>0) noise[x1][y1]+=sky[x1][y1]/GAIN;
      }
   }
}

void imgsub(int i) {
   int x1,y1;
   double q;

   calc1psf(rpsf[fld],stars[i].x,stars[i].y,stars[i].fx,stars[i].fy,stars[i].ffx,stars[i].ffy);
   for (x1=stars[i].x-rpsf[fld];x1<=stars[i].x+rpsf[fld];x1++) for (y1=stars[i].y-rpsf[fld];y1<=stars[i].y+rpsf[fld];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      res[x1][y1]-=(q=stars[i].s*(psf1[x1-stars[i].x][y1-stars[i].y]+ppsf[fld][stars[i].y/100][stars[i].x/100][stars[i].fx][stars[i].fy]));
      noise[x1][y1]+=dnoise[fld]*q*q;
   }
}

int starOK(int i) {
   int nbad=0,ntot=0,x1,y1;
   if (data[stars[i].x][stars[i].y]<=DMIN || data[stars[i].x][stars[i].y]>=DMAX) return 0;
   if (stars[i].x>797 || stars[i].y>797 || stars[i].x<52 || stars[i].y<52) return 0;
   for (x1=stars[i].x-rphot[fld];x1<=stars[i].x+rphot[fld];x1++) for (y1=stars[i].y+circ[0][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1<=stars[i].y+circ[1][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) {
      ntot++;
      if (data[x1][y1]<=DMIN || data[x1][y1]>=DMAX) nbad++;
   }
   if (nbad<ntot/3) return 1;
   return 0;
}

/*Quicksort from NR; 0-index by AED*/
#define SWAP(a,b) {memcpy(&temp,a,ssize);memcpy(a,b,ssize);memcpy(b,&temp,ssize);}
#define M 7
#define NSTACK 50

void sortstars() {
   int i,ir,j,k,l=0;
   int jstack=0,*istack,ssize;
   stype temp,a;

   ir=Nstars-1;
   ssize=sizeof(stype);
   if ((istack=(int*)calloc(sizeof(int),NSTACK))==NULL) merr();
   for (;;) {
      if (ir-l<M) {
	 for (j=l+1;j<=ir;j++) {
	    memcpy(&a,stars+j,ssize);
	    for (i=j-1;i>=l;i--) {
	       if (stars[i].s>=a.s) break;
	       memcpy(stars+i+1,stars+i,ssize);
	    }
	    memcpy(stars+i+1,&a,ssize);
	 }
	 if (jstack==0) break;
	 ir=istack[jstack--];
	 l=istack[jstack--];
      } else {
	 k=(l+ir)>>1;
	 SWAP(stars+k,stars+l+1)
	 if (stars[l+1].s<stars[ir].s) SWAP(stars+l+1,stars+ir)
	 if (stars[l].s<stars[ir].s) SWAP(stars+l,stars+ir)
	 if (stars[l+1].s<stars[l].s) SWAP(stars+l+1,stars+l)
	 i=l+1;
	 j=ir;
	 memcpy(&a,stars+l,ssize);
	 for (;;) {
	    do i++; while (stars[i].s>a.s);
	    do j--; while (stars[j].s<a.s);
	    if (j<i) break;
	    SWAP(stars+i,stars+j);
	 }
	 memcpy(stars+l,stars+j,ssize);
	 memcpy(stars+j,&a,ssize);
	 jstack+=2;
	 if (jstack > NSTACK) {
	    printf("****NSTACK too small in sortstars.");
	    exit(-1);
	 }
	 if (ir-i+1>=j-l) {
	    istack[jstack]=ir;
	    istack[jstack-1]=i;
	    ir=j-1;
	 } else {
	    istack[jstack]=j-1;
	    istack[jstack-1]=l;
	    l=i;
	 }
      }
   }
   free(istack);
}
#undef M
#undef NSTACK
#undef SWAP

int fixpsf(int Np,int*plist,int it) {
   int x,y,xx,yy,x3,y3,i,N,rad;
   float *list,*sig,av,sd,z,**off,**avpsf,toff=0,toffd=0;

   if (Np>4*rpsf[fld]) list=(float*)calloc(Np,sizeof(float));
   else list=(float*)calloc(4*rpsf[fld],sizeof(float));
   sig=(float*)calloc(Np,sizeof(float));
   off=(float**)calloc(2*rpsf[fld]+1,sizeof(float*));
   avpsf=(float**)calloc(2*rpsf[fld]+1,sizeof(float*));
   if (!list || !sig || !off || !avpsf) merr();
   off+=rpsf[fld];
   avpsf+=rpsf[fld];
   rad=(sub[fld]+2)/2;
   for (x=-rpsf[fld];x<=rpsf[fld];x++) {
      off[x]=(float*)calloc(2*rpsf[fld]+1,sizeof(float));
      avpsf[x]=(float*)calloc(2*rpsf[fld]+1,sizeof(float));
      if (!off[x] || !avpsf[x]) merr();
      off[x]+=rpsf[fld];
      avpsf[x]+=rpsf[fld];
      for (y=-rpsf[fld];y<=rpsf[fld];y++) {
	 N=0;
	 for (i=0;i<Np;i++) if (stars[plist[i]].s>0 && stars[plist[i]].s>stars[plist[i]].ss*fsig) {
	    int ii;
	    ii=plist[i];
	    xx=x+stars[ii].x;
	    yy=y+stars[ii].y;
	    sig[N]=stars[ii].s/sqrt(noise[xx][yy]);
	    list[N++]=res[xx][yy]/stars[ii].s;
	 }
	 av=sd=0;
	 xx=1;
	 while (xx) {
	    av=sd=z=0;
	    for (i=0;i<N;i++) {
	       av+=list[i]*sig[i];
	       z+=sig[i];
	    }
	    xx=0;
	    if (z>0) {
	       av/=z;
	       for (i=0;i<N;i++) sd+=(list[i]-av)*(list[i]-av)*sig[i];
	       sd=sqrt(sd/z)*5;
	       for (i=0;i<N;)
		  if (fabs(list[i]-av)>sd) {
		     list[i]=list[--N];
		     sig[i]=sig[N];
		     xx=1;
		  }
		  else i++;
	    }
	    else av=0;
	 }
	 off[x][y]=av;
	 avpsf[x][y]=0;
	 for (x3=0;x3<8;x3++) for (y3=0;y3<8;y3++) for (xx=-rad;xx<=rad;xx++) for (yy=-rad;yy<=rad;yy++) avpsf[x][y]+=psf[fld][x3][y3][xx][yy][x][y]+ppsf[fld][x3][y3][xx][yy];
	 avpsf[x][y]/=64.*(2*rad+1)*(2*rad+1);
	 toff+=avpsf[x][y];
	 toffd+=avpsf[x][y]+off[x][y];
      }
   }
   toff/=toffd;
   for (x=-rpsf[fld];x<=rpsf[fld];x++) for (y=-rpsf[fld];y<=rpsf[fld];y++) {
      off[x][y]=toff*off[x][y]+(toff-1)*avpsf[x][y];
      if (it>4) off[x][y]/=1+it*0.1;
      poff[x][y]+=off[x][y];
   }
   for (i=0;i<Nstars;i++) imgadd(i);
   for (x3=0;x3<8;x3++) for (y3=0;y3<8;y3++) for (xx=-rad;xx<=rad;xx++) for (yy=-rad;yy<=rad;yy++) for (x=-rpsf[fld];x<=rpsf[fld];x++) for (y=-rpsf[fld];y<=rpsf[fld];y++) psf[fld][x3][y3][xx][yy][x][y]+=off[x][y];
   getspsf(fld);
   for (i=0;i<Nstars;i++) imgsub(i);
   free(list);
   free(sig);
   i=0;
   for (x=-rpsf[fld];x<=rpsf[fld] && !i;x++) for (y=-rpsf[fld];y<=rpsf[fld] && !i;y++) if (fabs(off[x][y])>=0.001+0.0001*it) i=1;
   for (x=-rpsf[fld];x<=rpsf[fld];x++) free(off[x]-rpsf[fld]);
   free(off-rpsf[fld]);
   return i;
}

int goodstar(int i) {
   int j,r,x,y;
   r=rpsf[fld]+rphot[fld];
   if (getclass(i,NULL,NULL,NULL)) return 0;
   if (stars[i].sh>0.2 || stars[i].sh<-0.2 || stars[i].chi>2.5) return 0;
   if (stars[i].s<stars[i].ss*fsig || stars[i].s<1 || stars[i].x<XMIN2 || stars[i].x-rpsf[fld]<XMIN || stars[i].x>=XMAX2 || stars[i].x+rpsf[fld]>=XMAX || stars[i].y<YMIN2 || stars[i].y-rpsf[fld]<YMIN || stars[i].y>=YMAX2 || stars[i].y+rpsf[fld]>=YMAX) return 0;
   for (x=-rpsf[fld];x<=rpsf[fld];x++) for (y=-rpsf[fld];y<=rpsf[fld];y++) if (data[stars[i].x+x][stars[i].y+y]>=DMAX*0.855 || data[stars[i].x+x][stars[i].y+y]<=DMIN) return 0;
   for (j=0;j<i;j++) if (abs(stars[i].x-stars[j].x)<=r && abs(stars[i].y-stars[j].y)<=r) return 0;
   for (j=i+1;j<Nstars && stars[j].s*2>=stars[i].s;j++) if (abs(stars[i].x-stars[j].x)<=rpsf[fld] && abs(stars[i].y-stars[j].y)<=rpsf[fld]) return 0;
   return 1;
}

void calcpsf(void) {
#define minNpsf 50
#define goodNpsf 150
#define maxNpsf 300
   int *plist,*tlist,Np=0,Nt=0,i,it=0,x,y,cont=1;

   plist=(int*)calloc(Nstars,sizeof(int));
   tlist=(int*)calloc(Nstars,sizeof(int));
   sortstars();
   for (x=0;x<800;x++) for (y=0;y<800;y++) indx[x][y][0]=0;
   for (i=0;i<Nstars;i++) {
      addindx((int)stars[i].x,(int)stars[i].y,i);
      stars[i].flag=0;
   }
   for (i=0;i<Nstars;i++) {
      imgadd(i);
      if (((stars[i].s>=stars[i].ss*50 && Np<maxNpsf) || (stars[i].s>=stars[i].ss*10 && Np<goodNpsf) || Np<minNpsf) && goodstar(i)) {
	 plist[Np++]=i;
      }
      imgsub(i);
   }
   while (Np>0 && cont) {
      double mn=0,sd=0;

      cont=0;
      for (i=0;i<Np;i++) mn+=stars[plist[i]].sh;
      mn/=Np;
      for (i=0;i<Np;i++) sd+=(stars[plist[i]].sh-mn)*(stars[plist[i]].sh-mn);
      sd=2.00*sqrt(sd/Np);
      for (i=0;i<Np;i++) if (fabs(stars[plist[i]].sh-mn)>sd) {
	 Np--;
	 plist[i]=plist[Np];
	 i--;
	 cont=1;
      }
   }
   for (i=0;i<Np;i++) mark1pos(stars[plist[i]].x,stars[plist[i]].y);
   for (i=0;i<Nstars;i++) if (stars[i].flag) tlist[Nt++]=i;
   printf("%d PSF stars; %d total stars\n",Np,Nt);
   while (fixpsf(Np,plist,++it)) {
      for (i=0;i<Nt;i++) {
	 imgadd(tlist[i]);
	 if (stars[tlist[i]].s>0) subindx(stars[tlist[i]].x,stars[tlist[i]].y,tlist[i]);
	 phot(stars[tlist[i]].x,stars[tlist[i]].y,tlist[i],0,0);
	 imgsub(tlist[i]);
      }
   }
   free(plist);
   free(tlist);
   if (it==20) printf("Poorly converging PSF\n");
   printf("Adjusted central pixel of PSFs by %f\n",poff[0][0]);
}

int clean(int it) {
   int i,ii,j,b=0,ct=0,xx,yy,rad,x3,y3;
   real x,y,r,br;

   for (i=1;i<Nstars;i++) if (stars[i].x>=0) {
      rad=(int)(rmin[fld]+0.999);
      br=rmin[fld];
      for (xx=-rad;xx<=rad;xx++) if ((x3=stars[i].x+xx)>=XMIN && x3<XMAX) for (yy=-rad;yy<=rad;yy++) if ((xx || yy) && (y3=stars[i].y+yy)>=YMIN && y3<YMAX) for (ii=1;ii<=indx[x3][y3][0];ii++) {
	 j=indx[x3][y3][ii];
	 if (j<i && (x=fabs(stars[i].x+stars[i].fx/(float)sub[fld]-stars[j].x-stars[j].fx/(float)sub[fld]))<=rmin[fld] && (y=fabs(stars[i].y+stars[i].fy/(float)sub[fld]-stars[j].y-stars[j].fy/(float)sub[fld]))<=rmin[fld]) {
	    r=sqrt(x*x+y*y);
	    if (r<br) {
	       br=r;
	       b=j;
	    }
	 }
      }
      if (br<rmin[fld]) {
	 if (stars[i].s>0) subindx(stars[i].x,stars[i].y,i);
	 if (stars[b].s>0) subindx(stars[b].x,stars[b].y,b);
	 imgadd(i);
	 imgadd(b);
	 mark2pos(stars[i].x,stars[i].y,stars[b].x,stars[b].y);
	 avgstar(b,stars+i);
	 stars[b].s*=2;
	 stars[b].ss*=1.4;
	 if (it==1) {
	    if (stars[b].type==2 && stars[i].type==1) stars[b].type=1;
	 }
	 else stars[b].type=3;
	 phot(stars[b].x,stars[b].y,b,0,0);
	 imgsub(b);
	 stars[i].x=-1;
      }
   }
   ii=0;
   for (i=0;i<Nstars;i++) if (stars[i].x>=0 && stars[i].s>0) {
      if (ii<i) {
	 memcpy(stars+ii,stars+i,sizeof(stype));
	 repindx(stars[ii].x,stars[ii].y,i,ii);
      }
      ii++;
   }
   ct=Nstars-ii;
   Nstars=ii;
   if (ct) printf("%d stars eliminated\n",ct);
   return ct;
}

int solve(int it) {
   int i,reit=0,x,y;
   static int newpsf=0;

   if (it==1) newpsf=0;
   for (x=0;x<800;x++) for (y=0;y<800;y++) indx[x][y][0]=0;
   for (i=0;i<Nstars;i++) addindx(stars[i].x,stars[i].y,i);
   for (i=0;i<Nstars;i++) if (stars[i].flag) {
      reit++;
      imgadd(i);
      subindx(stars[i].x,stars[i].y,i);
      phot(stars[i].x,stars[i].y,i,0,it);
      imgsub(i);
   }
   if (clean(it) || it<2) return 2;
   if (!newpsf && !(OPTS&4)) {
      calcpsf();
      newpsf=1;
      for (i=0;i<Nstars;i++) stars[i].flag=1;
      return 2;
   }
   if (reit) return 1;
   return 0;
}

void find(void) {
   int xx,x,y,prev,nf;

   prev=Nstars;
   memcpy(ran,cran,sizeof(float[800][800]));
   for (x=0;x<800;x++) for (y=0;y<800;y++) indx[x][y][0]=0;
   for (x=XMIN+1;x<XMAX-1;x++) for (y=YMIN+1;y<YMAX-1;y++) if (res[x][y]>=sky[x][y]*10 && res[x][y]>res[x-1][y] && res[x][y]>res[x-1][y-1] && res[x][y]>res[x][y-1] && res[x][y]>res[x+1][y-1] && res[x][y]>res[x+1][y] && res[x][y]>res[x+1][y+1] && res[x][y]>res[x][y+1] && res[x][y]>res[x-1][y+1]) phot(x,y,-1,0,0);
   for (x=XMIN+1;x<XMAX-1;x++) for (y=YMIN+1;y<YMAX-1;y++) if (res[x][y]>=sky[x][y]*3 && res[x][y]<sky[x][y]*10 && res[x][y]>res[x-1][y] && res[x][y]>res[x-1][y-1] && res[x][y]>res[x][y-1] && res[x][y]>res[x+1][y-1] && res[x][y]>res[x+1][y] && res[x][y]>res[x+1][y+1] && res[x][y]>res[x][y+1] && res[x][y]>res[x-1][y+1]) phot(x,y,-1,0,0);
   for (x=XMIN+1;x<XMAX-1;x++) for (y=YMIN+1;y<YMAX-1;y++) if (res[x][y]>=sky[x][y] && res[x][y]<sky[x][y]*3 && res[x][y]>res[x-1][y] && res[x][y]>res[x-1][y-1] && res[x][y]>res[x][y-1] && res[x][y]>res[x+1][y-1] && res[x][y]>res[x+1][y] && res[x][y]>res[x+1][y+1] && res[x][y]>res[x][y+1] && res[x][y]>res[x-1][y+1]) phot(x,y,-1,0,0);
   for (x=XMIN+1;x<XMAX-1;x++) for (y=YMIN+1;y<YMAX-1;y++) if (res[x][y]<sky[x][y] && res[x][y]>res[x-1][y] && res[x][y]>res[x-1][y-1] && res[x][y]>res[x][y-1] && res[x][y]>res[x+1][y-1] && res[x][y]>res[x+1][y] && res[x][y]>res[x+1][y+1] && res[x][y]>res[x][y+1] && res[x][y]>res[x-1][y+1]) phot(x,y,-1,0,0);
   nf=Nstars;
   for (x=prev;x<nf;x++) imgsub(x);
   memcpy(ran,cran,sizeof(float[800][800]));
   for (xx=prev;xx<nf;xx++) for (x=stars[xx].x-rpsf[fld];x<=stars[xx].x+rpsf[fld];x++) for (y=stars[xx].y-rpsf[fld];y<stars[xx].y+rpsf[fld];y++) if (res[x][y]>res[x-1][y] && res[x][y]>res[x-1][y-1] && res[x][y]>res[x][y-1] && res[x][y]>res[x+1][y-1] && res[x][y]>res[x+1][y] && res[x][y]>res[x+1][y+1] && res[x][y]>res[x][y+1] && res[x][y]>res[x-1][y+1]) phot(x,y,-1,1,0);
   for (x=prev;x<nf;x++) imgadd(x);
   for (x=nf;x<Nstars;x++) if (res[stars[x].x][stars[x].y]<res[stars[x].x+1][stars[x].y] || res[stars[x].x][stars[x].y]<res[stars[x].x-1][stars[x].y] || res[stars[x].x][stars[x].y]<res[stars[x].x][stars[x].y+1] || res[stars[x].x][stars[x].y]<res[stars[x].x][stars[x].y-1] || res[stars[x].x][stars[x].y]<res[stars[x].x+1][stars[x].y+1] || res[stars[x].x][stars[x].y]<res[stars[x].x+1][stars[x].y-1] || res[stars[x].x][stars[x].y]<res[stars[x].x-1][stars[x].y+1] || res[stars[x].x][stars[x].y]<res[stars[x].x-1][stars[x].y-1]) {
      Nstars--;
      subindx(stars[x].x,stars[x].y,x);
      memcpy(stars+x,stars+Nstars,sizeof(stype));
      if (x<Nstars) repindx(stars[x].x,stars[x].y,Nstars,x);
      x--;
   }
   for (x=prev;x<Nstars;x++) imgsub(x);
   printf("%d stars found\n",Nstars-prev);
   for (x=prev;x<Nstars;x++) {
      y=stars[x].flag;
      mark1pos(stars[x].x,stars[x].y);
      stars[x].flag=y;
   }
   clean(1);
}

double fitstar(int i,double ffx,double ffy) {
   double cs,s,ss,c,cc,m,sh,np,A,a,b,ssky;
   int x1,y1;

   calc1psf(rphot[fld],stars[i].x,stars[i].y,stars[i].fx,stars[i].fy,ffx,ffy);
   ssky=getsky(stars[i].x,stars[i].y);
   cs=s=ss=c=cc=0;
   for (x1=stars[i].x-rphot[fld];x1<=stars[i].x+rphot[fld];x1++) for (y1=stars[i].y+circ[0][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1<=stars[i].y+circ[1][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      if (OPTS&1) {
	 cc+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
	 c+=psf1[x1-stars[i].x][y1-stars[i].y]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
	 s+=(res[x1][y1]-ssky)*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
	 ss+=noise[x1][y1]*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
      }
      else {
	 a=psf1[x1-stars[i].x][y1-stars[i].y];
	 b=a*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y]/noise[x1][y1];
	 cc+=b;
	 if (a>0) {
	    s+=(res[x1][y1]-ssky)*b;
	    a*=b;
	    c+=a;
	    cs+=a;
	    ss+=a*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
	 }
	 else cs+=a*b;
      }
   }
   if (ss<0 || c==0 || cs==0) return 0;
   if (OPTS&1) cs=c;
   cs/=cc;
   s/=c;
   ss=sqrt(ss)/c;
   c=sh=0;
   A=np=0;
   for (x1=stars[i].x-rphot[fld];x1<=stars[i].x+rphot[fld];x1++) for (y1=stars[i].y+circ[0][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1<=stars[i].y+circ[1][fld][stars[i].fx][stars[i].fy][x1-stars[i].x];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX && data[x1][y1]>DMIN && data[x1][y1]<DMAX) {
      a=psf1[x1-stars[i].x][y1-stars[i].y]*s;
      b=1/(noise[x1][y1]+dnoise[fld]*a*a);
      a=(res[x1][y1]-ssky)-a;
      c+=a*a*b*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
      np+=cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
      sh+=a/sqrt(noise[x1][y1])*(psf1[x1-stars[i].x][y1-stars[i].y]-cs)*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
      A+=(psf1[x1-stars[i].x][y1-stars[i].y]-cs)*(psf1[x1-stars[i].x][y1-stars[i].y]-cs)*s/sqrt(noise[x1][y1])*cwt[fld][stars[i].fx][stars[i].fy][x1-stars[i].x][y1-stars[i].y];
   }
   if (A>0) sh/=A;
   else sh=0;
   c=c/np;
   m=s/(ss*sqrt(c+.25));
   if (c>1) ss*=sqrt(c);
   stars[i].s=s;
   stars[i].ss=ss;
   stars[i].chi=sqrt(c);
   stars[i].sh=sh;
   stars[i].sky=ssky;
   return m;
}

void finalphot(int i) {
   int j;
   double dp=1;

   imgadd(i);
   while (dp>0.005) {
      int shrink=1;
      double t0,t,t1;

      t0=fitstar(i,stars[i].ffx-dp,stars[i].ffy);
      t=fitstar(i,stars[i].ffx,stars[i].ffy);
      t1=fitstar(i,stars[i].ffx+dp,stars[i].ffy);
      if (t0>t) {
	 stars[i].ffx-=dp;
	 shrink=0;
      }
      else if (t1>t) {
	 stars[i].ffx+=dp;
	 shrink=0;
      }
      else if (t0<t || t1<t) {
	 t=(t1-t0)/(4*t-2*t0-2*t1);
	 if (t<-.5) t=-.5;
	 else if (t>.5) t=.5;
	 stars[i].ffx+=t*dp;
      }
      if (stars[i].ffx>.5) {
	 stars[i].fx++;
	 stars[i].ffx-=1;
	 if (stars[i].fx>sub[fld]/2) {
	    stars[i].fx-=sub[fld];
	    stars[i].x++;
	 }
      }
      if (stars[i].ffx<-.5) {
	 stars[i].fx--;
	 stars[i].ffx+=1;
	 if (stars[i].fx<-(sub[fld]/2)) {
	    stars[i].fx+=sub[fld];
	    stars[i].x--;
	 }
      }
      t0=fitstar(i,stars[i].ffx,stars[i].ffy-dp);
      t=fitstar(i,stars[i].ffx,stars[i].ffy);
      t1=fitstar(i,stars[i].ffx,stars[i].ffy+dp);
      if (t0>t) {
	 stars[i].ffy-=dp;
	 shrink=0;
      }
      else if (t1>t) {
	 stars[i].ffy+=dp;
	 shrink=0;
      }
      else if (t0<t || t1<t) {
	 t=(t1-t0)/(4*t-2*t0-2*t1);
	 if (t<-.5) t=-.5;
	 else if (t>.5) t=.5;
	 stars[i].ffy+=t*dp;
      }
      if (stars[i].ffy>.5) {
	 stars[i].fy++;
	 stars[i].ffy-=1;
	 if (stars[i].fy>sub[fld]/2) {
	    stars[i].fy-=sub[fld];
	    stars[i].y++;
	 }
      }
      if (stars[i].ffy<-.5) {
	 stars[i].fy--;
	 stars[i].ffy+=1;
	 if (stars[i].fy<-(sub[fld]/2)) {
	    stars[i].fy+=sub[fld];
	    stars[i].y--;
	 }
      }
      if (shrink) dp*=.5;
      else dp*=.9;
   }
   fitstar(i,stars[i].ffx,stars[i].ffy);
   j=getclass(i,&stars[i].s,&stars[i].ss,&stars[i].chi);
   if (j>0) stars[i].type=j;
   imgsub(i);
}

real coldCTE(real x,real y,real ct,real bg) {
   real mult,XCTE,YCTE,lct,lbg,yr;
   double bgcorr[NFILTERS][2]={
      {0.000000e-06,2.869069e-04}, //F170W
      {8.540038e-07,1.010904e-04}, //F300W
      {8.540038e-07,1.171397e-04}, //F336W
      {8.540038e-07,9.556493e-05}, //F380W
      {7.339984e-06,2.279944e-05}, //F410M
      {7.339984e-06,2.279944e-05}, //F439W
      {7.339984e-06,2.279944e-05}, //F450W
      {7.339984e-06,2.279944e-05}, //F467M
      {5.304520e-06,2.433564e-05}, //F547M
      {5.304520e-06,2.433564e-05}, //F555W
      {5.304520e-06,2.433564e-05}, //F569W
      {5.304520e-06,2.433564e-05}, //F606W
      {2.354923e-06,2.104143e-05}, //F622W
      {2.354923e-06,2.104143e-05}, //F675W
      {2.354923e-06,2.104143e-05}, //F702W
      {4.414889e-06,1.967949e-05}, //F785LP
      {4.414889e-06,1.967949e-05}, //F791W
      {4.414889e-06,1.967949e-05}, //F814W
      {4.414889e-06,1.967949e-05}, //F850LP
      {4.414889e-06,1.967949e-05}  //F1042M
   };
   double fcorr[NFILTERS][2][4]={
      {{0.000000,0.000000,0.000000,0.000000},{0.000000,0.000000,0.000000,0.000000}}, //F170W
      {{0.000000,0.000000,0.000000,0.000000},{0.000000,0.000000,0.000000,0.000000}}, //F300W
      {{-0.010696,-0.028429,-0.044806,-0.072908},{-0.017399,-0.029386,-0.051494,-0.056319}}, //F336W
      {{-0.085207,-0.099002,-0.110461,-0.133910},{-0.075338,-0.101240,-0.113912,-0.121847}}, //F380W
      {{-0.067418,-0.081213,-0.092672,-0.116121},{-0.057549,-0.083451,-0.096123,-0.104058}}, //F410M
      {{-0.014118,-0.027912,-0.039372,-0.062820},{-0.004248,-0.030150,-0.042822,-0.050758}}, //F439W
      {{0.003330,-0.010464,-0.021924,-0.045372},{0.013200,-0.012702,-0.025374,-0.033310}}, //F450W
      {{0.058822,0.045028,0.033568,0.010120},{0.068692,0.042790,0.030118,0.022182}}, //F467M
      {{0.021796,-0.012867,0.002091,-0.008525},{0.012300,0.010481,-0.007393,-0.010866}}, //F547M
      {{0.002878,-0.031786,-0.016828,-0.027443},{-0.006618,-0.008437,-0.026311,-0.029784}}, //F555W
      {{0.041852,0.007189,0.022147,0.011532},{0.032357,0.030538,0.012664,0.009190}}, //F569W
      {{0.023378,-0.011285,0.003673,-0.006943},{0.013882,0.012063,-0.005811,-0.009284}}, //F606W
      {{-0.027339,-0.028044,-0.021984,-0.036220},{0.003571,-0.021912,-0.027475,-0.010892}}, //F622W
      {{-0.031295,-0.031999,-0.025939,-0.040175},{-0.000384,-0.025867,-0.031430,-0.014847}}, //F675W
      {{-0.025641,-0.026346,-0.020286,-0.034522},{0.005269,-0.020214,-0.025777,-0.009194}}, //F702W
      {{-0.010789,-0.015340,-0.014320,-0.017659},{0.021863,-0.009720,-0.004956,-0.011427}}, //F785LP
      {{0.011348,0.006797,0.007817,0.004478},{0.044000,0.012417,0.017181,0.010710}}, //F791W
      {{-0.009464,-0.014015,-0.012996,-0.016334},{0.023187,-0.008395,-0.003631,-0.010103}}, //F814W
      {{-0.006654,-0.011205,-0.010185,-0.013524},{0.025998,-0.005585,-0.000821,-0.007292}}, //F850LP
      {{-0.005653,-0.010205,-0.009185,-0.012524},{0.026998,-0.004584,0.000179,-0.006292}} //F1042M
   };

   if (OPTS&32) return 1.;
   //Convert to electrons in original image level;
   mult=EXP0/EXP;
   if (GAIN!=7) mult*=14;
   else mult*=7;
   bg-=ct*bgcorr[filt][(fld==0)?0:1];
   bg*=mult;
   ct*=mult;
   //Soften background and signal;
   if (bg<0) bg=1.;
   else bg=sqrt(1.+bg*bg);
   if (ct<0) ct=1.;
   else ct=sqrt(1.+ct*ct);
   lct=log(ct)-7;
   lbg=log(bg)-1;
   yr=YEAR-2;
   YCTE=(0.245972+exp(-0.444895*lct))*(exp(-0.035551*bg)+0.460522*exp(-0.231556*lbg))*(0.069726+0.023982*yr-0.000138*yr*yr);
   XCTE=exp(-0.009569*lbg)*0.025875;
   return pow(10,0.4*(fcorr[filt][IGN][fld]+YCTE*x/800.+XCTE*y/800.));
}

real CTE(real x,real y,real ct,real bg) {
   real mult,XCTE,YCTE,lct,lbg,yr;
   double bgcorr[NFILTERS][2][2]={
      {{0.000000e-06,2.869069e-04},{0.000000e-06,2.292851e-04}}, //F170W
      {{8.540038e-07,1.010904e-04},{2.347928e-06,8.078760e-05}}, //F300W
      {{8.540038e-07,1.171397e-04},{2.347928e-06,1.000053e-04}}, //F336W
      {{8.540038e-07,9.556493e-05},{2.347928e-06,1.000053e-04}}, //F380W
      {{7.339984e-06,2.279944e-05},{0.000000e-06,5.432855e-05}}, //F410M
      {{7.339984e-06,2.279944e-05},{0.000000e-06,5.432855e-05}}, //F439W
      {{7.339984e-06,2.279944e-05},{0.000000e-06,5.432855e-05}}, //F450W
      {{7.339984e-06,2.279944e-05},{0.000000e-06,5.432855e-05}}, //F467M
      {{5.304520e-06,2.433564e-05},{5.497646e-06,8.393854e-05}}, //F547M
      {{5.304520e-06,2.433564e-05},{5.497646e-06,7.363844e-05}}, //F555W
      {{5.304520e-06,2.433564e-05},{5.497646e-06,7.331492e-05}}, //F569W
      {{5.304520e-06,2.433564e-05},{5.497646e-06,6.200388e-05}}, //F606W
      {{2.354923e-06,2.104143e-05},{1.270904e-06,1.109481e-04}}, //F622W
      {{2.354923e-06,2.104143e-05},{1.270904e-06,1.109481e-04}}, //F675W
      {{2.354923e-06,2.104143e-05},{1.270904e-06,7.000852e-05}}, //F702W
      {{4.414889e-06,1.967949e-05},{1.781239e-05,6.953957e-05}}, //F785LP
      {{4.414889e-06,1.967949e-05},{1.781239e-05,9.479239e-05}}, //F791W
      {{4.414889e-06,1.967949e-05},{1.781239e-05,9.479239e-05}}, //F814W
      {{4.414889e-06,1.967949e-05},{1.781239e-05,9.479239e-05}}, //F850LP
      {{4.414889e-06,1.967949e-05},{1.781239e-05,9.479239e-05}}  //F1042M
   };
   double gcorr[4]={3.719247e-02,-1.203985e-02,6.861146e-03,3.992702e-03};
   double fcorr[2][NFILTERS]={
      {0.000000e+00,0.000000e+00,-2.304644e-02,-9.948792e-02,-7.099982e-02,-1.567542e-02,-9.765530e-03,4.599253e-02,-3.731368e-03,-9.215336e-03,7.585554e-03,9.387174e-03,-1.170134e-03,-6.744189e-03,-1.713857e-02,4.278063e-03,3.849086e-02,1.230014e-02,1.362391e-02,1.583694e-02},
      {0.000000e+00,0.000000e+00,4.467948e-02,-5.021531e-02,-2.172721e-02,3.359719e-02,3.950708e-02,9.526514e-02,1.863842e-02,1.315445e-02,2.995534e-02,3.175696e-02,1.809453e-02,1.252047e-02,2.126090e-03,-2.218946e-02,1.202334e-02,-1.416738e-02,-1.284361e-02,-1.063058e-02}};

   if (OPTS&32) return 1.;
   if (TEMP==0 && filters[filt].color!='X' && filters[filt].color!='U') return coldCTE(x,y,ct,bg);
   //Convert to electrons in original image level;
   mult=EXP0/EXP;
   if (GAIN!=7) mult*=14;
   else mult*=7;
   bg-=ct*bgcorr[filt][TEMP][(fld==0)?0:1];
   bg*=mult;
   ct*=mult;
   //Soften background and signal;
   if (bg<0) bg=1.;
   else bg=sqrt(1.+bg*bg);
   if (ct<0) ct=1.;
   else ct=sqrt(1.+ct*ct);
   //CTE from Whitmore, Heyer, & Casertano 1999;
   /*
   YCTE=0.023*pow(10,-0.256*log10(bg/7.));
   XCTE=0.025;
   if (s<28000) {
      YCTE*=(1+0.245*(-0.0087*log10(s/28000.))*(YEAR-0.04)*365.25);
      XCTE*=(1+0.341*(-0.0020*log10(s/28000.))*(YEAR-0.04)*365.25);
   }
   return 1.+YCTE*x/800.+XCTE*y/800.;
   */
   //HSTphot CTE determination;
   lct=log(ct)-7;
   lbg=log(bg)-1;
   yr=YEAR-2;
   if (TEMP==0) YCTE=0.018009+(0.087972+exp(-0.506635*lct))*exp(-0.042368*bg-0.035149*lbg)*(0.096532+0.041442*yr);
   else YCTE=0.102822+0.027988*exp(-0.958904*lct);
   if (TEMP==0) XCTE=exp(-0.195966*lct)*exp(-0.126435*lbg)*(0.023881+0.002327*yr);
   else XCTE=0.;
   return pow(10,0.4*(fcorr[TEMP][filt]+gcorr[fld]+YCTE*x/800.+XCTE*y/800.));
}

void outpt(FILE*of) {
   int i,Nelim[2]={0,0};
   stype t;
   real s,ss;

   sortstars();
   for (i=0;i<Nstars;i++) if (stars[i].s>0 && stars[i].s>=fsig*stars[i].ss && stars[i].x>=XMIN2 && stars[i].x<XMAX2 && stars[i].y>=YMIN2 && stars[i].y<YMAX2 && starOK(i)) {
      finalphot(i);
      if (stars[i].s>0 && stars[i].s>=fsig*stars[i].ss && stars[i].x>=XMIN2 && stars[i].x<XMAX2 && stars[i].y>=YMIN2 && stars[i].y<YMAX2 && starOK(i)) {
	 memcpy(&t,stars+i,sizeof(stype));
	 s=(t.s)*corr[fld];
	 ss=t.ss*corr[fld];
	 ss=ss/s*1.0857362;
	 t.sky+=sky[t.x][t.y];
	 s=-2.5*log10(s/EXP*CTE(t.x,t.y,t.s,sky[t.x][t.y]))+filters[filt].zero;
	 fprintf(of,"%d %6.2f %6.2f ",fld,t.y+.5+(t.fy+t.ffy)/(real)sub[fld],t.x+.5+(t.fx+t.ffx)/(real)sub[fld]);
	 if (t.s<999999.5) fprintf(of,"%8.1f",t.s);
	 else fprintf(of,"%8.2e",t.s);
	 fprintf(of," %6.2f %6.3f %6.3f %8.4f %7.1f %6.3f %d\n",t.sky,s,ss,t.chi,t.s/t.ss,t.sh,t.type);
	 Nelim[0]++;
      }
      else Nelim[1]++;
   }
   else Nelim[1]++;
   printf("%d stars written; %d stars deleted\n",Nelim[0],Nelim[1]);
   fflush(of);
}

int ferr(FILE *f,char *str) {
   printf("****%s\n",str);
   fclose(f);
   return 1;
}

int main(int argc,char**argv) {
   FILE *f,*of,*fpsf;
   char str[82];
   int MODE=0,line=0,tmp,it,x,y,HPP=0;

   if (argc!=5 && argc!=10) {
      printf("****Usage: %s <fits base> <output> <detection sig> <opts>\n <fld> <xmin> <xmax> <ymin> <ymax>\n",*argv);
      return 1;
   }
   fsig=atof(argv[3]);
   OPTS=atoi(argv[4]);
   if (argc==10) {
      MMIN=MMAX=atoi(argv[5]);
      YMIN2=atoi(argv[6]);
      YMAX2=atoi(argv[7]);
      XMIN2=atoi(argv[8]);
      XMAX2=atoi(argv[9]);
   }
   if (OPTS&1) rap[0]=2.1;
   else rap[0]=3.;
   rpsf[0]=6;
   sub[0]=5;
   rmin[0]=2.25;
   getpars();
   initrows();
   sprintf(str,"%s.fits",argv[1]);
   if ((f=fopen(str,"rb"))==NULL) {
      printf("****File not found\n");
      return 1;
   }
   while ((!feof(f)) && ((!MODE) || (line%36))) {
      fgets(str,81,f);
      line++;
      if (!MODE) {
	 if (!strncmp(str,"END     ",8)) MODE=1;
	 else if ((!strncmp(str,"SIMPLE  =",9)) && (str[29]!='T')) return ferr(f,"Illegal format type");
	 else if ((!strncmp(str,"BITPIX  =",9)) && (strncmp(str+27,"-32",3))) return ferr(f,"Illegal bits per pixel");
	 else if ((!strncmp(str,"NAXIS   =",9)) && (strncmp(str+27,"  3",3))) return ferr(f,"Illegal number of axes");
	 else if (!strncmp(str,"NAXIS",5)) {
	    tmp=atoi(str+9);
	    if (str[5]=='1') X=tmp;
	    else if (str[5]=='2') Y=tmp;
	    else if (str[5]=='3') Z=tmp;
	 }
	 else if (!strncmp(str,"ATODGAIN=",9)) GAIN=atof(str+9);
	 else if (!strncmp(str,"RNOISE  =",9)) RN=atof(str+9);
	 else if (!strncmp(str,"EPOCH   =",9)) {
	    YEAR=(atof(str+9)-49462.5)/365.25;
	    if (YEAR<0) YEAR=0;
	 }
	 else if (!strncmp(str,"EXPTIME =",9)) EXP=atof(str+9);
	 else if (!strncmp(str,"EXP0    =",9)) EXP0=atof(str+9);
	 else if (!strncmp(str,"RSDPFILL=",9)) DMIN=atoi(str+9);
	 else if (!strncmp(str,"SATURATE=",9)) DMAX=atoi(str+9);
	 else if (!strncmp(str,"TEMP    =",9)) TEMP=atoi(str+9);
	 else if (!strncmp(str,"FILTNAM1=",9)) {
	    int i=0;
	    while ((i<19) && (str[i+11]!=32)) {
	       FILT[i]=str[i+11];
	       i++;
	    }
	 }
	 else if (!strncmp(str,"HSTPHOT =",9)) HPP=1;
      }
   }
   if (!HPP) return ferr(f,"Not pre-processed by mask");
   if ((X!=800) || (Y!=800) || (Z!=4)) return ferr(f,"Illegal axis size");
   printf("Reading FITS file:\n");
   printf("Exp time=%fs, Gain=%f e/ADU, Noise=%f ADU\n",EXP,GAIN,RN);
   printf("Saturation at %d, min value=%d, temp=%d\n",DMAX,DMIN,TEMP);
   printf("Filter=\"%s\"\n",FILT);
   initfilters();
   filt=findfilt(FILT);
   sprintf(str,".%s",filters[filt].suffix);
   if (strcmp(str,argv[2]+strlen(argv[2])-strlen(str))) {
      printf("****Output filename should end with %s\n",str);
      exit(-1);
   }
   of=fopen(argv[2],"w");
   if (!of) {
      printf("****Error opening %s\n",argv[2]);
      exit(-1);
   }
   sprintf(str,"%s.psf",argv[2]);
   fpsf=fopen(str,"wb");
   if (!fpsf) {
      printf("****Error opening %s\n",str);
      exit(-1);
   }
   if (GAIN==7) {
      corr[0]/=1.987;
      corr[1]/=2.003;
      corr[2]/=2.006;
      corr[3]/=1.955;
      IGN=0;
   }
   else if (GAIN!=14) {
      printf("****Illegal Gain\n");
      exit(-1);
   }
   else IGN=1;
   getpsf();
   for (fld=0;fld<4;fld++) {
      ffread(data,sizeof(dtype),1,f);
      poff=(float**)calloc(2*rpsf[fld]+1,sizeof(float*));
      if (!poff) merr();
      poff+=rpsf[fld];
      for (tmp=-rpsf[fld];tmp<=rpsf[fld];tmp++) {
	 poff[tmp]=(float*)calloc(2*rpsf[fld]+1,sizeof(float));
	 if (!poff[tmp]) merr();
	 poff[tmp]+=rpsf[fld];
      }
      if (fld>=MMIN && fld<=MMAX) {
	 if (!fld) printf("** PC **\n");
	 else printf("** WFC%d **\n",fld+1);
	 XMIN=XMIN2-rpsf[fld]-rphot[fld];
	 if (XMIN<50) XMIN=50;
	 XMAX=XMAX2+rpsf[fld]+rphot[fld];
	 if (XMAX>800) XMAX=800;
	 YMIN=YMIN2-rpsf[fld]-rphot[fld];
	 if (YMIN<50) YMIN=50;
	 YMAX=YMAX2+rpsf[fld]+rphot[fld];
	 if (YMAX>800) YMAX=800;
	 for (x=0;x<800;x++) for (y=0;y<800;y++) res[x][y]=data[x][y];
	 initpixmap();
	 sprintf(str,"%s.sky",argv[1]);
	 readsky(str);
	 Nstars=0;
	 NO_SKY=1;
	 find();
	 NO_SKY=0;
	 sortstars();
	 line=0;
	 it=1;
	 while (it && line<MaxIT) {
	    it=solve(++line);
	    if (it>1 && line>MaxIT/2) line=MaxIT/2;
	 }
	 cleansat();
	 outpt(of);
      }
      for (tmp=-rpsf[fld];tmp<=rpsf[fld];tmp++) {
	 ffwrite(poff[tmp]-rpsf[fld],sizeof(float),rpsf[fld]*2+1,fpsf);
	 free(poff[tmp]-rpsf[fld]);
      }
      free(poff-rpsf[fld]);
   }
   fclose(f);
   fclose(of);
   fclose(fpsf);
   return 0;
}
