#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#ifdef LINUX
#include "linux.h"
#else
#include "sun.h"
#endif
#include "filters.h"
#define XMIN 50
#define XMAX 800
#define YMIN 50
#define YMAX 800
#define MMIN 0

typedef float dtype[800][800];
typedef struct{int x,y,fx,fy,type;double s,aps,sig,ffx,ffy;} stype;
typedef float ****psftype;

stype *stars,*sstars;
int X,Y,Z,DMAX,BAD,RMODE=0;
dtype data,rec;
psftype psf[4][8][8];
char FILT[20]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
float GAIN=7,RN;
int rpsf[4],sub[4],fld,Gauss_PSF=0,filt;

void merr(int i) {
   printf("Mem error %d\n",i);
   exit(0);
}

void getpars(void) {
   rpsf[1]=rpsf[2]=rpsf[3]=(rpsf[0]*2+1)/3;
   sub[1]=sub[2]=sub[3]=sub[0]*2;
   stars=(stype*)calloc(1,sizeof(stype));
   if (stars==NULL) merr(0);
}

void readsky(char *fn) {
   int f,x,y;
   FILE *sf;

   if ((sf=fopen(fn,"rb"))==NULL) {
      printf("Cannot open file %s\n",fn);
      exit(-1);
   }
   if (RMODE==0) for (f=0;f<=fld;f++) ffread(rec,800*800,4,sf);
   else for (x=0;x<800;x++) for (y=0;y<800;y++) rec[x][y]=0;
   for (x=XMIN;x<XMAX;x++) for (y=YMIN;y<YMAX;y++) data[x][y]-=rec[x][y];
   fclose(sf);
}

void getpsf(void) {
   int f,x,y,z,rad,px,py;

   for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
      rad=(sub[f]+2)/2;
      psf[f][px][py]=(float****)calloc(rad*2+1,sizeof(float***));
      if (psf[f][px][py]==NULL) merr(1);
      psf[f][px][py]+=rad;
      for (x=-rad;x<=rad;x++) {
	 psf[f][px][py][x]=(float***)calloc(rad*2+1,sizeof(float**));
	 if (psf[f][px][py][x]==NULL) merr(2);
	 psf[f][px][py][x]+=rad;
	 for (y=-rad;y<=rad;y++) {
	    psf[f][px][py][x][y]=(float**)calloc(2*rpsf[f]+1,sizeof(float*));
	    if (psf[f][px][py][x][y]==NULL) merr(3);
	    psf[f][px][py][x][y]+=rpsf[f];
	    for (z=-rpsf[f];z<=rpsf[f];z++) {
	       psf[f][px][py][x][y][z]=(float*)calloc(2*rpsf[f]+1,sizeof(float));
	       if (psf[f][px][py][x][y][z]==NULL) merr(4);
	       psf[f][px][py][x][y][z]+=rpsf[f];
	    }
	 }
      }
   }
   if (Gauss_PSF) {
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 int x1,y1;
	 float sig,n;

	 rad=(sub[f]+2)/2;
	 if (f) sig=0.54;
	 else sig=0.84;
	 for (x=-rad;x<=rad;x++) for (y=-rad;y<=rad;y++) {
	    n=0;
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) n+=(psf[f][px][py][x][y][x1][y1]=exp(-.5*((x1-x/(float)sub[f])*(x1-x/(float)sub[f])+(y1-y/(float)sub[f])*(y1-y/(float)sub[f]))/sig/sig));
	    for (x1=-rpsf[f];x1<=rpsf[f];x1++) for (y1=-rpsf[f];y1<=rpsf[f];y1++) psf[f][px][py][x][y][x1][y1]/=n;
	 }
      }
   }
   else {
      FILE *pf;
      int x1;
      char fn[80];

      sprintf(fn,"%s/psfs/%s.%d.%d.psf",BASEDIR,filters[filt].suffix,rpsf[0],sub[0]);
      if ((pf=fopen(fn,"rb"))==NULL) {
	 printf("File error: %s\n",fn);
	 exit(0);
      }
      for (f=0;f<4;f++) for (px=0;px<8;px++) for (py=0;py<8;py++) {
	 for (x=-(sub[f]+2)/2;x<=(sub[f]+2)/2;x++) for (y=-(sub[f]+2)/2;y<=(sub[f]+2)/2;y++) for (x1=-rpsf[f];x1<=rpsf[f];x1++) ffread(psf[f][px][py][x][y][x1]-rpsf[f],2*rpsf[f]+1,sizeof(float),pf);
      }
      fclose(pf);
   }
}

void psfcorr(char *str) {
   char fn[81];
   FILE *fpsf;
   int f,x,y,x2,y2,x3,y3,rad;
   float c[101];

   sprintf(fn,"%s.psf",str);
   if ((fpsf=fopen(fn,"rb"))==NULL) return;
   for (f=0;f<4;f++) for (x=-rpsf[f];x<=rpsf[f];x++) {
      rad=(sub[f]+2)/2;
      ffread(c,sizeof(float),2*rpsf[f]+1,fpsf);
      for (y=-rpsf[f];y<=rpsf[f];y++) for (x2=0;x2<8;x2++) for (y2=0;y2<8;y2++) for (x3=-rad;x3<=rad;x3++) for (y3=-rad;y3<=rad;y3++) psf[f][x2][y2][x3][y3][x][y]+=c[y+rpsf[f]];
   }
   fclose(fpsf);
}

void imgsub(long i) {
   int x1,y1,x0,y0;
   double q,mx,my;

   if (stars[i].type==5) return;
   if (stars[i].type==4) {
      data[stars[i].x][stars[i].y]-=stars[i].s;
      rec[stars[i].x][stars[i].y]+=stars[i].s;
      return;
   }
   if (stars[i].ffx<0) {
      x0=stars[i].fx-1;
      mx=stars[i].ffx+1;
   }
   else {
      x0=stars[i].fx;
      mx=stars[i].ffx;
   }
   if (stars[i].ffy<0) {
      y0=stars[i].fy-1;
      my=stars[i].ffy+1;
   }
   else {
      y0=stars[i].fy;
      my=stars[i].ffy;
   }
   for (x1=stars[i].x-rpsf[fld];x1<=stars[i].x+rpsf[fld];x1++) for (y1=stars[i].y-rpsf[fld];y1<=stars[i].y+rpsf[fld];y1++) if (x1>=XMIN && x1<XMAX && y1>=YMIN && y1<YMAX) if (data[x1][y1]>BAD && data[x1][y1]<DMAX) {
      q=psf[fld][stars[i].y/100][stars[i].x/100][x0][y0][x1-stars[i].x][y1-stars[i].y]*(1-mx)*(1-my)+psf[fld][stars[i].y/100][stars[i].x/100][x0+1][y0][x1-stars[i].x][y1-stars[i].y]*mx*(1-my)+psf[fld][stars[i].y/100][stars[i].x/100][x0][y0+1][x1-stars[i].x][y1-stars[i].y]*(1-mx)*my+psf[fld][stars[i].y/100][stars[i].x/100][x0+1][y0+1][x1-stars[i].x][y1-stars[i].y]*mx*my;
      rec[x1][y1]+=(q=stars[i].s*q);
      if (RMODE==0) data[x1][y1]-=q;
   }
}

void readstars(char *fn) {
   FILE *f;
   int i,j;
   double x,y,q,tp=1;
   if ((f=fopen(fn,"r"))==NULL) {
      printf("Cannot find %s\n",fn);
      exit(0);
   }
   fscanf(f,"%d",&i);
   while (i!=fld && !feof(f)) {
      for (j=0;j<10;j++) fscanf(f,"%lf",&x);
      fscanf(f,"%d",&i);
   }
   while (i==fld && !feof(f)) {
      if (!RMODE) fscanf(f,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",&y,&x,&stars[0].s,&q,&q,&q,&q,&q,&q,&tp);
      else fscanf(f,"%lf %lf %lf",&y,&x,&stars[0].s);
      stars[0].x=(int)x;
      stars[0].fx=(int)((x-stars[0].x-.5)*sub[fld]+20.5)-20;
      stars[0].ffx=(x-stars[0].x-.5)*sub[fld]-stars[0].fx;
      stars[0].y=(int)y;
      stars[0].fy=(int)((y-stars[0].y-.5)*sub[fld]+20.5)-20;
      stars[0].ffy=(y-stars[0].y-.5)*sub[fld]-stars[0].fy;
      stars[0].type=tp;
      imgsub(0);
      fscanf(f,"%d",&i);
   }
   fclose(f);
}

void addnoise(void) {
   int x,y,i,ct;
   float z;
   for (x=0;x<800;x++) for (y=0;y<800;y++) {
      z=(rec[x][y]*GAIN+RN)*2;
      ct=0;
      for (i=0;i<(int)z;i++) if (rand()/256%2) ct++;
      if (rand()/250%2) rec[x][y]=(z-(int)z+ct-RN)/GAIN;
      else rec[x][y]=(ct-RN)/GAIN;
   }
}

int ferr(FILE *f,char *str) {
   printf("%s\n",str);
   fclose(f);
   return 1;
}

int main(int argc,char**argv) {
   FILE *f,*fres,*frec;
   char str[82];
   int MODE=0,line=0,tmp,x,y,HPP=0;

   if (argc!=5 && argc!=6) {
      printf("Usage: %s <fits base> <photometry> <residual> <reconstructed> <<noise>>\n",*argv);
      printf("       %s <fits base> <fake photometry> \"\" <new image> <<noise>>\n",*argv);
      return 1;
   }
   rpsf[0]=6;
   sub[0]=5;
   getpars();
   sprintf(str,"%s.fits",argv[1]);
   if ((f=fopen(str,"rb"))==NULL) {
      printf("%s not found\n",str);
      return 1;
   }
   if (argv[3][0]==0) {
      fres=NULL;
      RMODE=1;
   }
   else if ((fres=fopen(argv[3],"wb"))==NULL) {
      printf("Cannot open %s\n",argv[3]);
      return 1;
   }
   if ((frec=fopen(argv[4],"wb"))==NULL) {
      printf("Cannot open %s\n",argv[4]);
      return 1;
   }
   while ((!feof(f)) && ((!MODE) || (line%36))) {
      fgets(str,81,f);
      line++;
      if (!MODE) {
	 if (!strncmp(str,"END     ",8)) MODE=1;
	 else if ((!strncmp(str,"SIMPLE  =",9)) && (str[29]!='T')) return ferr(f,"Illegal format type");
	 else if ((!strncmp(str,"BITPIX  =",9)) && (strncmp(str+27,"-32",3))) return ferr(f,"Illegal bits per pixel");
	 else if ((!strncmp(str,"NAXIS   =",9)) && (strncmp(str+27,"  3",3))) return ferr(f,"Illegal number of axes");
	 else if (!strncmp(str,"NAXIS",5)) {
	    tmp=atoi(str+9);
	    if (str[5]=='1') X=tmp;
	    else if (str[5]=='2') Y=tmp;
	    else if (str[5]=='3') Z=tmp;
	 }
	 else if (!strncmp(str,"ATODGAIN=",9)) GAIN=atof(str+9);
	 else if (!strncmp(str,"RNOISE  =",9)) RN=atof(str+9);
	 else if (!strncmp(str,"RSDPFILL=",9)) BAD=atoi(str+9);
	 else if (!strncmp(str,"SATURATE=",9)) DMAX=atoi(str+9);
	 else if (!strncmp(str,"FILTNAM1=",9)) {
	    int i=0;
	    while ((i<19) && (str[i+11]!=32)) {
	       FILT[i]=str[i+11];
	       i++;
	    }
	 }
	 else if (!strncmp(str,"HSTPHOT =",9)) HPP=1;
      }
   }
   if (!HPP) return ferr(f,"Not pre-processed by mask");
   RN=RN*RN*GAIN*GAIN;
   if ((X!=800) || (Y!=800) || (Z!=4)) return ferr(f,"Illegal axis size");
   printf("Reading FITS file:\n");
   printf("Saturation at %d, bad value=%d\n",DMAX,BAD);
   printf("Filter=\"%s\"\n",FILT);
   initfilters();
   filt=findfilt(FILT);
   getpsf();
   psfcorr(argv[1]);
   if (fres!=NULL) {
      fputs("SIMPLE  =                    T / Standard FITS format                           ",fres);
      fputs("BITPIX  =                  -32 / 32 bit IEEE floating point numbers             ",fres);
      fputs("NAXIS   =                    3 / Number of axes                                 ",fres);
      fputs("NAXIS1  =                  800 /                                                ",fres);
      fputs("NAXIS2  =                  800 /                                                ",fres);
      fputs("NAXIS3  =                    4 /                                                ",fres);
      fputs("EXTEND  =                    T / There may be standard extensions               ",fres);
      fputs("END                                                                             ",fres);
      for (line=8;line<36;line++) fputs("                                                                                ",fres);
   }
   fputs("SIMPLE  =                    T / Standard FITS format                           ",frec);
   fputs("BITPIX  =                  -32 / 32 bit IEEE floating point numbers             ",frec);
   fputs("NAXIS   =                    3 / Number of axes                                 ",frec);
   fputs("NAXIS1  =                  800 /                                                ",frec);
   fputs("NAXIS2  =                  800 /                                                ",frec);
   fputs("NAXIS3  =                    4 /                                                ",frec);
   fputs("EXTEND  =                    T / There may be standard extensions               ",frec);
   fputs("END                                                                             ",frec);
   for (line=8;line<36;line++) fputs("                                                                                ",frec);
   srand(time(NULL));
   for (fld=0;fld<4;fld++) {
      ffread(data,sizeof(dtype),1,f);
      sprintf(str,"%s.sky",argv[1]);
      readsky(str);
      readstars(argv[2]);
      if (fres) ffwrite(data,sizeof(dtype),1,fres);
      if (argc==6 && atoi(argv[5])) addnoise();
      if (RMODE==1) for (x=0;x<800;x++) for (y=0;y<800;y++) {
	 if (data[x][y]<=BAD) rec[x][y]=BAD-1;
	 else if (data[x][y]>=DMAX) rec[x][y]=DMAX+1;
	 else rec[x][y]+=data[x][y];
      }
      ffwrite(rec,sizeof(dtype),1,frec);
   }
   fclose(frec);
   if (fres) fclose(fres);
   fclose(f);
   return 0;
}
